Package: azr
Title: Credential Chain for Seamless 'OAuth 2.0' Authentication to
        'Azure Services'
Version: 0.3.4
Authors@R: 
    person("Pedro", "Baltazar",, "pedrobtz@gmail.com", role = c("aut", "cre"))
Description: Implements a credential chain for 'Azure OAuth 2.0' authentication 
      based on the package 'httr2''s 'OAuth' framework. Sequentially attempts authentication 
      methods until one succeeds. During development allows interactive 
      browser-based flows ('Device Code' and 'Auth Code' flows) and non-interactive 
      flow ('Client Secret') in batch mode.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://pedrobtz.github.io/azr/, https://github.com/pedrobtz/azr
BugReports: https://github.com/pedrobtz/azr/issues
Depends: R (>= 4.1)
Imports: utils, R6, cli, httr2 (>= 1.2.2), jsonlite, rlang
Suggests: data.table, httpuv, clipr, processx, testthat (>= 3.0.0), vcr
        (>= 2.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-09 04:33:57 UTC; pbtz
Author: Pedro Baltazar [aut, cre]
Maintainer: Pedro Baltazar <pedrobtz@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-09 16:30:30 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,redacted)
S3method(print,redacted)
export(AuthCodeCredential)
export(AzureCLICredential)
export(CachedTokenCredential)
export(ClientSecretCredential)
export(DefaultCredential)
export(DeviceCodeCredential)
export(RefreshTokenCredential)
export(api_client)
export(api_resource)
export(api_service)
export(api_storage_client)
export(az_cli_account_show)
export(az_cli_get_cached_token)
export(az_cli_get_token)
export(az_cli_is_login)
export(az_cli_login)
export(az_cli_logout)
export(azr_graph_client)
export(azr_storage_client)
export(cached_token_credential_chain)
export(credential_chain)
export(default_azure_client_id)
export(default_azure_client_secret)
export(default_azure_config_dir)
export(default_azure_host)
export(default_azure_oauth_client)
export(default_azure_scope)
export(default_azure_tenant_id)
export(default_azure_url)
export(default_credential_chain)
export(default_non_auth)
export(default_redirect_uri)
export(default_refresh_token)
export(default_response_handler)
export(get_credential_auth)
export(get_credential_provider)
export(get_request_authorizer)
export(get_token)
export(get_token_provider)
export(is_hosted_session)
importFrom(rlang,"%||%")

Credential <- R6::R6Class(
  classname = "Credential",
  public = list(
    .id = NULL,
    .name = NULL,
    .scope = NULL,
    .scope_str = NULL,
    .resource = NULL,
    .client_id = NULL,
    .client_secret = NULL,
    .tenant_id = NULL,
    .use_cache = "disk",
    .cache_key = NULL,
    .oauth_client = NULL,
    .oauth_host = NULL,
    .oauth_endpoint = NULL,
    .oauth_url = NULL,
    .token_url = NULL,
    .redirect_uri = NULL,
    .classname = NULL,
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = NULL,
      client_secret = NULL,
      use_cache = c("disk", "memory"),
      offline = FALSE,
      oauth_endpoint = NULL,
      name = NULL
    ) {
      if (!rlang::is_interactive() && self$is_interactive()) {
        cli::cli_abort(
          "Credential {.cls {class(self)[[1]]}} requires an interactive session"
        )
      }

      self$.classname <- paste(class(self), collapse = "/")

      self$.scope <- scope %||% default_azure_scope(resource = "azure_arm")

      if (isTRUE(offline)) {
        self$.scope <- unique(c(self$.scope, "offline_access"))
      }

      self$.scope_str <- collapse_scope(self$.scope)
      self$.resource <- get_scope_resource(self$.scope)

      self$.client_id <- client_id %||% default_azure_client_id()
      self$.client_secret <- client_secret %||% default_azure_client_secret()

      self$.tenant_id <- tenant_id %||% default_azure_tenant_id()
      self$.use_cache <- rlang::arg_match(use_cache)

      self$.cache_key <- list(
        client_id = self$.client_id,
        tenant_id = self$.tenant_id,
        scope = self$.scope,
        classname = self$.classname
      )
      self$.id <- rlang::hash(self$.cache_key)

      self$.name <- name %||% self$.id

      self$.oauth_host <- default_azure_host()
      self$.token_url <- default_azure_url(
        endpoint = "token",
        oauth_host = self$.oauth_host,
        tenant_id = self$.tenant_id
      )

      self$.oauth_endpoint <- oauth_endpoint %||% self$.oauth_endpoint

      if (!is.null(self$.oauth_endpoint)) {
        self$.oauth_url <- default_azure_url(
          endpoint = self$.oauth_endpoint,
          oauth_host = self$.oauth_host,
          tenant_id = self$.tenant_id
        )
      }

      self$.oauth_client <- httr2::oauth_client(
        name = self$.name,
        id = self$.client_id,
        secret = self$.client_secret,
        token_url = self$.token_url,
        auth = "body"
      )

      self$validate()

      # Lock all public fields to prevent modification
      lockBinding(".id", self)
      lockBinding(".name", self)
      lockBinding(".scope", self)
      lockBinding(".scope_str", self)
      lockBinding(".resource", self)
      lockBinding(".client_id", self)
      lockBinding(".client_secret", self)
      lockBinding(".tenant_id", self)
      lockBinding(".use_cache", self)
      lockBinding(".cache_key", self)
      lockBinding(".oauth_client", self)
      lockBinding(".oauth_host", self)
      lockBinding(".oauth_endpoint", self)
      lockBinding(".oauth_url", self)
      lockBinding(".token_url", self)
      lockBinding(".classname", self)
    },
    validate = function() {
      validate_scope(self$.scope)
      validate_tenant_id(self$.tenant_id)
      invisible(self)
    },
    is_interactive = function() {
      FALSE
    },
    print = function() {
      cli::cli_text(cli::style_bold(
        "<",
        paste(class(self), collapse = "/"),
        ">"
      ))

      nms <- r6_get_public_fields(
        cls = r6_get_class(
          self,
          envir = getNamespace(methods::getPackageName())
        )
      )

      pfields <- rlang::env_get_list(env = self, nms = nms)
      names(pfields) <- sub("^\\.", "", names(pfields))

      # Filter out NULL/empty values and redact sensitive fields
      pfields <- Filter(length, pfields)
      redacted <- list_redact(pfields, c("client_secret", "key"))

      bullets(redacted)
      invisible(self)
    }
  )
)


credential_chain <- function(...) {
  res <- rlang::enquos(...)
  class(res) <- c("credential_chain", class(res))
  res
}


collapse_scope <- function(scope) {
  paste(scope, collapse = " ")
}

.login_cache <- new.env(parent = emptyenv())

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(cli::format_inline(
    "{.pkg azr} {utils::packageVersion('azr')} | Azure OAuth 2.0 credential chain"
  ))

  packageStartupMessage(cli::format_bullets_raw(c(
    i = "Environment configuration:"
  )))
  for (bullet in cli::format_bullets_raw(get_env_config())) {
    packageStartupMessage(bullet)
  }
}

#' Create a Microsoft Graph API Client
#'
#' @description
#' Creates a configured client for the Microsoft Graph API with authentication
#' and versioned endpoints (v1.0 and beta). This function returns an
#' [api_service] object that provides access to Microsoft Graph resources
#' through versioned endpoints.
#'
#' @details
#' The function creates a Microsoft Graph service using these components:
#'
#' - **[api_client]**: A general-purpose API client configured with the Graph API
#'   host URL (`https://graph.microsoft.com`) and authentication provider.
#'
#' - **[api_graph_resource]**: A specialized resource class that extends [api_resource]
#'   with Microsoft Graph-specific methods. Currently implements:
#'   - `me(select = NULL)`: Fetch the current user's profile. The `select` parameter
#'     accepts a character vector of properties to return (e.g., `c("displayName", "mail")`).
#'
#' - **[api_service]**: A service container that combines the client and resources
#'   with versioned endpoints (v1.0 and beta). The service is locked using
#'   [lockEnvironment()] to prevent modification after creation.
#'
#' @param scopes A character string specifying the OAuth2 scope suffix to be appended
#'   to the Graph API URL. Defaults to `".default"`, which requests all permissions
#'   the app has been granted. The full scope will be `https://graph.microsoft.com/{scopes}`.
#' @param chain A [credential_chain] instance for authentication. If NULL,
#'   a default credential chain will be created using [DefaultCredential].
#' @param ... Additional arguments passed to the [api_client] constructor.
#'
#' @return An [api_service] object configured for Microsoft Graph API with
#'   v1.0 and beta endpoints. The object is locked using [lockEnvironment()] to
#'   prevent modification after creation. Access endpoints via `$v1.0` or `$beta`.
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a Graph API client with default credentials
#' graph <- azr_graph_client()
#'
#' # Fetch current user profile from v1.0 endpoint
#' me <- graph$v1.0$me()
#'
#' # Fetch specific properties using OData $select
#' me <- graph$v1.0$me(select = c("displayName", "mail", "userPrincipalName"))
#'
#' # Use beta endpoint for preview features
#' me_beta <- graph$beta$me(select = c("displayName", "mail"))
#'
#' # Create with a custom credential chain
#' custom_chain <- credential_chain(
#'   AzureCLICredential$new(scope = "https://graph.microsoft.com/.default")
#' )
#' graph <- azr_graph_client(chain = custom_chain)
#'
#' # Use specific scopes instead of .default
#' graph <- azr_graph_client(scopes = "User.Read Mail.Read")
#' }
# azr_graph_client ----
azr_graph_client <- function(scopes = ".default", ..., chain = NULL) {
  graph_url <- "https://graph.microsoft.com"

  # Construct the full scope URL
  if (length(scopes) > 1) {
    scopes <- paste(scopes, collapse = " ")
  }
  scope <- paste0(graph_url, "/", scopes)

  provider <- DefaultCredential$new(
    scope = scope,
    chain = chain
  )

  client <- api_client$new(
    host_url = graph_url,
    provider = provider,
    ...
  )

  service <- api_service$new(
    client = client,
    chain = chain,
    endpoints = list(
      "v1.0" = api_graph_resource,
      "beta" = api_graph_resource
    )
  )

  lockEnvironment(service, bindings = TRUE)

  service
}


#' Microsoft Graph API Resource
#'
#' @description
#' An R6 class that extends [api_resource] to provide specialized methods
#' for the Microsoft Graph API. This class adds convenience methods for
#' common Graph operations.
#'
#' @keywords internal
# api_graph_resource ----
api_graph_resource <- R6::R6Class(
  classname = "api_graph_resource",
  inherit = api_resource,
  public = list(
    #' @description
    #' Fetch the current user's profile
    #'
    #' @param select A character vector of properties to select (e.g., c("displayName", "mail")).
    #'   If NULL, all properties are returned.
    #'
    #' @return The response from the /me endpoint
    me = function(select = NULL) {
      if (!is.null(select)) {
        if (!is.character(select) || length(select) == 0) {
          cli::cli_abort("{.arg select} must be a non-empty character vector.")
        }
        query <- list(`$select` = paste(select, collapse = ","))
      } else {
        query <- NULL
      }

      self$.client$.fetch("me", query = query)
    }
  )
)

#' Azure API Resource
#'
#' @description
#' An R6 class that wraps an `api_client` and adds an additional path segment
#' (like "beta" or "v1.0") to all requests. This is useful for APIs that version
#' their endpoints or have different API surfaces under different paths.
#'
#' @details
#' The `api_resource` class creates a modified base request by appending an
#' endpoint path to the client's base request. All subsequent API calls through
#' this resource will automatically include this path prefix.
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a client
#' client <- api_client$new(
#'   host_url = "https://graph.microsoft.com"
#' )
#'
#' # Create a resource with v1.0 API endpoint
#' resource_v1 <- api_resource$new(
#'   client = client,
#'   endpoint = "v1.0"
#' )
#'
#' # Create a resource with beta API endpoint
#' resource_beta <- api_resource$new(
#'   client = client,
#'   endpoint = "beta"
#' )
#'
#' # Make requests - the endpoint is automatically prepended
#' response <- resource_v1$.fetch(
#'   path = "/me",
#'   method = "get"
#' )
#' }
api_resource <- R6::R6Class(
  classname = "api_resource",
  # > private ----
  private = list(
    endpoint = NULL
  ),
  # > public ----
  public = list(
    #' @field .client The cloned api_client instance with modified base_req
    .client = NULL,
    #' @description
    #' Create a new API resource instance
    #'
    #' @param client An `api_client` object that provides the base HTTP client
    #'   functionality. This will be cloned to avoid modifying the original.
    #' @param endpoint A character string specifying the API endpoint or path
    #'   segment to append (e.g., `"v1.0"`, `"beta"`).
    #'
    #' @return A new `api_resource` object
    initialize = function(client, endpoint) {
      if (is.null(client)) {
        cli::cli_abort("{.arg client} must not be {.val NULL}.")
      }
      if (!R6::is.R6(client)) {
        cli::cli_abort("{.arg client} must be an R6 object.")
      }
      if (!inherits(client, "api_client")) {
        cli::cli_abort("{.arg client} must be an {.cls api_client} object.")
      }

      # If endpoint is already set, do nothing (ignore the argument)
      if (!is.null(private$endpoint)) {
        # Do nothing - keep existing endpoint
      } else {
        if (is.null(endpoint)) {
          cli::cli_abort("{.arg endpoint} must not be {.val NULL}.")
        }
        if (!is.character(endpoint)) {
          cli::cli_abort("{.arg endpoint} must be a character string.")
        }
        if (length(endpoint) != 1L) {
          cli::cli_abort(
            "{.arg endpoint} must be a single character string, not length {length(endpoint)}."
          )
        }
        if (!nzchar(endpoint)) {
          cli::cli_abort("{.arg endpoint} must not be an empty string.")
        }

        private$endpoint <- endpoint
      }

      # Clone the client and modify its base_req to include the endpoint path
      self$.client <- client$clone()
      self$.client$.base_req <- self$.client$.base_req |>
        httr2::req_url_path_append(private$endpoint)

      # Lock all fields to prevent modification
      lockBinding(".client", self)
      lockBinding("endpoint", private)
    }
  )
)

#' Detect if running in a hosted session
#'
#' Determines whether the current R session is running in a hosted environment
#' such as Google Colab, VS Code, Kubernetes, or RStudio Server (non-localhost).
#'
#' @return A logical value: `TRUE` if running in a hosted session (Google Colab,
#'   VS Code, Kubernetes, or remote RStudio Server), `FALSE` otherwise.
#'
#' @details
#' This function checks for (in order):
#' * Option override: if `azr.hosted` option is set, returns `isTRUE()` of its value
#'
#' * Google Colab: presence of the `COLAB_RELEASE_TAG` environment variable
#' * VS Code: presence of the `VSCODE_INJECTION` or `VSCODE_PROXY_URI` environment variable
#' * Kubernetes: presence of the `KUBERNETES_SERVICE_HOST` environment variable
#' * RStudio Server: `RSTUDIO_PROGRAM_MODE` is "server" and
#'   `RSTUDIO_HTTP_REFERER` does not contain "localhost"
#'
#' @export
#' @examples
#' if (is_hosted_session()) {
#'   message("Running in a hosted environment")
#' }
is_hosted_session <- function() {
  # Check for user override via option
  opt <- getOption("azr.hosted")
  if (!is.null(opt)) {
    return(isTRUE(opt))
  }

  # Check for Google Colab
  if (nzchar(Sys.getenv("COLAB_RELEASE_TAG"))) {
    return(TRUE)
  }

  # Check for VS Code
  if (nzchar(Sys.getenv("VSCODE_INJECTION"))) {
    return(TRUE)
  }

  if (nzchar(Sys.getenv("VSCODE_PROXY_URI"))) {
    return(TRUE)
  }

  # Check for Kubernetes
  if (nzchar(Sys.getenv("KUBERNETES_SERVICE_HOST"))) {
    return(TRUE)
  }

  # Check for RStudio Server (non-localhost)
  Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server" &&
    !grepl("localhost", Sys.getenv("RSTUDIO_HTTP_REFERER"), fixed = TRUE)
}


bullets <- function(x) {
  as_simple <- function(x) {
    if (is.atomic(x) && length(x) == 1) {
      if (is.character(x)) {
        paste0("\"", x, "\"")
      } else {
        format(x)
      }
    } else {
      if (inherits(x, "redacted")) {
        format(x)
      } else {
        paste0("<", class(x)[[1L]], ">")
      }
    }
  }

  vals <- vapply(x, as_simple, character(1))
  names <- format(names(x))
  names <- gsub(" ", "\u00a0", names, fixed = TRUE) # non-breaking space

  for (i in seq_along(x)) {
    cli::cli_li("{.field {names[[i]]}}: {vals[[i]]}")
  }
  invisible(NULL)
}


list_redact <- function(x, names, case_sensitive = TRUE) {
  if (case_sensitive) {
    i <- match(names, names(x))
  } else {
    i <- match(tolower(names), tolower(names(x)))
  }
  i <- i[!is.na(i)]
  i <- setdiff(i, which(is_empty_vec(x)))
  x[i] <- list(redacted())
  x
}


redacted <- function() {
  structure(list(), class = "redacted")
}


#' @exportS3Method format redacted
format.redacted <- function(x, ...) {
  cli::col_grey("<REDACTED>")
}


#' @exportS3Method print redacted
print.redacted <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
  invisible(x)
}

is_port_available <- function(port, host = "127.0.0.1") {
  # Try to connect to the port (if something is listening, connection succeeds)
  suppressWarnings(
    tryCatch(
      {
        con <- socketConnection(
          host = host,
          port = port,
          server = FALSE,
          blocking = TRUE,
          open = "r+",
          timeout = 1
        )
        close(con)
        # If connection succeeds, port is IN USE
        FALSE
      },
      error = function(e) {
        # If connection fails, port is AVAILABLE
        TRUE
      }
    )
  )
}

token_has_expired <- function(token, delay = 5) {
  if (!inherits(token, "httr2_token")) {
    cli::cli_abort(
      "{.arg token} must be an {.cls httr2_token} object, not {.cls {class(token)[[1L]]}}.",
      call = rlang::caller_env()
    )
  }
  if (is.null(token$expires_at)) {
    FALSE
  } else {
    (as.integer(Sys.time()) + delay) > token$expires_at
  }
}


random_port <- function(
  min = 10000L,
  max = 49151L,
  host = "127.0.0.1",
  n = 20
) {
  min <- max(1L, min)
  max <- min(max, 65535L)
  try_ports <- sample(x = seq.int(min, max), n)
  for (port in try_ports) {
    if (is_port_available(port, host)) {
      return(port)
    }
  }
  cli::cli_abort("Cannot find an available port.")
}

#' API Service Base Class
#'
#' @description
#' Base R6 class for creating API service wrappers. This class provides a
#' foundation for building service-specific API clients with authentication,
#' endpoint management, and configuration.
#'
#' @export
api_service <- R6::R6Class(
  classname = "api_service",
  lock_objects = FALSE,
  cloneable = FALSE,
  private = list(
    .chain = NULL,
    .endpoints = NULL,
    .config = NULL
  ),
  public = list(
    #' @field .client An [api_client] instance for making API requests
    .client = NULL,

    #' @description
    #' Create a new API service instance
    #'
    #' @param client An [api_client] instance. If `NULL`, a new client will be created.
    #' @param chain A [credential_chain] instance for authentication. Optional.
    #' @param endpoints A named list where names are endpoint paths (e.g., "v1.0", "beta")
    #'   and values are R6 class objects (not instances) to use for creating resources.
    #'   Defaults to an empty list. If the value is `NULL`, [api_resource] will be used.
    #' @param config A list of configuration options. Defaults to an empty list.
    #'
    #' @return A new `api_service` object
    initialize = function(
      client = NULL,
      chain = NULL,
      endpoints = list(),
      config = list()
    ) {
      self$.client <- client
      private$.chain <- chain
      private$.config <- config

      # Validate that endpoints is a list
      if (!is.list(endpoints)) {
        cli::cli_abort("{.arg endpoints} must be a list.")
      }

      # Get endpoint paths (names)
      endpoint_paths <- names(endpoints)

      # Validate endpoint paths
      validate_endpoint_paths(endpoint_paths)

      private$.endpoints <- endpoints

      # Loop over endpoints and add api_resources
      if (length(endpoints) > 0) {
        for (i in seq_along(endpoints)) {
          endpoint_path <- endpoint_paths[i]
          resource_class <- endpoints[[i]]

          # If resource_class is NULL, use default api_resource
          if (is.null(resource_class)) {
            resource_class <- api_resource
          }

          # Check if resource_class is already an R6 object instance
          if (R6::is.R6(resource_class)) {
            # Already an R6 object, assign directly
            self[[endpoint_path]] <- resource_class
          } else {
            # Validate that resource_class is an R6ClassGenerator
            if (!R6::is.R6Class(resource_class)) {
              cli::cli_abort(
                "Endpoint {.val {endpoint_path}} must specify an R6 class, not {.cls {class(resource_class)}}."
              )
            }

            # Create resource instance for this endpoint
            resource <- resource_class$new(
              client = self$.client,
              endpoint = endpoint_path
            )

            # Add as a field with the endpoint path name
            self[[endpoint_path]] <- resource
          }

          # Lock the endpoint field to make it read-only
          lockBinding(endpoint_path, self)
        }
      }

      # Lock all fields to make them read-only
      lockBinding(".client", self)
      lockBinding(".chain", private)
      lockBinding(".endpoints", private)
      lockBinding(".config", private)
    }
  )
)


#' Validate Endpoint Paths
#'
#' @description
#' Validates that endpoint paths (names of the endpoints list) are single strings
#' without spaces and contain only valid URL path characters.
#'
#' @param endpoint_paths A character vector of endpoint path strings to validate
#'
#' @return Invisibly returns `NULL` if validation passes, otherwise throws an error
#'
#' @noRd
validate_endpoint_paths <- function(endpoint_paths) {
  if (length(endpoint_paths) == 0) {
    return(invisible(NULL))
  }

  for (i in seq_along(endpoint_paths)) {
    endpoint_path <- endpoint_paths[i]

    # Check if it's a character string
    if (!is.character(endpoint_path) || length(endpoint_path) != 1) {
      cli::cli_abort(
        "Endpoint path at position {i} must be a single character string"
      )
    }

    # Check if it's not empty
    if (!nzchar(endpoint_path)) {
      cli::cli_abort(
        "Endpoint path at position {i} must not be an empty string"
      )
    }

    # Check for spaces
    if (grepl("\\s", endpoint_path)) {
      cli::cli_abort(
        "Endpoint path at position {i} ({.val {endpoint_path}}) contains spaces"
      )
    }

    # Check for valid URL path characters (alphanumeric, hyphens, underscores, dots, slashes)
    if (!grepl("^[a-zA-Z0-9._/-]+$", endpoint_path)) {
      cli::cli_abort(
        c(
          "Endpoint path at position {i} ({.val {endpoint_path}}) contains invalid characters.",
          "i" = "Only alphanumeric, '.', '_', '-', and '/' are allowed."
        )
      )
    }
  }

  invisible(NULL)
}

#' Azure API Client
#'
#' @description
#' An R6 class that provides a base HTTP client for interacting with Azure APIs.
#' This client handles authentication, request building, retry logic, logging,
#' and error handling for Azure API requests.
#'
#' @details
#' The `api_client` class is designed to be a base class for Azure service-specific
#' clients. It provides:
#' - Automatic authentication using Azure credentials
#' - Configurable retry logic with exponential backoff
#' - Request and response logging
#' - JSON, XML, and HTML content type handling
#' - Standardized error handling
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a client with default credentials
#' client <- api_client$new(
#'   host_url = "https://management.azure.com"
#' )
#'
#' # Create a client with a credential provider
#' cred_provider <- get_credential_provider(
#'   scope = "https://management.azure.com/.default"
#' )
#' client <- api_client$new(
#'   host_url = "https://management.azure.com",
#'   provider = cred_provider
#' )
#'
#' # Create a client with custom credentials function
#' client <- api_client$new(
#'   host_url = "https://management.azure.com",
#'   credentials = my_credential_function,
#'   timeout = 120,
#'   max_tries = 3
#' )
#'
#' # Create a client with custom response handler
#' custom_handler <- function(content) {
#'   # Custom processing logic - e.g., keep data frames as-is
#'   content
#' }
#' client <- api_client$new(
#'   host_url = "https://management.azure.com",
#'   response_handler = custom_handler
#' )
#'
#' # Make a GET request
#' response <- client$.fetch(
#'   path = "/subscriptions/{subscription_id}/resourceGroups",
#'   subscription_id = "my-subscription-id",
#'   query = list(`api-version` = "2021-04-01"),
#'   method = "get"
#' )
#' }
api_client <- R6::R6Class(
  # nolint cyclocomp_linter
  classname = "api_client",
  cloneable = TRUE,
  # > public ----
  public = list(
    #' @field .host_url Base URL for the API
    .host_url = NULL,
    #' @field .base_req Base httr2 request object
    .base_req = NULL,
    #' @field .provider Credential provider R6 object
    .provider = NULL,
    #' @field .credentials Credentials function for authentication
    .credentials = NULL,
    #' @field .options Request options (timeout, connecttimeout, max_tries)
    .options = NULL,
    #' @field .response_handler Optional callback function to process response content
    .response_handler = NULL,
    #' @description
    #' Create a new API client instance
    #'
    #' @param host_url A character string specifying the base URL for the API
    #'   (e.g., `"https://management.azure.com"`).
    #' @param provider An R6 credential provider object that inherits from the
    #'   `Credential` or `DefaultCredential` class. If provided, the credential's
    #'   `req_auth` method will be used for authentication. Takes precedence over
    #'   `credentials`.
    #' @param credentials A function that adds authentication to requests. If
    #'   both `provider` and `credentials` are `NULL`, uses [default_non_auth()].
    #'   The function should accept an httr2 request object and return a modified
    #'   request with authentication.
    #' @param timeout An integer specifying the request timeout in seconds.
    #'   Defaults to `60`.
    #' @param connecttimeout An integer specifying the connection timeout in
    #'   seconds. Defaults to `30`.
    #' @param max_tries An integer specifying the maximum number of retry
    #'   attempts for failed requests. Defaults to `5`.
    #' @param response_handler An optional function to process the parsed response
    #'   content. The function should accept one argument (the parsed response) and
    #'   return the processed content. If `NULL`, uses [default_response_handler()]
    #'   which converts data frames to data.table objects. Defaults to `NULL`.
    #'
    #' @return A new `api_client` object
    initialize = function(
      host_url,
      provider = NULL,
      credentials = NULL,
      timeout = 60L,
      connecttimeout = 30L,
      max_tries = 5L,
      response_handler = NULL
    ) {
      if (!missing(host_url)) {
        self$.host_url <- host_url
      }

      self$.options <- list(
        timeout = timeout,
        connecttimeout = connecttimeout,
        max_tries = max_tries
      )
      if (is.null(self$.host_url)) {
        cli::cli_abort("Argument {.arg host_url} must not be NULL.")
      }
      if (length(self$.host_url) != 1L) {
        cli::cli_abort(
          "Argument {.arg host_url} must be a single value, not length {length(self$.host_url)}."
        )
      }
      if (!is.character(self$.host_url)) {
        cli::cli_abort(
          "Argument {.arg host_url} must be a character string, not {.cls {class(self$.host_url)}}."
        )
      }

      # Handle credentials function
      if (!is.null(credentials)) {
        if (!is.function(credentials)) {
          cli::cli_abort(
            "Argument {.arg credentials} must be a function, not {.cls {class(credentials)}}."
          )
        }
        self$.credentials <- credentials
      } else if (!is.null(provider)) {
        # Handle credential provider if provided
        if (
          !R6::is.R6(provider) ||
            !(inherits(provider, "Credential") ||
              inherits(provider, "DefaultCredential"))
        ) {
          cli::cli_abort(
            "Argument 'provider' must be an R6 object that inherits from 'Credential' or 'DefaultCredential' class."
          )
        }
        self$.provider <- provider
        self$.credentials <- function(req) provider$req_auth(req)
      } else {
        self$.credentials <- default_non_auth
      }

      # Handle response handler function
      if (is.null(response_handler)) {
        response_handler <- default_response_handler
      }

      stopifnot(is.function(response_handler))
      self$.response_handler <- response_handler

      self$.base_req <- httr2::request(self$.host_url) |>
        httr2::req_options(
          timeout = timeout,
          connecttimeout = connecttimeout
        ) |>
        httr2::req_retry(
          max_tries = max_tries,
          retry_on_failure = TRUE,
          backoff = function(i) {
            x <- backoff_default(i)
            cli::cli_alert_warning(
              "Request failed. Retrying in {.val {x}} secs. (attempt {i} of {self$.options$max_tries})."
            )
            return(x)
          }
        ) |>
        httr2::req_error(body = function(resp) {
          cli::cli_alert_danger(
            "<<< status = {.val {resp$status}} | time = {format_timing(resp$timing)} secs."
          )
          if (httr2::resp_has_body(resp)) {
            err_content <- httr2::resp_body_string(resp)
            cli::cli_alert_danger("<<< body:")
            cli::cli_verbatim(err_content)
          }
          invisible()
        })

      # Lock all public fields to prevent modification
      lockBinding(".host_url", self)
      lockBinding(".base_req", self)
      lockBinding(".provider", self)
      lockBinding(".credentials", self)
      lockBinding(".options", self)
      lockBinding(".response_handler", self)
    },
    #' @description
    #' Make an HTTP request to the API
    #'
    #' @param path A character string specifying the API endpoint path. Supports
    #'   [rlang::englue()] syntax for variable interpolation using named arguments
    #'   passed via `...`.
    #' @param ... Named arguments used for path interpolation with [rlang::englue()].
    #' @param query A named list of query parameters to append to the URL.
    #' @param body Request body data. Sent as JSON in the request body.
    #'   Can be a list or character string (JSON).
    #' @param headers A named list of additional HTTP headers to include in the
    #'   request.
    #' @param method A character string specifying the HTTP method. One of
    #'   `"get"`, `"post"`, `"put"`, `"patch"`, or `"delete"`. Defaults to `"get"`.
    #' @param verbosity An integer specifying the verbosity level for request
    #'   debugging (passed to [httr2::req_perform()]). Defaults to `0`.
    #' @param content A character string specifying what to return. One of:
    #'   - `"body"` (default): Return the parsed response body
    #'   - `"headers"`: Return response headers
    #'   - `"response"`: Return the full httr2 response object
    #'   - `"request"`: Return the prepared request object without executing it
    #' @param content_type A character string specifying how to parse the response
    #'   body. If `NULL`, uses the response's Content-Type header. Common values:
    #'   `"application/json"`, `"application/xml"`, `"text/html"`.
    #'
    #' @return Depends on the `content` parameter:
    #'   - `"body"`: Parsed response body (list, data.frame, or character)
    #'   - `"headers"`: List of response headers
    #'   - `"response"`: Full [httr2::response()] object
    #'   - `"request"`: [httr2::request()] object
    .fetch = function(
      path,
      ...,
      query = NULL,
      body = NULL,
      headers = NULL,
      method = "get",
      verbosity = 0L,
      content = c("body", "headers", "response", "request"),
      content_type = NULL
    ) {
      content <- match.arg(content)

      req <- self$.build_request(
        path,
        ...,
        query = query,
        body = body,
        headers = headers,
        method = method
      )

      if (content == "request") {
        return(req)
      }

      resp <- self$.send_request(req, verbosity = verbosity)

      self$.resp_content(resp, content = content, content_type = content_type)
    },
    #' @description
    #' Extract content from a response object
    #'
    #' @param resp An [httr2::response()] object
    #' @param content A character string specifying what to return. One of:
    #'   - `"body"`: Return the parsed response body
    #'   - `"headers"`: Return response headers
    #'   - `"response"`: Return the full httr2 response object
    #' @param content_type A character string specifying how to parse the response
    #'   body. Only used when `content = "body"`. If `NULL`, uses the response's
    #'   Content-Type header.
    #'
    #' @return Depends on the `content` parameter:
    #'   - `"body"`: Parsed response body (list, data.frame, or character)
    #'   - `"headers"`: List of response headers
    #'   - `"response"`: Full [httr2::response()] object
    .resp_content = function(resp, content, content_type = NULL) {
      switch(
        content,
        body = self$.resp_body_content(resp, content_type = content_type),
        headers = httr2::resp_headers(resp),
        response = resp
      )
    },
    #' @description
    #' Build an HTTP request object
    #'
    #' @param path A character string specifying the API endpoint path. Supports
    #'   [rlang::englue()] syntax for variable interpolation using named arguments
    #'   passed via `...`.
    #' @param ... Named arguments used for path interpolation with [rlang::englue()].
    #' @param query A named list of query parameters to append to the URL.
    #' @param body Request body data. Sent as JSON in the request body.
    #'   Can be a list or character string (JSON).
    #' @param headers A named list of additional HTTP headers to include in the
    #'   request.
    #' @param method A character string specifying the HTTP method. One of
    #'   `"get"`, `"post"`, `"put"`, `"patch"`, or `"delete"`. Defaults to `"get"`.
    #'
    #' @return An [httr2::request()] object ready for execution
    .build_request = function(
      path,
      ...,
      query = NULL,
      body = NULL,
      headers = NULL,
      method = "get"
    ) {
      path <- rlang::englue(path, env = as.environment(list(...)))

      req <- self$.base_req |>
        httr2::req_url_path_append(path) |>
        httr2::req_method(method)

      if (!is.null(query)) {
        req <- httr2::req_url_query(req, !!!query)
      }

      if (!is.null(body)) {
        if (!is.character(body)) {
          body <- jsonlite::toJSON(
            drop_null(body),
            null = "null",
            auto_unbox = TRUE
          )
        }
        stopifnot(length(body) == 1L)
        req <- httr2::req_body_raw(
          req,
          body = body,
          type = "application/json"
        )
      }

      if (!is.null(headers)) {
        req <- httr2::req_headers(req, !!!headers)
      }

      return(req)
    },
    #' @description
    #' Perform an HTTP request and log the results
    #'
    #' @param req An [httr2::request()] object to execute
    #' @param verbosity An integer specifying the verbosity level for request
    #'   debugging (passed to [httr2::req_perform()]). Defaults to `0`.
    #'
    #' @return An [httr2::response()] object containing the API response
    .send_request = function(req, verbosity) {
      cli::cli_alert_info(">>> {.strong {req$method}} {.url {req$url}}")

      if (!is.null(req$body) && req$body$content_type == "application/json") {
        cli::cli_alert_info(">>> body:")
        cli::cli_verbatim(format_json_body(
          req$body$data,
          params = req$body$params
        ))
      }

      resp <- httr2::req_perform(self$.credentials(req), verbosity = verbosity)

      size <- format_size(resp$body, units = "Kb")
      time <- format_timing(resp$timing)
      cli::cli_alert_success(
        "<<< status = {.val {resp$status}} | time = {time} secs. | size = {size} Kb"
      )

      return(resp)
    },
    #' @description
    #' Extract and parse response content
    #'
    #' @param resp An [httr2::response()] object
    #' @param content_type A character string specifying how to parse the response
    #'   body. If `NULL`, uses the response's Content-Type header. Common values:
    #'   `"application/json"`, `"application/xml"`, `"text/html"`.
    #'
    #' @return Parsed response body. Format depends on content type:
    #'   - JSON: List or data.frame
    #'   - XML: xml2 document
    #'   - HTML: xml2 document
    #'   - Other: Character string
    .resp_body_content = function(resp, content_type = NULL) {
      if (!httr2::resp_has_body(resp)) {
        cli::cli_inform("Response has no body.")
        return(invisible(NULL))
      }

      if (is.null(content_type)) {
        content_type <- httr2::resp_content_type(resp)
      }

      ans <- switch(
        content_type,
        "application/json" = httr2::resp_body_json(
          resp,
          simplifyVector = TRUE,
          flatten = FALSE
        ),
        "application/xml" = httr2::resp_body_xml(resp),
        "text/html" = httr2::resp_body_html(resp),
        httr2::resp_body_string(resp)
      )

      # Apply response handler callback
      ans <- self$.response_handler(ans)

      return(ans)
    },
    #' @description
    #' Get authentication token from the credential provider
    #'
    #' @return An [httr2::oauth_token()] object if a provider is available,
    #'   otherwise returns `NULL` with a warning.
    .get_token = function() {
      if (!is.null(self$.provider)) {
        return(self$.provider$get_token())
      } else {
        cli::cli_warn(
          "No credential provider available. Cannot retrieve token."
        )
        return(invisible(NULL))
      }
    }
  )
)


# utilities ----
backoff_default <- function(i, max_time = 60.0, b = 2.5) {
  round(min(stats::runif(1L, min = 1.0, max = b^i), max_time), digits = 1L)
}

format_timing <- function(timing) {
  y <- timing[["total"]]
  if (is.null(y) || is.na(y)) {
    cli::col_blue("???")
  } else {
    cli::format_inline("{.val {as.numeric(format(y, digits = 3L))}}")
  }
}

format_size <- function(body, units = "Kb") {
  x <- format(utils::object.size(body), units = units)
  cli::col_blue(strsplit(x, split = " ", fixed = TRUE)[[1]][1])
}

format_json_body <- function(x, params = NULL, max_size = 12L, max_items = 6L) {
  inner_format <- function(z) {
    if (length(z) > max_items) {
      z <- c(utils::head(z, 3), list("..."), utils::tail(z, 3))
    }
    lapply(z, function(i) {
      if (is.list(i)) {
        inner_format(i)
      } else if (length(i) > max_size) {
        c(utils::head(i), "...", utils::tail(i))
      } else {
        i
      }
    })
  }

  par_auto_box <- params$auto_unbox %||% TRUE
  par_null <- params$null %||% "null"

  if (is.character(x)) {
    x <- jsonlite::fromJSON(
      x,
      simplifyDataFrame = FALSE,
      simplifyMatrix = FALSE
    )
  }

  res <- jsonlite::toJSON(
    inner_format(x),
    pretty = TRUE,
    auto_unbox = par_auto_box,
    null = par_null
  )
  res <- gsub("\\", "...", res, fixed = TRUE)
  res <- gsub('"..."', "...", res, fixed = TRUE)

  return(res)
}

#' Default No Authentication
#'
#' @description
#' A pass-through credential function that performs no authentication.
#' This function returns the request object unchanged, allowing API calls
#' to be made without adding any authentication headers or tokens.
#'
#' @param req An [httr2::request()] object
#'
#' @return The same [httr2::request()] object, unmodified
#'
#' @export
default_non_auth <- function(req) {
  req
}


try_as_data_table <- function(x) {
  tryCatch(
    data.table::as.data.table(x),
    error = function(e) {
      cli::cli_warn(c(
        "Failed to convert data.frame to data.table: {e$message}",
        "i" = "Returning original data.frame"
      ))
      x
    }
  )
}

#' Default response handler
#'
#' Converts `data.frame` results in the parsed response to `data.table` objects
#' when the `data.table` package is available. Applied automatically by
#' [api_client] unless overridden via the `response_handler` argument.
#'
#' @param content Parsed response content from an API call.
#'
#' @return The processed content, with any `data.frame` objects converted to
#'   `data.table` if the `data.table` package is installed.
#'
#' @export
default_response_handler <- function(content) {
  if (!rlang::is_installed("data.table")) {
    return(content)
  }
  if (is.data.frame(content)) {
    return(try_as_data_table(content))
  }
  if (is.list(content)) {
    content <- lapply(content, function(x) {
      if (is.data.frame(x)) try_as_data_table(x) else x
    })
  }
}

drop_null <- function(x) Filter(Negate(is.null), x)

validate_tenant_id <- function(x) {
  if (!rlang::is_string(x)) {
    cli::cli_abort(
      "{.arg x} must be a single string, not {.obj_type_friendly {x}}"
    )
  }

  if (!grepl("^[A-Za-z0-9.-]+$", x)) {
    cli::cli_abort("Tenant ID {.val {x}} is not valid")
  }

  invisible(TRUE)
}


validate_scope <- function(x) {
  if (!rlang::is_character(x)) {
    cli::cli_abort(
      "{.arg x} must be a character vector, not {.obj_type_friendly {x}}"
    )
  }

  invalid <- !grepl("^[A-Za-z0-9_.:/-]+$", x)
  if (any(invalid)) {
    cli::cli_abort("Scope {.val {x[invalid]}} is not valid")
  }

  invisible(TRUE)
}


get_scope_resource <- function(scope) {
  x <- grep("^http", scope, value = TRUE, ignore.case = TRUE)

  if (length(x) != 1L) {
    return(NULL)
  }

  u <- httr2::url_parse(x)
  u$path <- NULL
  u$query <- NULL
  u$fragment <- NULL

  res <- httr2::url_build(u)
  sub("/$", "", res)
}


r6_get_initialize_arguments <- function(cls) {
  if (is.null(cls)) {
    return(NULL)
  }

  if (!R6::is.R6Class(cls)) {
    cli::cli_abort(
      "{.arg cls} must be an R6 class, not {.obj_type_friendly {cls}}"
    )
  }

  if (is.null(cls$public_methods$initialize)) {
    return(r6_get_initialize_arguments(cls$get_inherit()))
  }

  names(formals(cls$public_methods$initialize))
}


r6_get_public_fields <- function(cls) {
  if (!R6::is.R6Class(cls)) {
    cli::cli_abort(
      "{.arg cls} must be an R6 class, not {.obj_type_friendly {cls}}"
    )
  }

  res <- names(cls$public_fields)
  sup <- cls$get_inherit()

  if (!is.null(sup)) {
    return(c(res, r6_get_public_fields(sup)))
  }

  res
}


r6_get_class <- function(obj, envir = rlang::caller_env()) {
  if (!R6::is.R6(obj)) {
    cli::cli_abort(
      "{.arg obj} must be an R6 object, not {.obj_type_friendly {obj}}"
    )
  }
  get(class(obj)[[1]], envir = envir)
}


is_empty <- function(x) {
  is.null(x) ||
    (rlang::is_scalar_vector(x) &&
      (rlang::is_empty(x) || is.na(x) || !nzchar(x)))
}


is_empty_vec <- function(x) {
  vapply(x, is_empty, logical(1), USE.NAMES = FALSE)
}


get_env_config <- function() {
  tenant_id_env <- Sys.getenv("AZURE_TENANT_ID", unset = "")
  client_secret_env <- Sys.getenv("AZURE_CLIENT_SECRET", unset = "")
  authority_host_env <- Sys.getenv("AZURE_AUTHORITY_HOST", unset = "")
  config_dir_env <- Sys.getenv("AZURE_CONFIG_DIR", unset = "")
  federated_token_file_env <- Sys.getenv(
    "AZURE_FEDERATED_TOKEN_FILE",
    unset = ""
  )

  c(
    env_override_entry(
      "AZURE_TENANT_ID",
      env_val = tenant_id_env,
      default_val = azure_client$tenant_id
    ),
    "*" = if (nzchar(client_secret_env)) {
      paste0("AZURE_CLIENT_SECRET: ", cli::col_grey("<<REDACTED>>"))
    } else {
      paste0("AZURE_CLIENT_SECRET: ", cli::col_grey("(not set)"))
    },
    env_override_entry(
      "AZURE_AUTHORITY_HOST",
      env_val = authority_host_env,
      default_val = azure_authority_hosts$azure_public_cloud
    ),
    "*" = if (nzchar(config_dir_env)) {
      cli::format_inline("AZURE_CONFIG_DIR: {.val {config_dir_env}}")
    } else {
      cli::format_inline(
        "AZURE_CONFIG_DIR: {.val {default_azure_config_dir()}} (default)"
      )
    },
    "*" = if (nzchar(federated_token_file_env)) {
      cli::format_inline(
        "AZURE_FEDERATED_TOKEN_FILE: {.val {federated_token_file_env}}"
      )
    } else {
      paste0("AZURE_FEDERATED_TOKEN_FILE: ", cli::col_grey("(not set)"))
    }
  )
}


# Formats a bullet entry for a field that can be set via an env var and
# otherwise falls back to a built-in default.
env_override_entry <- function(var_name, env_val, default_val) {
  has_env <- nzchar(env_val)

  if (!has_env) {
    return(c(
      "*" = cli::format_inline(
        "{var_name}: {.val {default_val}} {cli::col_grey('(default)')}"
      )
    ))
  }

  check <- paste0(" ", cli::col_green("\u2713"))

  c(
    "*" = paste0(var_name, ":"),
    " " = paste0(cli::format_inline("  env: {.val {env_val}}"), check)
  )
}

#' Interactive credential base class
#'
#' @description
#' Base class for interactive authentication credentials. This class should not
#' be instantiated directly; use [DeviceCodeCredential] or [AuthCodeCredential]
#' instead.
#'
#' @keywords internal
InteractiveCredential <- R6::R6Class(
  classname = "InteractiveCredential",
  inherit = Credential,
  # public ----
  public = list(
    #' @field use_refresh_token Logical indicating whether to use the login flow (acquire
    #'  tokens via refresh token exchange).
    use_refresh_token = TRUE,
    #' @field interactive Logical indicating whether this credential requires
    #'  user interaction.
    interactive = TRUE,
    #' @description
    #' Shared initializer for interactive credentials
    #'
    #' @param scope A character string specifying the OAuth2 scope.
    #' @param tenant_id Azure AD tenant ID.
    #' @param client_id Application (client) ID. Defaults to the Azure CLI
    #'   public client ID.
    #' @param use_cache Cache type: `"disk"` or `"memory"`.
    #' @param offline Whether to request offline access (refresh tokens).
    #' @param interactive Whether this credential requires user interaction.
    #' @param use_refresh_token Whether to use the login flow (acquire tokens via refresh token
    #'   exchange). Set to `FALSE` to use the access token flow directly.
    #' @param flow_fun The httr2 OAuth flow function (e.g. [httr2::oauth_flow_device]).
    #' @param req_auth_fun The httr2 request auth function (e.g. [httr2::req_oauth_device]).
    #' @param oauth_endpoint The OAuth endpoint name passed to the parent credential.
    #' @param name The credential name passed to the parent credential.
    #' @param extra_flow_params A named list of additional parameters merged into
    #'   `private$.flow_params` after `scope` and `auth_url`.
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = default_azure_cli_client_id(),
      use_cache = "disk",
      offline = TRUE,
      interactive = TRUE,
      use_refresh_token = TRUE,
      flow_fun,
      req_auth_fun,
      oauth_endpoint,
      name,
      extra_flow_params = list()
    ) {
      self$interactive <- interactive
      private$.flow <- if (self$interactive) {
        flow_fun
      } else {
        \(...) cli::cli_abort("non-interactive session")
      }
      private$.req_auth_fun <- req_auth_fun
      self$use_refresh_token <- use_refresh_token

      private$.login_scope <- c(default_azure_scope(), "offline_access")

      super$initialize(
        scope = scope,
        tenant_id = tenant_id,
        client_id = client_id,
        use_cache = use_cache,
        offline = offline,
        oauth_endpoint = oauth_endpoint,
        name = name
      )
      private$.flow_params <- c(
        list(scope = self$.scope_str, auth_url = self$.oauth_url),
        extra_flow_params
      )
    },
    #' @description
    #' Check if the credential requires user interaction
    #'
    #' @return Logical indicating whether this credential is interactive
    is_interactive = function() {
      self$interactive
    },
    #' @description
    #' Get an access token using the flow configured by the subclass.
    #' Returns a valid in-object cached token immediately if one exists for the
    #' requested scope. Otherwise attempts token acquisition in three steps:
    #' (1) return a valid httr2-cached token without any interaction; (2)
    #' silently refresh using an existing refresh token; (3) fall back to the
    #' configured interactive flow. When `reauth = TRUE` all caches are bypassed
    #' and the interactive flow is used directly.
    #'
    #' @param scope A character string specifying the OAuth2 scope. Defaults to
    #'   `NULL`, which uses the scope configured on the credential.
    #' @param reauth A logical value indicating whether to force
    #'   reauthentication, bypassing the cache and silent refresh. Defaults to
    #'   `FALSE`.
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function(scope = NULL, reauth = FALSE) {
      cache_key <- rlang::hash(scope %||% self$.scope)

      if (!isTRUE(reauth)) {
        cached <- private$.token_cache[[cache_key]]
        if (private$token_is_valid(cached)) {
          return(cached)
        }
      }

      token <- private$fetch_token(scope = scope, reauth = reauth)
      private$.token_cache[[cache_key]] <- token
      token
    },
    #' @description
    #' Add OAuth authentication to an httr2 request using the flow configured
    #' by the subclass
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with OAuth authentication configured
    req_auth = function(req) {
      token <- self$get_token()
      httr2::req_auth_bearer_token(req, token$access_token)
    }
  ),
  # private ----
  private = list(
    .flow = NULL,
    .flow_params = NULL,
    .req_auth_fun = NULL,
    .login_scope = NULL,
    .token_cache = list(),
    token_is_valid = function(token) {
      if (is.null(token) || !inherits(token, "httr2_token")) {
        return(FALSE)
      }
      !token_has_expired(token)
    },
    fetch_token = function(scope = NULL, reauth = FALSE) {
      if (isTRUE(self$use_refresh_token) || !is.null(scope)) {
        token <- private$do_flow_refresh_token(scope = scope)
        if (inherits(token, "httr2_token")) {
          return(token)
        } else {
          cli::cli_abort("Failed to acquire token via login refresh flow.")
        }
      }
      private$do_flow_access_token(
        scope = NULL,
        reauth = reauth,
        interactive = self$interactive
      )
    },
    # runs access token flow
    do_flow_access_token = function(
      scope = NULL,
      reauth = FALSE,
      interactive = TRUE
    ) {
      flow_params <- private$.flow_params
      cache_key <- self$.cache_key

      if (!is.null(scope)) {
        flow_params$scope <- collapse_scope(scope)
        cache_key$scope <- collapse_scope(scope)
      }
      flow <- if (interactive) {
        private$.flow
      } else {
        \(...) cli::cli_abort("non-interactive session")
      }
      httr2::oauth_token_cached(
        client = self$.oauth_client,
        flow = flow,
        cache_disk = self$.use_cache == "disk",
        cache_key = cache_key,
        flow_params = flow_params,
        reauth = reauth
      )
    },
    # Get an access token for `scope` using the refresh token from this
    # credential's own cached token. Returns an httr2_token or NULL if not
    # possible.
    do_flow_refresh_token = function(scope = NULL) {
      scope <- collapse_scope(scope %||% self$.scope)
      login_token <- tryCatch(
        private$do_flow_access_token(scope = private$.login_scope),
        error = function(e) NULL
      )

      if (is.null(login_token) || is.null(login_token$refresh_token)) {
        return(NULL)
      }

      withCallingHandlers(
        tryCatch(
          httr2::oauth_flow_refresh(
            client = self$.oauth_client,
            refresh_token = login_token$refresh_token,
            scope = scope
          ),
          error = function(e) NULL
        ),
        warning = function(w) {
          if (
            grepl(
              "Refresh token has changed",
              conditionMessage(w),
              fixed = TRUE
            )
          ) {
            invokeRestart("muffleWarning")
          }
        }
      )
    }
  )
)


#' Device code credential authentication
#'
#' @description
#' Authenticates a user through the device code flow. This flow is designed for
#' devices that don't have a web browser or have input constraints.
#'
#' @details
#' The device code flow displays a code that the user must enter on another
#' device with a web browser to complete authentication. This is ideal for
#' CLI applications, headless servers, or devices without a browser.
#'
#' The credential supports token caching to avoid repeated authentication.
#' Tokens can be cached to disk or in memory.
#'
#' @export
#' @examples
#' # DeviceCodeCredential requires an interactive session
#' \dontrun{
#' # Create credential with default settings
#' cred <- DeviceCodeCredential$new()
#'
#' # Get an access token (will prompt for 'device code' flow)
#' token <- cred$get_token()
#'
#' # Force re-authentication
#' token <- cred$get_token(reauth = TRUE)
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' req <- cred$req_auth(req)
#' }
DeviceCodeCredential <- R6::R6Class(
  classname = "DeviceCodeCredential",
  inherit = InteractiveCredential,
  public = list(
    #' @description
    #' Create a new device code credential
    #'
    #' @param scope A character string specifying the OAuth2 scope. Defaults to `NULL`.
    #' @param tenant_id A character string specifying the Azure Active Directory
    #'   tenant ID. Defaults to `NULL`.
    #' @param client_id A character string specifying the application (client) ID.
    #'   Defaults to the Azure CLI public client ID.
    #' @param use_cache A character string specifying the cache type. Use `"disk"`
    #'   for disk-based caching or `"memory"` for in-memory caching. Defaults to `"disk"`.
    #' @param offline A logical value indicating whether to request offline access
    #'   (refresh tokens). Defaults to `TRUE`.
    #' @param interactive A logical value indicating whether this credential
    #'   requires user interaction. Defaults to `TRUE`.
    #' @param use_refresh_token A logical value indicating whether to use the login flow
    #'   (acquire tokens via refresh token exchange). Defaults to `TRUE`.
    #'
    #' @return A new `DeviceCodeCredential` object
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = default_azure_cli_client_id(),
      use_cache = "disk",
      offline = TRUE,
      interactive = TRUE,
      use_refresh_token = TRUE
    ) {
      super$initialize(
        scope = scope,
        tenant_id = tenant_id,
        client_id = client_id,
        use_cache = use_cache,
        offline = offline,
        interactive = interactive,
        use_refresh_token = use_refresh_token,
        flow_fun = httr2::oauth_flow_device,
        req_auth_fun = httr2::req_oauth_device,
        oauth_endpoint = "devicecode",
        name = "azr-device-code"
      )
    }
  )
)


#' Authorization code credential authentication
#'
#' @description
#' Authenticates a user through the OAuth 2.0 authorization code flow. This
#' flow opens a web browser for the user to sign in.
#'
#' @details
#' The authorization code flow is the standard OAuth 2.0 flow for interactive
#' authentication. It requires a web browser and is suitable for applications
#' where the user can interact with a browser window.
#'
#' The credential supports token caching to avoid repeated authentication.
#' Tokens can be cached to disk or in memory. A redirect URI is required for
#' the OAuth flow to complete.
#'
#' @export
#' @examples
#' # AuthCodeCredential requires an interactive session
#' \dontrun{
#' # Create credential with default settings
#' cred <- AuthCodeCredential$new(
#'   tenant_id = "your-tenant-id",
#'   client_id = "your-client-id",
#'   scope = "https://management.azure.com/.default"
#' )
#'
#' # Get an access token (will open browser for authentication)
#' token <- cred$get_token()
#'
#' # Force reauthentication
#' token <- cred$get_token(reauth = TRUE)
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' req <- cred$req_auth(req)
#' }
AuthCodeCredential <- R6::R6Class(
  classname = "AuthCodeCredential",
  inherit = InteractiveCredential,
  public = list(
    #' @description
    #' Create a new authorization code credential
    #'
    #' @param scope A character string specifying the OAuth2 scope. Defaults to `NULL`.
    #' @param tenant_id A character string specifying the Azure Active Directory
    #'   tenant ID. Defaults to `NULL`.
    #' @param client_id A character string specifying the application (client) ID.
    #'   Defaults to the Azure CLI public client ID.
    #' @param use_cache A character string specifying the cache type. Use `"disk"`
    #'   for disk-based caching or `"memory"` for in-memory caching. Defaults to `"disk"`.
    #' @param offline A logical value indicating whether to request offline access
    #'   (refresh tokens). Defaults to `TRUE`.
    #' @param redirect_uri A character string specifying the redirect URI registered
    #'   with the application. Defaults to [default_redirect_uri()].
    #' @param interactive A logical value indicating whether this credential
    #'   requires user interaction. Defaults to `TRUE`.
    #' @param use_refresh_token A logical value indicating whether to use the login flow
    #'   (acquire tokens via refresh token exchange). Defaults to `TRUE`.
    #'
    #' @return A new `AuthCodeCredential` object
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = default_azure_cli_client_id(),
      use_cache = "disk",
      offline = TRUE,
      redirect_uri = default_redirect_uri(),
      interactive = TRUE,
      use_refresh_token = TRUE
    ) {
      super$initialize(
        scope = scope,
        tenant_id = tenant_id,
        client_id = client_id,
        use_cache = use_cache,
        offline = offline,
        interactive = interactive,
        use_refresh_token = use_refresh_token,
        flow_fun = httr2::oauth_flow_auth_code,
        req_auth_fun = httr2::req_oauth_auth_code,
        oauth_endpoint = "authorize",
        name = "azr-auth-code",
        extra_flow_params = list(redirect_uri = redirect_uri)
      )
      self$.redirect_uri <- redirect_uri
      lockBinding(".redirect_uri", self)
    }
  )
)

#' Azure Default Client Configuration
#'
#' @description
#' Default client ID and tenant ID used for Azure authentication when not
#' explicitly provided. The client ID is Microsoft's public Azure CLI client ID.
#'
#' @keywords internal
azure_client <- list(
  tenant_id = "common",
  client_id = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"
)

#' Azure Authority Host URLs
#'
#' @description
#' Login endpoint URLs for different Azure cloud environments.
#'
#' @keywords internal
azure_authority_hosts <- list(
  azure_china = "login.chinacloudapi.cn",
  azure_government = "login.microsoftonline.us",
  azure_public_cloud = "login.microsoftonline.com"
)

#' Azure Storage Endpoint Defaults
#'
#' @description
#' Default endpoint suffixes for Azure Storage services.
#'
#' @keywords internal
azure_storage_endpoints <- list(
  dfs = "dfs.core.windows.net"
)

#' Common Azure OAuth Scopes
#'
#' @description
#' Predefined OAuth scopes for common Azure services.
#'
#' @keywords internal
azure_scopes <- list(
  azure_arm = "https://management.azure.com/.default",
  azure_graph = "https://graph.microsoft.com/.default",
  azure_storage = "https://storage.azure.com/.default",
  azure_key_vault = "https://vault.azure.net/.default",
  azure_openai = "https://cognitiveservices.azure.com/.default"
)

#' Azure Environment Variable Names
#'
#' @description
#' Standard environment variable names used for Azure credential discovery.
#'
#' @keywords internal
environment_variables <- list(
  azure_client_id = "AZURE_CLIENT_ID",
  azure_client_secret = "AZURE_CLIENT_SECRET",
  azure_tenant_id = "AZURE_TENANT_ID",
  azure_authority_host = "AZURE_AUTHORITY_HOST",
  azure_refresh_token = "AZURE_REFRESH_TOKEN",
  client_secret_vars = c(
    "AZURE_CLIENT_ID",
    "AZURE_CLIENT_SECRET",
    "AZURE_TENANT_ID"
  ),
  cert_vars = c(
    "AZURE_CLIENT_ID",
    "AZURE_CLIENT_CERTIFICATE_PATH",
    "AZURE_TENANT_ID"
  ),
  azure_username = "AZURE_USERNAME",
  azure_password = "AZURE_PASSWORD",
  username_password_vars = c(
    "AZURE_CLIENT_ID",
    "AZURE_USERNAME",
    "AZURE_PASSWORD"
  ),
  azure_federated_token_file = "AZURE_FEDERATED_TOKEN_FILE"
)

# RefreshTokenCredential ----
#' Refresh token credential authentication
#'
#' @description
#' Authenticates using an existing refresh token. This credential is useful when
#' you have obtained a refresh token through another authentication flow and want
#' to use it to get new access tokens without interactive authentication.
#'
#' @details
#' The refresh token credential uses the OAuth 2.0 refresh token flow to obtain
#' new access tokens. It requires a valid refresh token that was previously
#' obtained through an interactive flow (e.g., authorization code or device code).
#'
#' This is particularly useful for:
#' \itemize{
#'   \item Non-interactive sessions where you have a pre-obtained refresh token
#'   \item Long-running applications that need to refresh tokens automatically
#'   \item Scenarios where you want to avoid repeated interactive authentication
#' }
#'
#' @field .refresh_token Character string containing the refresh token.
#'
#' @export
#' @examples
#' \dontrun{
#' # Create credential with a refresh token
#' cred <- RefreshTokenCredential$new(
#'   refresh_token = "your-refresh-token",
#'   scope = "https://management.azure.com/.default",
#'   tenant_id = "your-tenant-id",
#'   client_id = "your-client-id"
#' )
#'
#' # Get an access token
#' token <- cred$get_token()
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' resp <- httr2::req_perform(cred$req_auth(req))
#' }
RefreshTokenCredential <- R6::R6Class(
  classname = "RefreshTokenCredential",
  inherit = Credential,
  public = list(
    .refresh_token = NULL,

    #' @description
    #' Create a new refresh token credential
    #'
    #' @param refresh_token A character string containing the refresh token.
    #'   Defaults to [default_refresh_token()] which reads from the
    #'   `AZURE_REFRESH_TOKEN` environment variable.
    #' @param scope A character string specifying the OAuth2 scope. Defaults to `NULL`.
    #' @param tenant_id A character string specifying the Azure Active Directory
    #'   tenant ID. Defaults to `NULL`.
    #' @param client_id A character string specifying the application (client) ID.
    #'   Defaults to `NULL`.
    #'
    #' @return A new `RefreshTokenCredential` object
    initialize = function(
      refresh_token = default_refresh_token(),
      scope = NULL,
      tenant_id = NULL,
      client_id = NULL
    ) {
      self$.refresh_token <- refresh_token
      super$initialize(
        scope = scope,
        tenant_id = tenant_id,
        client_id = client_id
      )
      lockBinding(".refresh_token", self)
    },

    #' @description
    #' Validate the credential configuration
    #'
    #' @details
    #' Checks that the refresh token is provided and not NA or NULL. Calls the
    #' parent class validation method.
    validate = function() {
      super$validate()

      if (is.null(self$.refresh_token) || rlang::is_na(self$.refresh_token)) {
        cli::cli_abort("Argument {.arg refresh_token} cannot be NULL or NA.")
      }
    },

    #' @description
    #' Get an access token using the refresh token flow
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function() {
      httr2::oauth_flow_refresh(
        client = self$.oauth_client,
        refresh_token = self$.refresh_token,
        scope = self$.scope_str
      )
    },

    #' @description
    #' Add OAuth refresh token authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with OAuth refresh token authentication configured
    req_auth = function(req) {
      httr2::req_oauth_refresh(
        req = req,
        client = self$.oauth_client,
        refresh_token = self$.refresh_token,
        scope = self$.scope_str
      )
    }
  )
)

#' Default credential authentication
#'
#' @description
#' An R6 class that provides lazy initialization of credential providers.
#' The credential provider is created on first access using the default
#' credential chain.
#'
#' @details
#' This class wraps the credential discovery process in an R6 object with
#' a lazily evaluated `provider` field. The provider is only created when
#' first accessed, using the same logic as [get_token_provider()].
#'
#' @export
#' @examples
#' # Create a DefaultCredential object
#' cred <- DefaultCredential$new(
#'   scope = "https://graph.microsoft.com/.default",
#'   tenant_id = "my-tenant-id"
#' )
#'
#' \dontrun{
#' # Get a token (triggers lazy initialization)
#' token <- cred$get_token()
#'
#' # Authenticate a request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' resp <- httr2::req_perform(cred$req_auth(req))
#'
#' # Or access the provider directly
#' provider <- cred$provider
#' }
#'
#' @field .scope Character string specifying the authentication scope.
#' @field .tenant_id Character string specifying the tenant ID.
#' @field .client_id Character string specifying the client ID.
#' @field .client_secret Character string specifying the client secret.
#' @field .use_cache Character string indicating the caching strategy.
#' @field .offline Logical indicating whether to request offline access.
#' @field .chain A credential chain object for authentication.
DefaultCredential <- R6::R6Class(
  classname = "DefaultCredential",
  public = list(
    .scope = NULL,
    .tenant_id = NULL,
    .client_id = NULL,
    .client_secret = NULL,
    .use_cache = NULL,
    .offline = NULL,
    .chain = NULL,

    #' @description
    #' Create a new DefaultCredential object
    #'
    #' @param scope Optional character string specifying the authentication scope.
    #' @param tenant_id Optional character string specifying the tenant ID for
    #'   authentication.
    #' @param client_id Optional character string specifying the client ID for
    #'   authentication.
    #' @param client_secret Optional character string specifying the client secret
    #'   for authentication.
    #' @param use_cache Character string indicating the caching strategy. Defaults
    #'   to `"disk"`. Options include `"disk"` for disk-based caching or `"memory"`
    #'   for in-memory caching.
    #' @param offline Logical. If `TRUE`, adds 'offline_access' to the scope to request a 'refresh_token'.
    #'   Defaults to `TRUE`.
    #' @param chain A list of credential objects, where each element must inherit
    #'   from the `Credential` base class. Credentials are attempted in the order
    #'   provided until `get_token` succeeds.
    #'
    #' @return A new `DefaultCredential` object
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = NULL,
      client_secret = NULL,
      use_cache = "disk",
      offline = TRUE,
      chain = default_credential_chain()
    ) {
      self$.scope <- scope
      self$.tenant_id <- tenant_id
      self$.client_id <- client_id
      self$.client_secret <- client_secret
      self$.use_cache <- use_cache
      self$.offline <- offline
      self$.chain <- chain
    },

    #' @description
    #' Get an access token using the credential chain
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function() {
      self$provider$get_token()
    },

    #' @description
    #' Add authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with authentication configured
    req_auth = function(req) {
      self$provider$req_auth(req)
    }
  ),
  active = list(
    #' @field provider Lazily initialized credential provider
    provider = function() {
      if (is.null(private$.provider_cache)) {
        private$.provider_cache <- get_credential_provider(
          scope = self$.scope,
          tenant_id = self$.tenant_id,
          client_id = self$.client_id,
          client_secret = self$.client_secret,
          use_cache = self$.use_cache,
          offline = self$.offline,
          chain = self$.chain
        )
      }
      private$.provider_cache
    }
  ),
  private = list(
    .provider_cache = NULL
  )
)


#' Get Default Token Provider Function
#'
#' Creates a token provider function that retrieves authentication credentials
#' and returns a callable token getter. This function handles the credential
#' discovery process and returns the token acquisition method from the
#' discovered credential object.
#'
#' @param scope Optional character string specifying the authentication scope.
#' @param tenant_id Optional character string specifying the tenant ID for
#'   authentication.
#' @param client_id Optional character string specifying the client ID for
#'   authentication.
#' @param client_secret Optional character string specifying the client secret
#'   for authentication.
#' @param use_cache Character string indicating the caching strategy. Defaults
#'   to `"disk"`. Options include `"disk"` for disk-based caching or `"memory"`
#'   for in-memory caching.
#' @param offline Logical. If `TRUE`, adds 'offline_access' to the scope to request a 'refresh_token'.
#'   Defaults to `TRUE`.
#' @param chain A list of credential objects, where each element must inherit
#'   from the `Credential` base class. Credentials are attempted in the order
#'   provided until `get_token` succeeds.
#'
#' @return A function that retrieves and returns an authentication token when
#'   called.
#'
#' @seealso [get_request_authorizer()], [get_token()]
#'
#' @examples
#' # In non-interactive sessions, this function will return an error if the
#' # environment is not set up with valid credentials. In an interactive session
#' # the user will be prompted to attempt one of the interactive authentication flows.
#' \dontrun{
#' token_provider <- get_token_provider(
#'   scope = "https://graph.microsoft.com/.default",
#'   tenant_id = "my-tenant-id",
#'   client_id = "my-client-id",
#'   client_secret = "my-secret"
#' )
#' token <- token_provider()
#' }
#'
#' @export
get_token_provider <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
) {
  provider <- get_credential_provider(
    scope = scope,
    tenant_id = tenant_id,
    client_id = client_id,
    client_secret = client_secret,
    use_cache = use_cache,
    offline = offline,
    chain = chain
  )

  function() {
    provider$get_token()
  }
}

#' Get Default Request Authorizer Function
#'
#' Creates a request authorizer function that retrieves authentication credentials
#' and returns a callable request authorization method. This function handles the
#' credential discovery process and returns the request authentication method
#' from the discovered credential object.
#'
#' @param scope Optional character string specifying the authentication scope.
#' @param tenant_id Optional character string specifying the tenant ID for
#'   authentication.
#' @param client_id Optional character string specifying the client ID for
#'   authentication.
#' @param client_secret Optional character string specifying the client secret
#'   for authentication.
#' @param use_cache Character string indicating the caching strategy. Defaults
#'   to `"disk"`. Options include `"disk"` for disk-based caching or `"memory"`
#'   for in-memory caching.
#' @param offline Logical. If `TRUE`, adds 'offline_access' to the scope to request a 'refresh_token'.
#'   Defaults to `TRUE`.
#' @param chain A list of credential objects, where each element must inherit
#'   from the `Credential` base class. Credentials are attempted in the order
#'   provided until `get_token` succeeds.
#'
#' @return A function that authorizes HTTP requests with appropriate credentials
#'   when called.
#'
#'
#' @seealso [get_token_provider()], [get_token()]
#'
#' @examples
#' # In non-interactive sessions, this function will return an error if the
#' # environment is not setup with valid credentials. And in an interactive session
#' # the user will be prompted to attempt one of the interactive authentication flows.
#' \dontrun{
#' req_auth <- get_request_authorizer(
#'   scope = "https://graph.microsoft.com/.default"
#' )
#' req <- req_auth(httr2::request("https://graph.microsoft.com/v1.0/me"))
#' }
#'
#' @export
get_request_authorizer <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
) {
  provider <- get_credential_provider(
    scope = scope,
    tenant_id = tenant_id,
    client_id = client_id,
    client_secret = client_secret,
    use_cache = use_cache,
    offline = offline,
    chain = chain
  )

  function(req) {
    provider$req_auth(req)
  }
}

#' Get Authentication Token
#'
#' Retrieves an authentication token using the default token provider. This is
#' a convenience function that combines credential discovery and token
#' acquisition in a single step.
#'
#' @param scope Optional character string specifying the authentication scope.
#' @param tenant_id Optional character string specifying the tenant ID for
#'   authentication.
#' @param client_id Optional character string specifying the client ID for
#'   authentication.
#' @param client_secret Optional character string specifying the client secret
#'   for authentication.
#' @param use_cache Character string indicating the caching strategy. Defaults
#'   to `"disk"`. Options include `"disk"` for disk-based caching or `"memory"`
#'   for in-memory caching.
#' @param offline Logical. If `TRUE`, adds 'offline_access' to the scope to request a 'refresh_token'.
#'   Defaults to `TRUE`.
#' @param chain A list of credential objects, where each element must inherit
#'   from the `Credential` base class. Credentials are attempted in the order
#'   provided until `get_token` succeeds.
#'
#' @return An [httr2::oauth_token()] object.
#'
#' @seealso [get_token_provider()], [get_request_authorizer()]
#'
#' @examples
#' # In non-interactive sessions, this function will return an error if the
#' # environment is not setup with valid credentials. And in an interactive session
#' # the user will be prompted to attempt one of the interactive authentication flows.
#' \dontrun{
#' token <- get_token(
#'   scope = "https://graph.microsoft.com/.default",
#'   tenant_id = "my-tenant-id",
#'   client_id = "my-client-id",
#'   client_secret = "my-secret"
#' )
#' }
#'
#' @export
get_token <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
) {
  provider <- get_credential_provider(
    scope = scope,
    tenant_id = tenant_id,
    client_id = client_id,
    client_secret = client_secret,
    use_cache = use_cache,
    offline = offline,
    chain = chain
  )

  provider$get_token()
}


#' Get Credential Authentication Function
#'
#' @description
#' Creates a function that retrieves authentication tokens and formats them as
#' HTTP Authorization headers. This function handles credential discovery and
#' returns a callable method that generates Bearer token headers when invoked.
#'
#' @inheritParams get_token
#'
#' @return A function that, when called, returns a named list with an
#'   `Authorization` element containing the Bearer token, suitable for use
#'   with [httr2::req_headers()].
#'
#' @seealso [get_token()], [get_request_authorizer()], [get_token_provider()]
#'
#' @examples
#' \dontrun{
#' # Create an authentication function
#' auth_fn <- get_credential_auth(
#'   scope = "https://graph.microsoft.com/.default"
#' )
#'
#' # Call it to get headers
#' auth_headers <- auth_fn()
#'
#' # Use with httr2
#' req <- httr2::request("https://graph.microsoft.com/v1.0/me") |>
#'   httr2::req_headers(!!!auth_headers)
#' }
#'
#' @export
get_credential_auth <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
) {
  get_token <- get_token_provider(
    scope = scope,
    tenant_id = tenant_id,
    client_id = client_id,
    client_secret = client_secret,
    use_cache = use_cache,
    offline = offline,
    chain = chain
  )

  function() {
    token <- get_token()
    list(Authorization = paste0("Bearer ", token$access_token))
  }
}


#' Get Credential Provider
#'
#' @description
#' Discovers and returns an authenticated credential object from a chain of
#' credential providers. This function attempts each credential in the chain
#' until one successfully authenticates, returning the first successful
#' credential object.
#'
#' @param scope Optional character string specifying the authentication scope.
#' @param tenant_id Optional character string specifying the tenant ID for
#'   authentication.
#' @param client_id Optional character string specifying the client ID for
#'   authentication.
#' @param client_secret Optional character string specifying the client secret
#'   for authentication.
#' @param use_cache Character string indicating the caching strategy. Defaults
#'   to `"disk"`. Options include `"disk"` for disk-based caching or `"memory"`
#'   for in-memory caching.
#' @param offline Logical. If `TRUE`, adds 'offline_access' to the scope to
#'   request a 'refresh_token'. Defaults to `TRUE`.
#' @param oauth_host Optional character string specifying the OAuth host URL.
#' @param oauth_endpoint Optional character string specifying the OAuth endpoint.
#' @param chain A list of credential objects, where each element must inherit
#'   from the `Credential` base class. Credentials are attempted in the order
#'   provided until `get_token` succeeds. If `NULL`, uses
#'   [default_credential_chain()].
#' @param interactive A logical value indicating whether interactive credentials
#'   are allowed. Defaults to `TRUE`.
#' @param verbose A logical value indicating whether to print verbose messages
#'   during credential discovery. Defaults to `getOption("azr.verbose", FALSE)`.
#'
#' @return A credential object that inherits from the `Credential` class and
#'   has successfully authenticated.
#'
#' @seealso [get_token_provider()], [get_request_authorizer()],
#'   [default_credential_chain()]
#'
#' @examples
#' \dontrun{
#' # Get a credential provider with default settings
#' cred <- get_credential_provider(
#'   scope = "https://graph.microsoft.com/.default",
#'   tenant_id = "my-tenant-id"
#' )
#'
#' # Use the credential to get a token
#' token <- cred$get_token()
#' }
#'
#' @export
get_credential_provider <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  oauth_host = NULL,
  oauth_endpoint = NULL,
  chain = NULL,
  interactive = TRUE,
  verbose = getOption("azr.verbose", FALSE)
) {
  if (is.null(chain) || length(chain) == 0L) {
    chain <- default_credential_chain()
  }

  if (!inherits(chain, "credential_chain")) {
    cli::cli_abort(
      "Argument {.arg chain} must be of class {.cls credential_chain}."
    )
  }

  errors <- list()
  attempt <- try_build_credential(chain, envir = rlang::current_env())
  errors <- attempt$errors

  for (crd_name in names(attempt$credentials)) {
    obj <- attempt$credentials[[crd_name]]

    if (verbose) {
      cli::cli_inform(c(
        " " = "Attempting to get token from {.strong {crd_name}}..."
      ))
    }

    token <- tryCatch(
      obj$get_token(),
      error = function(e) {
        errors[[crd_name]] <<- conditionMessage(e)
        NULL
      },
      interrupt = function(e) {
        errors[[crd_name]] <<- "Authentication interrupted by user"
        NULL
      }
    )

    if (inherits(token, "httr2_token")) {
      if (verbose) {
        cli::cli_inform(c(
          "v" = "Successfully authenticated with {.strong {crd_name}}."
        ))
      }
      return(obj)
    }

    if (verbose) {
      cli::cli_inform(c(
        "x" = "{.strong {crd_name}} failed: {errors[[crd_name]]}"
      ))
    }
  }

  # All credentials failed, report all errors
  error_msgs <- c(
    "All authentication methods in the chain failed!"
  )

  for (cred_name in names(errors)) {
    error_msgs <- c(
      error_msgs,
      "i" = paste0("{.strong ", cred_name, "}:"),
      "x" = errors[[cred_name]]
    )
  }

  cli::cli_abort(error_msgs, class = "azr_credential_chain_failed")
}

try_build_credential <- function(chain, envir) {
  verbose <- rlang::env_get(envir, "verbose", default = FALSE)
  credentials <- list()
  errors <- list()

  for (i in seq_along(chain)) {
    crd_expr <- chain[[i]]
    crd_name <- names(chain)[i]
    if (is.null(crd_name) || !nzchar(crd_name)) {
      crd_name <- paste0("credential_", i)
    }

    if (verbose) {
      cli::cli_inform(c(
        "i" = "Trying credential {.strong {crd_name}} ({i}/{length(chain)})..."
      ))
    }

    crd <- try(
      {
        mask <- rlang::new_data_mask(
          rlang::env_clone(
            asNamespace("azr"),
            parent = rlang::quo_get_env(crd_expr)
          )
        )
        rlang::eval_tidy(crd_expr, data = mask)
      },
      silent = TRUE
    )

    if (R6::is.R6Class(crd)) {
      if (verbose) {
        cli::cli_inform(c(
          " " = "Instantiating R6 class {.cls {crd$classname}}."
        ))
      }

      obj <- try(
        new_instance(crd, env = envir),
        silent = TRUE
      )

      if (inherits(obj, "try-error")) {
        errors[[crd_name]] <- conditionMessage(attr(obj, "condition"))
        if (verbose) {
          cli::cli_inform(c(
            "x" = "Failed to instantiate {.strong {crd_name}}: {errors[[crd_name]]}"
          ))
        }
        next
      }

      if (!inherits(obj, "Credential")) {
        errors[[crd_name]] <- "Object does not inherit from Credential class"
        if (verbose) {
          cli::cli_inform(c(
            "x" = "{.strong {crd_name}}: {errors[[crd_name]]}"
          ))
        }
        next
      }
    } else if (R6::is.R6(crd) && inherits(crd, "Credential")) {
      if (verbose) {
        cli::cli_inform(c(
          " " = "Using existing R6 instance of class {.cls {class(crd)[1]}}."
        ))
      }
      obj <- crd
    } else {
      errors[[crd_name]] <- "Invalid credential type"
      if (verbose) {
        cli::cli_inform(c(
          "x" = "{.strong {crd_name}}: {errors[[crd_name]]}"
        ))
      }
      next
    }

    if (obj$is_interactive() && !rlang::is_interactive()) {
      errors[[crd_name]] <- "Credential requires interactive session"
      if (verbose) {
        cli::cli_inform(c(
          "x" = "Skipping {.strong {crd_name}}: {errors[[crd_name]]}"
        ))
      }
      next
    }

    credentials[[crd_name]] <- obj
  }

  list(
    credentials = credentials,
    errors = errors
  )
}


#' Create Default Credential Chain
#'
#' Creates the default chain of credentials to attempt during authentication.
#' The credentials are tried in order until one successfully authenticates.
#' The default chain includes:
#' \enumerate{
#'   \item Client Secret Credential - Uses client ID and secret
#'   \item Authorization Code Credential - Interactive browser-based authentication
#'   \item Device Code Credential - Interactive device code flow
#'   \item Azure CLI Credential - Uses credentials from Azure CLI
#' }
#'
#' @return A `credential_chain` object containing the default sequence of
#'   credential providers.
#'
#' @seealso [credential_chain()], [get_token_provider()]
#'
#' @export
default_credential_chain <- function() {
  credential_chain(
    client_secret = ClientSecretCredential,
    auth_code = AuthCodeCredential,
    device_code = DeviceCodeCredential,
    azure_cli = AzureCLICredential
  )
}


#' Create Custom Credential Chain
#'
#' Creates a custom chain of credential providers to attempt during
#' authentication. Credentials are tried in the order they are provided
#' until one successfully authenticates. This allows you to customize
#' the authentication flow beyond the default credential chain.
#'
#' @param ... Named credential objects or credential classes. Each element
#'   should be a credential class (e.g., `ClientSecretCredential`) or an
#'   instantiated credential object that inherits from the `Credential`
#'   base class. The names are used for identification purposes.
#'
#' @return A `credential_chain` object containing the specified sequence
#'   of credential providers.
#'
#' @seealso [default_credential_chain()], [get_token_provider()]
#'
#' @examples
#' # Create a custom chain with only non-interactive credentials
#' custom_chain <- credential_chain(
#'   client_secret = ClientSecretCredential,
#'   azure_cli = AzureCLICredential
#' )
#'
#' # Use the custom chain to get a token
#' \dontrun{
#' token <- get_token(
#'   scope = "https://graph.microsoft.com/.default",
#'   chain = custom_chain
#' )
#' }
#'
#' @export
credential_chain <- function(...) {
  res <- rlang::enquos(...)

  if (length(res) == 0L) {
    cli::cli_abort(
      c(
        "Credential chain cannot be empty.",
        "i" = "Provide at least one credential class or instance.",
        "i" = "Use {.fn default_credential_chain} for a pre-configured chain."
      )
    )
  }

  class(res) <- c("credential_chain", class(res))
  res
}

new_instance <- function(cls, env = rlang::caller_env()) {
  cls_args <- r6_get_initialize_arguments(cls)

  if (is.null(cls_args)) {
    return(cls$new())
  }

  cls_values <- rlang::env_get_list(nms = cls_args, default = NULL, env = env)
  cls_values <- Filter(Negate(is.null), cls_values)

  eval(rlang::call2(cls$new, !!!cls_values))
}

#' Get default Azure tenant ID
#'
#' @description
#' Retrieves the Azure tenant ID in priority order:
#' 1. `AZURE_TENANT_ID` environment variable
#' 2. Built-in fallback (`"common"`)
#'
#' @return A character string with the tenant ID
#'
#' @export
#' @examples
#' default_azure_tenant_id()
default_azure_tenant_id <- function() {
  Sys.getenv(
    environment_variables$azure_tenant_id,
    unset = azure_client$tenant_id
  )
}


#' Get default Azure client ID
#'
#' @description
#' Retrieves the Azure client ID in priority order:
#' 1. `AZURE_CLIENT_ID` environment variable
#' 2. Built-in fallback (Microsoft's public Azure CLI client ID)
#'
#' @return A character string with the client ID
#'
#' @export
#' @examples
#' default_azure_client_id()
default_azure_client_id <- function() {
  Sys.getenv(
    environment_variables$azure_client_id,
    unset = azure_client$client_id
  )
}


default_azure_cli_client_id <- function() {
  azure_client$client_id
}


#' Get default Azure client secret
#'
#' @description
#' Retrieves the Azure client secret from the `AZURE_CLIENT_SECRET` environment
#' variable, or returns `NA_character_` if not set.
#'
#' @return A character string with the client secret, or `NA_character_` if not set
#'
#' @export
#' @examples
#' default_azure_client_secret()
default_azure_client_secret <- function() {
  res <- Sys.getenv(
    environment_variables$azure_client_secret,
    unset = NA_character_
  )

  if (is.na(res)) {
    NULL
  } else {
    res
  }
}


#' Get default Azure refresh token
#'
#' @description
#' Retrieves the Azure refresh token from the `AZURE_REFRESH_TOKEN` environment
#' variable, or returns `NULL` if not set.
#'
#' @return A character string with the refresh token, or `NULL` if not set
#'
#' @export
#' @examples
#' default_refresh_token()
default_refresh_token <- function() {
  res <- Sys.getenv(
    environment_variables$azure_refresh_token,
    unset = NA_character_
  )

  if (is.na(res)) {
    NULL
  } else {
    res
  }
}


#' Get default Azure OAuth scope
#'
#' @description
#' Returns the default OAuth scope for a specified Azure resource.
#'
#' @param resource A character string specifying the Azure resource. Must be one of:
#'   `"azure_arm"` (Azure Resource Manager), `"azure_graph"` (Microsoft Graph),
#'   `"azure_storage"` (Azure Storage), or `"azure_key_vault"` (Azure Key Vault).
#'   Defaults to `"azure_arm"`.
#'
#' @return A character string with the OAuth scope URL
#'
#' @export
#' @examples
#' default_azure_scope()
#' default_azure_scope("azure_graph")
default_azure_scope <- function(resource = "azure_arm") {
  resource <- rlang::arg_match(resource, values = names(azure_scopes))
  azure_scopes[[resource]]
}


#' Create default Azure OAuth client
#'
#' @description
#' Creates an [httr2::oauth_client()] configured for Azure authentication.
#'
#' @param client_id A character string specifying the client ID. Defaults to
#'   [default_azure_client_id()].
#' @param client_secret A character string specifying the client secret. Defaults
#'   to `NULL`.
#' @param name A character string specifying the client name. Defaults to `NULL`.
#'
#' @return An [httr2::oauth_client()] object
#'
#' @export
#' @examples
#' client <- default_azure_oauth_client()
#' client <- default_azure_oauth_client(
#'   client_id = "my-client-id",
#'   client_secret = "my-secret"
#' )
default_azure_oauth_client <- function(
  client_id = default_azure_client_id(),
  client_secret = NULL,
  name = NULL
) {
  httr2::oauth_client(
    name = name,
    id = client_id,
    token_url = default_azure_url("token"),
    secret = client_secret,
    auth = "body"
  )
}


#' Get default Azure OAuth URLs
#'
#' @description
#' Constructs Azure OAuth 2.0 endpoint URLs for a given tenant and authority host.
#'
#' @param endpoint A character string specifying which endpoint URL to return.
#'   Must be one of: `"authorize"`, `"token"`, or `"devicecode"`. If `NULL`
#'   (default), returns a list of all endpoint URLs.
#' @param oauth_host A character string specifying the Azure authority host.
#'   Defaults to [default_azure_host()].
#' @param tenant_id A character string specifying the tenant ID. Defaults to
#'   [default_azure_tenant_id()].
#'
#' @return If `endpoint` is specified, returns a character string with the URL.
#'   If `endpoint` is `NULL`, returns a named list of all endpoint URLs.
#'
#' @export
#' @examples
#' # Get all URLs
#' default_azure_url()
#'
#' # Get specific endpoint
#' default_azure_url("token")
#'
#' # Custom tenant
#' default_azure_url("authorize", tenant_id = "my-tenant-id")
default_azure_url <- function(
  endpoint = NULL,
  oauth_host = default_azure_host(),
  tenant_id = default_azure_tenant_id()
) {
  validate_tenant_id(tenant_id)

  if (!rlang::is_string(oauth_host) || !nzchar(oauth_host)) {
    cli::cli_abort(
      "{.arg oauth_host} must be a non-empty string, not {.obj_type_friendly {oauth_host}}"
    )
  }

  oauth_host <- sub("/+$", "", oauth_host)
  oauth_host <- sub("^https?://", "", oauth_host)
  oauth_base <- rlang::englue("https://{oauth_host}/{tenant_id}/oauth2/v2.0")

  urls <- c(
    authorize = paste0(oauth_base, "/authorize"),
    token = paste0(oauth_base, "/token"),
    devicecode = paste0(oauth_base, "/devicecode")
  )

  if (!is.null(endpoint)) {
    endpoint <- rlang::arg_match(endpoint, values = names(urls))
    return(urls[[endpoint]])
  }

  as.list(urls)
}


#' Get default Azure authority host
#'
#' @description
#' Retrieves the Azure authority host in priority order:
#' 1. `AZURE_AUTHORITY_HOST` environment variable
#' 2. Built-in fallback (`login.microsoftonline.com`)
#'
#' @return A character string with the authority host URL
#'
#' @export
#' @examples
#' default_azure_host()
default_azure_host <- function() {
  host <- Sys.getenv(
    environment_variables$azure_authority_host,
    unset = azure_authority_hosts$azure_public_cloud
  )

  if (!rlang::is_string(host) || !nzchar(host)) {
    cli::cli_abort(
      "{.arg host} must be a non-empty string, not {.obj_type_friendly {host}}"
    )
  }

  host <- sub("/+$", "", host)
  sub("^https?://", "", host)
}

#' Get default Azure Storage DFS endpoint suffix
#'
#' @description
#' Returns the default endpoint suffix used to construct Azure Data Lake Storage
#' Gen2 DFS URLs.
#'
#' @return A character string with the DFS endpoint suffix.
#'
#' @export
#' @examples
#' default_storage_endpoint()
default_storage_endpoint <- function() {
  azure_storage_endpoints$dfs
}

#' Get default Azure configuration directory
#'
#' @description
#' Retrieves the Azure configuration directory from the `AZURE_CONFIG_DIR`
#' environment variable, or falls back to the platform-specific default.
#'
#' @return A character string with the Azure configuration directory path
#'
#' @export
#' @examples
#' default_azure_config_dir()
default_azure_config_dir <- function() {
  Sys.getenv(
    "AZURE_CONFIG_DIR",
    unset = if (.Platform$OS.type == "windows") {
      normalizePath(
        file.path(Sys.getenv("USERPROFILE"), ".azure"),
        winslash = "/",
        mustWork = FALSE
      )
    } else {
      "~/.azure"
    }
  )
}

#' Get default MSAL token cache path
#'
#' @description
#' Returns the path to the MSAL token cache file shared by the Azure CLI and
#' Azure SDKs. Defaults to `msal_token_cache.json` inside the Azure config
#' directory (see [default_azure_config_dir()]).
#'
#' @return A character string with the path to the MSAL token cache file.
#'
#' @seealso [default_azure_config_dir()], [write_msal_token()]
#'
#' @export
default_msal_token_cache <- function() {
  file.path(default_azure_config_dir(), "msal_token_cache.json")
}


#' Get default OAuth redirect URI
#'
#' @description
#' Constructs a redirect URI for OAuth flows. If the provided URI doesn't have
#' a port, assigns a random port using [httpuv::randomPort()].
#'
#' @param redirect_uri A character string specifying the redirect URI. Defaults
#'   to [httr2::oauth_redirect_uri()].
#'
#' @return A character string with the redirect URI
#'
#' @export
#' @examples
#' default_redirect_uri()
default_redirect_uri <- function(redirect_uri = httr2::oauth_redirect_uri()) {
  parsed <- httr2::url_parse(redirect_uri)

  if (is.null(parsed$port)) {
    parsed$port <- random_port()
  }

  httr2::url_build(parsed)
}

#' Get default federated token file path
#'
#' @description
#' Retrieves the path to the federated identity token file from the
#' `AZURE_FEDERATED_TOKEN_FILE` environment variable, or returns `NULL` if
#' not set. Used by [WorkloadIdentityCredential].
#'
#' @return A character string with the file path, or `NULL` if not set
#'
#' @export
#' @examples
#' default_federated_token_file()
default_federated_token_file <- function() {
  res <- Sys.getenv(
    environment_variables$azure_federated_token_file,
    unset = NA_character_
  )

  if (is.na(res)) NULL else res
}

#' Azure CLI credential authentication
#'
#' @description
#' Authenticates using the Azure CLI (`az`) command-line tool. This credential
#' requires the Azure CLI to be installed and the user to be logged in via
#' `az login`.
#'
#' @details
#' The credential uses the `az account get-access-token` command to retrieve
#' access tokens. It will use the currently active Azure CLI account and
#' subscription unless a specific tenant is specified.
#'
#' @export
#' @examples
#' # Create credential with default settings
#' cred <- AzureCLICredential$new()
#'
#' # Create credential with specific scope and tenant
#' cred <- AzureCLICredential$new(
#'   scope = "https://management.azure.com/.default",
#'   tenant_id = "your-tenant-id"
#' )
#'
#' # To get a token or authenticate a request it is required that
#' # 'az login' is successfully executed, otherwise it will return an error.
#' \dontrun{
#' # Get an access token
#' token <- cred$get_token()
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' resp <- httr2::req_perform(cred$req_auth(req))
#' }
AzureCLICredential <- R6::R6Class(
  classname = "AzureCLICredential",
  inherit = Credential,
  public = list(
    #' @field interactive Logical indicating whether to check login status and
    #'   perform login if needed
    interactive = FALSE,
    #' @field .process_timeout Timeout in seconds for Azure CLI command execution
    .process_timeout = 10,

    #' @description
    #' Create a new Azure CLI credential
    #'
    #' @param scope A character string specifying the OAuth2 scope. Defaults to
    #'   `NULL`, which uses the scope set during initialization.
    #' @param tenant_id A character string specifying the Azure Active Directory
    #'   tenant ID. Defaults to `NULL`, which uses the default tenant from Azure CLI.
    #' @param process_timeout A numeric value specifying the timeout in seconds
    #'   for the Azure CLI process. Defaults to `10`.
    #' @param interactive A logical value indicating whether to check if the user is
    #'   logged in and perform login if needed. Defaults to `FALSE`.
    #' @param use_bridge A logical value indicating whether to use the device code
    #'   bridge webpage during login. If `TRUE`, launches an intermediate local webpage
    #'   that displays the device code and facilitates copy-pasting before redirecting
    #'   to the Microsoft device login page. Only used when `interactive = TRUE`. Defaults to `FALSE`.
    #'
    #' @return A new `AzureCLICredential` object
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      process_timeout = NULL,
      interactive = FALSE,
      use_bridge = FALSE
    ) {
      self$interactive <- interactive
      super$initialize(
        scope = scope,
        tenant_id = tenant_id
      )
      self$.process_timeout <- process_timeout %||% self$.process_timeout

      if (isTRUE(self$is_interactive())) {
        # Check if user is logged in
        if (!az_cli_is_login(timeout = self$.process_timeout)) {
          cli::cli_alert_info("User is not logged in to Azure CLI")
          az_cli_login(use_bridge = use_bridge)
        }
      }
    },
    #' @description
    #' Get an access token from Azure CLI
    #'
    #' @param scope A character string specifying the OAuth2 scope. If `NULL`,
    #'   uses the scope specified during initialization.
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function(scope = NULL) {
      # Check if user is logged in
      if (!az_cli_is_login(timeout = self$.process_timeout)) {
        cli::cli_abort(
          c(
            "User is not logged in to Azure CLI",
            "i" = "Please run {.code $login()} or use {.code az login} in your terminal"
          ),
          class = "azr_cli_not_logged_in"
        )
      }

      rlang::try_fetch(
        az_cli_get_token(
          scope = scope %||% self$.scope,
          tenant_id = self$.tenant_id,
          timeout = self$.process_timeout
        ),
        error = function(cnd) {
          cli::cli_abort(cnd$message, call = call("get_token"))
        }
      )
    },
    #' @description
    #' Add authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #' @param scope A character string specifying the OAuth2 scope. If `NULL`,
    #'   uses the scope specified during initialization.
    #'
    #' @return The request object with authentication header added
    req_auth = function(req, scope = NULL) {
      # Check if user is logged in
      if (!az_cli_is_login(timeout = self$.process_timeout)) {
        cli::cli_abort(
          c(
            "User is not logged in to Azure CLI",
            "i" = "Please run {.code $login()} or use {.code az login} in your terminal"
          ),
          class = "azr_cli_not_logged_in"
        )
      }

      token <- self$get_token(scope)
      httr2::req_auth_bearer_token(req, token$access_token)
    },
    #' @description
    #' Show the currently active Azure CLI account information
    #'
    #' @param timeout A numeric value specifying the timeout in seconds for the
    #'   Azure CLI command. If `NULL`, uses the process timeout specified during
    #'   initialization.
    #'
    #' @return A list containing the account information from Azure CLI
    account_show = function(timeout = NULL) {
      az_cli_account_show(timeout = timeout %||% self$.process_timeout)
    },
    #' @description
    #' Perform Azure CLI login using device code flow
    #'
    #' @return Invisibly returns the exit status (0 for success, non-zero for failure)
    login = function() {
      az_cli_login()
    },
    #' @description
    #' Check if the credential requires user interaction
    #'
    #' @return Logical indicating whether this credential is interactive
    is_interactive = function() {
      self$interactive
    },
    #' @description
    #' Log out from Azure CLI
    #'
    #' @return Invisibly returns `NULL`
    logout = function() {
      az_cli_logout()
    }
  )
)


#' Get Access Token from Azure CLI
#'
#' @description
#' Retrieves an access token from Azure CLI using the `az account get-access-token`
#' command. This is a lower-level function that directly interacts with the Azure
#' CLI to obtain OAuth2 tokens.
#'
#' @details
#' This function executes the Azure CLI command and parses the JSON response to
#' create an httr2 OAuth token object. The token includes the access token,
#' token type, and expiration time.
#'
#' @param scope A character string specifying the OAuth2 scope for which to
#'   request the access token (e.g., `"https://management.azure.com/.default"`).
#' @param tenant_id A character string specifying the Azure Active Directory
#'   tenant ID. If `NULL`, uses the default tenant from Azure CLI. Defaults to `NULL`.
#' @param timeout A numeric value specifying the timeout in seconds for the
#'   Azure CLI process. Defaults to `10`.
#'
#' @return An [httr2::oauth_token()] object containing:
#'   - `access_token`: The OAuth2 access token string
#'   - `token_type`: The type of token (typically "Bearer")
#'   - `.expires_at`: POSIXct timestamp when the token expires
#'
#' @export
az_cli_get_token <- function(scope, tenant_id = NULL, timeout = 10L) {
  args <- c("account", "get-access-token", "--output", "json")
  az_path <- az_cli_available()

  validate_scope(scope)
  args <- append(args, c("--scope", scope))

  if (!is.null(tenant_id)) {
    validate_tenant_id(tenant_id)
    args <- append(args, c("--tenant", tenant_id))
  }

  output <- suppressWarnings(system2(
    command = az_path,
    args = args,
    stdout = TRUE,
    stderr = TRUE,
    timeout = timeout
  ))

  status <- attr(output, "status")

  # Check for command failure
  if (!is.null(status) && status != 0L) {
    if (status == 124L) {
      cli::cli_abort(
        sprintf("Azure CLI command timed out after %d seconds", timeout),
        class = "azr_cli_timeout_error"
      )
    }

    # Provide context with the error
    error_msg <- paste(output, collapse = "\n")
    cli::cli_abort(
      sprintf("Azure CLI command failed (exit code %d): %s", status, error_msg),
      class = "azr_cli_error"
    )
  }

  # Check for empty output
  if (length(output) == 0 || !nzchar(paste(output, collapse = ""))) {
    cli::cli_abort(
      "Azure CLI returned empty output. Ensure you are logged in with 'az login'",
      class = "azr_cli_empty_output"
    )
  }

  # Parse JSON with error handling
  token <- tryCatch(
    jsonlite::fromJSON(paste(output, collapse = "\n")),
    error = function(e) {
      cli::cli_abort(
        sprintf("Failed to parse Azure CLI output as JSON: %s", e$message),
        class = "azr_cli_parse_error"
      )
    }
  )

  # Validate required fields
  required_fields <- c("accessToken", "tokenType", "expiresOn")
  missing_fields <- setdiff(required_fields, names(token))

  if (length(missing_fields) > 0) {
    cli::cli_abort(
      sprintf(
        "Azure CLI response missing required fields: %s",
        paste(missing_fields, collapse = ", ")
      ),
      class = "azr_cli_invalid_response"
    )
  }

  expires_at <- as.POSIXct(token$expiresOn)

  token_info <- tryCatch(
    find_msal_token(access_token = token$accessToken),
    error = function(e) NULL
  )

  token_args <- list(
    access_token = token$accessToken,
    token_type = token$tokenType,
    expires_at = expires_at
  )
  if (!is.null(token_info[["refresh_token"]])) {
    token_args$refresh_token <- token_info[["refresh_token"]]
  }
  if (!is.null(token_info[["scope"]])) {
    token_args$scope <- token_info[["scope"]]
  }

  do.call(httr2::oauth_token, token_args)
}


az_cli_available <- function() {
  az_path <- Sys.which("az")

  if (!nzchar(az_path)) {
    cli::cli_abort(
      c(
        "Azure CLI not found on PATH",
        "i" = "Install Azure CLI: https://docs.microsoft.com/cli/azure/install-azure-cli",
        "i" = "Or ensure 'az' is in your system PATH"
      ),
      class = "azr_cli_not_found"
    )
  }

  az_path
}


#' Check if User is Logged in to Azure CLI
#'
#' @description
#' Checks whether the user is currently logged in to Azure CLI by attempting
#' to retrieve account information.
#'
#' @param timeout A numeric value specifying the timeout in seconds for the
#'   Azure CLI command. Defaults to `10`.
#'
#' @return A logical value: `TRUE` if the user is logged in, `FALSE` otherwise
#'
#' @export
az_cli_is_login <- function(timeout = 10L) {
  tryCatch(
    {
      az_cli_account_show(timeout = timeout)
      TRUE
    },
    error = function(e) {
      FALSE
    }
  )
}


#' Azure CLI Device Code Login
#'
#' @description
#' Performs an interactive Azure CLI login using device code flow.
#' Automatically captures the device code, copies it to the clipboard,
#' and opens the browser for authentication.
#'
#' @details
#' This function runs `az login --use-device-code`, monitors the output
#' to extract the device code, copies it to the clipboard, and opens
#' the authentication URL in the default browser.
#'
#' @param tenant_id A character string specifying the Azure Active Directory
#'   tenant ID to authenticate against. If `NULL` (default), uses the default
#'   tenant from Azure CLI configuration.
#' @param use_bridge A logical value indicating whether to use the device code
#'   bridge webpage. If `TRUE`, launches an intermediate local webpage that
#'   displays the device code and facilitates copy-pasting before redirecting
#'   to the Microsoft device login page. If `FALSE` (default), copies the code
#'   directly to the clipboard and opens the Microsoft login page.
#' @param verbose A logical value indicating whether to print detailed process
#'   output to the console, including error messages from the Azure CLI process.
#'   If `FALSE` (default), only essential messages are displayed.
#'
#' @return Invisibly returns the exit status (0 for success, non-zero for failure)
#'
#' @export
az_cli_login <- function(
  tenant_id = NULL,
  use_bridge = FALSE,
  verbose = FALSE
) {
  if (!rlang::is_interactive()) {
    cli::cli_abort(
      c(
        "Azure CLI login requires an interactive session",
        "i" = "This function cannot be used in non-interactive environments"
      ),
      class = "azr_cli_non_interactive"
    )
  }

  az_path <- az_cli_available()
  az_args <- c("login", "--use-device-code", "--output", "json")

  if (!is.null(tenant_id)) {
    validate_tenant_id(tenant_id)
    az_args <- c(az_args, c("--tenant", tenant_id))
  }

  cli::cli_alert_info("Starting Azure CLI login process...")

  output_file <- tempfile(fileext = ".json")

  rlang::check_installed("processx")

  # Start the process in the background
  p <- processx::process$new(
    az_path,
    az_args,
    stdout = output_file,
    stderr = "|"
  )

  # Variables to track state
  code_found <- FALSE
  login_url <- "https://microsoft.com/devicelogin"
  all_stdout <- character()

  # Loop to read output while the process is alive
  while (p$is_alive()) {
    # Poll for output (wait up to 100ms)
    p$poll_io(100)

    # Read stdout and stderr
    lines_err <- p$read_error_lines()

    # Collect stdout for later parsing
    if (length(lines_err) > 0) {
      all_stdout <- c(all_stdout, lines_err)
    }

    # Print output to console
    if (isTRUE((verbose) && lines_err) > 0) {
      cli::cli_alert_warning("Error output:")
      for (line in lines_err) {
        cli::cli_text(line)
      }
    }

    # Look for the device code
    if (!code_found && length(lines_err) > 0) {
      # Combine lines to handle wrapping
      combined_text <- paste(lines_err, collapse = " ")

      # Look for the code pattern (usually 9 alphanumeric characters)
      match <- regexec(
        "enter the code ([A-Z0-9]+) to authenticate",
        combined_text
      )
      match_text <- regmatches(combined_text, match)[[1]]

      if (length(match_text) > 1) {
        device_code <- match_text[2]
        code_found <- TRUE

        cli::cli_alert_success("Found Device Code: {.val {device_code}}")

        if (isTRUE(use_bridge)) {
          tryCatch(
            {
              launch_device_code(device_code)
              cli::cli_alert_info(
                "Launch web bridge to copy code to clipboard!"
              )
            },
            error = function(e) {
              cli::cli_alert_warning(
                "Could not use browser. Please copy manually."
              )
            }
          )
        } else {
          # Copy to clipboard
          rlang::check_installed("clipr")
          tryCatch(
            {
              clipr::write_clip(device_code)
              cli::cli_alert_info("Code copied to clipboard! [Cmd/Ctrl + V]")
            },
            error = function(e) {
              cli::cli_alert_warning(
                "Could not write to clipboard. Please copy manually."
              )
            }
          )

          cli::cli_alert_info("Opening browser to {.url {login_url}}...")
          utils::browseURL(login_url)
        }
      }
    }

    # Sleep briefly to prevent 100% CPU usage
    Sys.sleep(0.1)
  }

  # Check exit status
  exit_status <- p$get_exit_status()
  error_lines <- p$read_error_lines()

  # Print any remaining error lines
  if (isTRUE(verbose) && length(error_lines) > 0) {
    cli::cli_alert_warning("Error output:")
    for (line in error_lines) {
      cli::cli_text(line)
    }
  }

  if (exit_status == 0) {
    cli::cli_alert_success("Login Completed Successfully!")

    result <- tryCatch(
      jsonlite::fromJSON(output_file),
      error = function(e) {
        NULL
      }
    )
    return(invisible(result))
  } else {
    cli::cli_alert_danger("Process finished with error (Status: {exit_status})")
  }

  invisible(exit_status)
}


#' Show Azure CLI Account Information
#'
#' @description
#' Retrieves information about the currently active Azure CLI account and
#' subscription. This function runs `az account show` and parses the JSON
#' output into an R list.
#'
#' @details
#' The function returns details about the current Azure subscription including:
#' - Subscription ID and name
#' - Tenant ID
#' - Account state (e.g., "Enabled")
#' - User information
#' - Cloud environment details
#'
#' @param timeout An integer specifying the timeout in seconds for the Azure
#'   CLI command. Defaults to `10`.
#'
#' @return A list containing the account information from Azure CLI
#'
#' @export
az_cli_account_show <- function(timeout = 10L) {
  az_path <- az_cli_available()
  args <- c("account", "show", "--output", "json")

  output <- suppressWarnings(system2(
    command = az_path,
    args = args,
    stdout = TRUE,
    stderr = TRUE,
    timeout = timeout
  ))

  status <- attr(output, "status")

  # Check for command failure
  if (!is.null(status) && status != 0L) {
    if (status == 124L) {
      cli::cli_abort(
        sprintf("Azure CLI command timed out after %d seconds", timeout),
        class = "azr_cli_timeout_error"
      )
    }

    # Provide context with the error
    error_msg <- paste(output, collapse = "\n")
    cli::cli_abort(
      sprintf("Azure CLI command failed (exit code %d): %s", status, error_msg),
      class = "azr_cli_error"
    )
  }

  # Check for empty output
  if (length(output) == 0 || !nzchar(paste(output, collapse = ""))) {
    cli::cli_abort(
      "Azure CLI returned empty output. Ensure you are logged in with 'az login'",
      class = "azr_cli_empty_output"
    )
  }

  # Parse JSON with error handling
  account_info <- tryCatch(
    jsonlite::fromJSON(paste(output, collapse = "\n")),
    error = function(e) {
      cli::cli_abort(
        sprintf("Failed to parse Azure CLI output as JSON: %s", e$message),
        class = "azr_cli_parse_error"
      )
    }
  )

  return(account_info)
}


#' Azure CLI Logout
#'
#' @description
#' Logs out from Azure CLI by removing all stored credentials and account
#' information. This function runs `az logout`.
#'
#' @details
#' After logging out, you will need to run [az_cli_login()] again to
#' authenticate and use Azure CLI credentials.
#'
#' @return Invisibly returns `NULL`
#'
#' @export
az_cli_logout <- function() {
  az_path <- az_cli_available()
  args <- c("logout")

  cli::cli_alert_info("Logging out from Azure CLI...")

  output <- suppressWarnings(system2(
    command = az_path,
    args = args,
    stdout = TRUE,
    stderr = TRUE
  ))

  status <- attr(output, "status")

  if (!is.null(status) && status != 0L) {
    error_msg <- paste(output, collapse = "\n")

    cli::cli_bullets(c(
      c(
        "!" = "Azure CLI logout failed (exit code {.val {status}})",
        "x" = error_msg
      )
    ))
  } else {
    cli::cli_alert_success("Successfully logged out from Azure CLI")
  }

  invisible(NULL)
}


#' Write an httr2 Token to the MSAL Token Cache
#'
#' @description
#' Writes an [httr2::oauth_token()] object into the MSAL token cache JSON file
#' (`msal_token_cache.json`) shared by the Azure SDK and Azure CLI. The
#' resulting entry is readable by other Azure tools (Python SDK, Azure CLI,
#' and the rest of this package via [az_cli_get_cached_token()]).
#'
#' @details
#' The function adds or overwrites `AccessToken`, `RefreshToken` (when the
#' token carries a refresh token), `Account`, and `AppMetadata` sections.
#' Existing entries for other accounts or clients are preserved.
#'
#' The `home_account_id` follows the MSAL convention
#' `"<object_id>.<tenant_id>"` where `object_id` is the Azure AD OID of the
#' authenticated principal. Cache entry keys are built in the same format used
#' by the Azure CLI and MSAL Python:
#' \itemize{
#'   \item AccessToken: `<home_account_id>-<environment>-accesstoken-<client_id>-<realm>-<target>`
#'   \item RefreshToken: `<home_account_id>-<environment>-refreshtoken-<client_id>--`
#'   \item Account: `<home_account_id>-<environment>-<realm>`
#'   \item AppMetadata: `appmetadata-<environment>-<client_id>`
#' }
#'
#' @param token An [httr2::oauth_token()] object. Must contain `access_token`,
#'   `token_type`, and `.expires_at`. May optionally contain `refresh_token`
#'   and `scope`. All cache fields (`home_account_id`, `tenant_id`, `username`,
#'   `client_id`, `scope`, `environment`) are derived from the JWT claims
#'   (`oid`, `tid`, `upn`/`preferred_username`, `appid`/`azp`, `scp`/`scope`,
#'   `iss`) and the token object itself.
#' @param cache_file Path to the MSAL token cache JSON file. Defaults to
#'   [default_msal_token_cache()].
#'
#' @return Invisibly returns the path to the cache file.
#'
#' @seealso [az_cli_get_cached_token()], [httr2::oauth_token()]
#'
#' @export
write_msal_token <- function(
  token,
  cache_file = default_msal_token_cache()
) {
  fields <- extract_msal_token_fields(token)
  cache <- build_msal_cache_entries(token, fields, cache_file)
  write_msal_cache(cache, cache_file)
}

# Extracts and validates identity fields from an httr2_token object by decoding
# its JWT access token claims.
extract_msal_token_fields <- function(token) {
  if (!inherits(token, "httr2_token")) {
    cli::cli_abort(
      "{.arg token} must be an {.cls httr2_token} object, not {.cls {class(token)[[1L]]}}.",
      call = rlang::caller_env()
    )
  }

  claims <- decode_jwt_claims(token$access_token)

  oid <- claims[["oid"]]
  tid <- claims[["tid"]]
  client_id <- claims[["appid"]] %||% claims[["azp"]]
  if (is.null(oid) || is.null(tid) || is.null(client_id)) {
    missing <- names(Filter(
      is.null,
      list(oid = oid, tid = tid, appid = client_id)
    ))
    cli::cli_abort(
      c(
        "Cannot derive account identity from the access token.",
        "i" = "The JWT is missing {.field {missing}} claim{?s}."
      )
    )
  }

  home_account_id <- paste0(oid, ".", tid)

  iss <- claims[["iss"]]
  environment <- if (!is.null(iss)) {
    host <- sub("^https?://", "", iss)
    sub("/.*$", "", host)
  } else {
    host <- sub("^https?://", "", default_azure_host())
    sub("/$", "", host)
  }

  list(
    home_account_id = home_account_id,
    local_account_id = oid,
    tenant_id = tid,
    client_id = client_id,
    username = claims[["upn"]] %||% claims[["preferred_username"]],
    environment = environment,
    scope_str = paste(token$scope %||% character(0L), collapse = " ")
  )
}

# Merges new MSAL cache entries built from a token and its extracted fields into
# an existing cache list (read from disk if present).
build_msal_cache_entries <- function(token, fields, cache_file) {
  now_unix <- as.character(as.integer(Sys.time()))
  expires_unix <- as.character(as.integer(token$.expires_at))
  ext_expires <- as.character(as.integer(token$.expires_at) + 86400L)

  # All key components are lowercased per MSAL convention
  env_l <- tolower(fields$environment)
  client_l <- tolower(fields$client_id)
  realm_l <- tolower(fields$tenant_id)
  acct_l <- tolower(fields$home_account_id)
  target_l <- tolower(fields$scope_str)

  at_key <- paste(
    acct_l,
    env_l,
    "accesstoken",
    client_l,
    realm_l,
    target_l,
    sep = "-"
  )
  # Refresh token key has empty family_id and empty target, producing double trailing dash
  rt_key <- paste(acct_l, env_l, "refreshtoken", client_l, "", "", sep = "-")
  account_key <- paste(acct_l, env_l, realm_l, sep = "-")
  app_key <- paste("appmetadata", env_l, client_l, sep = "-")

  cache <- if (file.exists(cache_file)) {
    tryCatch(read_msal_cache(cache_file), error = function(e) list())
  } else {
    list()
  }

  for (section in c(
    "AccessToken",
    "RefreshToken",
    "IdToken",
    "Account",
    "AppMetadata"
  )) {
    if (is.null(cache[[section]])) cache[[section]] <- list()
  }

  cache$AccessToken[[at_key]] <- list(
    home_account_id = fields$home_account_id,
    environment = fields$environment,
    client_id = fields$client_id,
    target = fields$scope_str,
    realm = fields$tenant_id,
    token_type = token$token_type %||% "Bearer",
    cached_at = now_unix,
    expires_on = expires_unix,
    extended_expires_on = ext_expires,
    secret = token$access_token,
    credential_type = "AccessToken"
  )

  if (nzchar(token$refresh_token %||% "")) {
    cache$RefreshToken[[rt_key]] <- list(
      home_account_id = fields$home_account_id,
      environment = fields$environment,
      client_id = fields$client_id,
      target = fields$scope_str,
      secret = token$refresh_token,
      credential_type = "RefreshToken"
    )
  }

  acct_entry <- list(
    home_account_id = fields$home_account_id,
    environment = fields$environment,
    realm = fields$tenant_id,
    authority_type = "MSSTS",
    local_account_id = fields$local_account_id
  )
  if (!is.null(fields$username)) {
    acct_entry$username <- fields$username
  }
  cache$Account[[account_key]] <- acct_entry

  cache$AppMetadata[[app_key]] <- list(
    client_id = fields$client_id,
    environment = fields$environment
  )

  cache
}

# Serialises a cache list to the MSAL JSON cache file, creating parent
# directories as needed.
write_msal_cache <- function(cache, cache_file) {
  cache_dir <- dirname(cache_file)
  if (!dir.exists(cache_dir)) {
    dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  }
  jsonlite::write_json(cache, cache_file, pretty = TRUE, auto_unbox = TRUE)
  invisible(cache_file)
}


# Reads and parses the MSAL token cache JSON file. Aborts if the file does not
# exist or cannot be parsed.
read_msal_cache <- function(cache_file) {
  if (!file.exists(cache_file)) {
    cli::cli_abort(
      c(
        "MSAL token cache file not found.",
        "i" = "Expected file at: {.path {cache_file}}"
      ),
      class = "azr_cli_cache_not_found_error"
    )
  }
  tryCatch(
    jsonlite::fromJSON(cache_file, simplifyDataFrame = FALSE),
    error = function(e) {
      cli::cli_abort(
        c(
          "Failed to parse MSAL token cache: {e$message}",
          "i" = "The cache file may be corrupted: {.path {cache_file}}"
        ),
        class = "azr_cli_cache_parse_error"
      )
    }
  )
}

# Decodes the payload of a JWT (without verifying the signature) and returns
# the claims as a named list. Returns an empty list if the token is not a
# well-formed JWT.
decode_jwt_claims <- function(jwt) {
  parts <- strsplit(jwt, ".", fixed = TRUE)[[1L]]
  if (length(parts) < 2L) {
    return(list())
  }
  payload <- parts[[2L]]
  # base64url -> base64: replace chars and add padding
  payload <- gsub("-", "+", payload, fixed = TRUE)
  payload <- gsub("_", "/", payload, fixed = TRUE)
  pad <- (4L - nchar(payload) %% 4L) %% 4L
  payload <- paste0(payload, strrep("=", pad))
  tryCatch(
    jsonlite::fromJSON(
      rawToChar(jsonlite::base64_dec(payload)),
      simplifyDataFrame = FALSE
    ),
    error = function(e) list()
  )
}

# Reads the MSAL token cache JSON and returns a list with `refresh_token` and
# `scope` matching the given credentials. If `access_token` is provided its
# value is matched against the `secret` field of every `AccessToken` entry to
# resolve `home_account_id`, `client_id`, and `scope`; otherwise both
# `home_account_id` and `client_id` must be supplied directly.
find_msal_token <- function(
  cache_file = default_msal_token_cache(),
  home_account_id = NULL,
  client_id = NULL,
  access_token = NULL
) {
  cache <- read_msal_cache(cache_file)

  scope <- NULL
  if (!is.null(access_token)) {
    for (tok in cache$AccessToken %||% list()) {
      if (identical(tok$secret, access_token)) {
        home_account_id <- tok$home_account_id
        client_id <- tok$client_id
        scope <- tok$target
        break
      }
    }
  }

  if (is.null(home_account_id) || is.null(client_id)) {
    return(NULL)
  }

  refresh_token <- NULL
  for (rt in cache$RefreshToken %||% list()) {
    if (
      identical(rt$home_account_id, home_account_id) &&
        identical(rt$client_id, client_id)
    ) {
      refresh_token <- rt$secret
      break
    }
  }

  list(refresh_token = refresh_token, scope = scope)
}

#' Get Cached Token from MSAL Token Cache
#'
#' @description
#' Reads the MSAL token cache file (`msal_token_cache.json`) from the Azure
#' configuration directory and returns a matching access token as an
#' [httr2::oauth_token()] object.
#'
#' @details
#' The MSAL token cache is a JSON file maintained by the Azure CLI that stores
#' access tokens and refresh tokens. This function reads cached access tokens
#' directly from the file without invoking the Azure CLI, which can be useful
#' in environments where the CLI is slow or unavailable but tokens have been
#' previously cached.
#'
#' When multiple tokens are found, the function selects the token that expires
#' latest. If `scope` is provided, only tokens matching that scope/resource are
#' returned.
#'
#' @param scope A character string specifying the OAuth2 scope to filter tokens.
#'   If `NULL` (default), returns the latest-expiring token regardless of scope.
#' @param tenant_id A character string specifying the tenant ID to filter tokens.
#'   If `NULL` (default), matches any tenant.
#' @param client_id A character string specifying the client ID to filter tokens.
#'   If `NULL` (default), matches any client.
#' @param config_dir A character string specifying the Azure configuration
#'   directory. Defaults to [default_azure_config_dir()].
#'
#' @return An [httr2::oauth_token()] object containing:
#'   - `access_token`: The OAuth2 access token string
#'   - `token_type`: The type of token (typically "Bearer")
#'   - `.expires_at`: POSIXct timestamp when the token expires
#'
#' @export
az_cli_get_cached_token <- function(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  config_dir = default_azure_config_dir()
) {
  cache_file <- file.path(config_dir, "msal_token_cache.json")

  if (!file.exists(cache_file)) {
    cli::cli_abort(
      c(
        "MSAL token cache file not found at {.path {cache_file}}",
        "i" = "Ensure you have logged in with {.code az login}"
      ),
      class = "azr_cli_cache_not_found"
    )
  }

  cache <- read_msal_cache(cache_file)

  tokens <- cache$AccessToken
  if (is.null(tokens) || length(tokens) == 0L) {
    cli::cli_abort(
      c(
        "No access tokens found in MSAL token cache",
        "i" = "Ensure you have logged in with {.code az login}"
      ),
      class = "azr_cli_cache_empty"
    )
  }

  # Filter by scope/resource
  if (!is.null(scope)) {
    resource <- get_scope_resource(scope)
    tokens <- Filter(
      function(tok) {
        target <- tok$target %||% ""
        # Match if scope appears in target or resource matches
        grepl(scope, target, fixed = TRUE) ||
          (!is.null(resource) && grepl(resource, target, fixed = TRUE))
      },
      tokens
    )
  }

  # Filter by tenant_id (realm field in MSAL cache)
  if (!is.null(tenant_id)) {
    tokens <- Filter(
      function(tok) {
        identical(tok$realm, tenant_id)
      },
      tokens
    )
  }

  # Filter by client_id
  if (!is.null(client_id)) {
    tokens <- Filter(
      function(tok) {
        identical(tok$client_id, client_id)
      },
      tokens
    )
  }

  if (length(tokens) == 0L) {
    cli::cli_abort(
      c(
        "No matching access token found in MSAL token cache",
        "i" = "Try logging in with {.code az login} to refresh your tokens"
      ),
      class = "azr_cli_cache_no_match"
    )
  }

  # Select the token with the latest expiry
  expires <- vapply(
    tokens,
    function(tok) {
      as.numeric(tok$expires_on %||% "0")
    },
    numeric(1)
  )

  best <- tokens[[which.max(expires)]]

  expires_at <- as.POSIXct(
    as.numeric(best$expires_on),
    origin = "1970-01-01",
    tz = "UTC"
  )

  token_info <- find_msal_token(
    cache_file,
    access_token = best$secret
  )

  httr2::oauth_token(
    access_token = best$secret,
    token_type = best$token_type %||% "Bearer",
    refresh_token = token_info[["refresh_token"]],
    .expires_at = expires_at
  )
}


launch_device_code <- function(code) {
  html_content <- system.file("www/code.html", package = "azr") |>
    readLines() |>
    paste(collapse = "\n") |>
    sprintf(code)

  temp_file <- tempfile(fileext = ".html")
  writeLines(html_content, temp_file)
  utils::browseURL(temp_file)
}

# ClientSecretCredential ----
#' Client secret credential authentication
#'
#' @description
#' Authenticates a service principal using a client ID and client secret.
#' This credential is commonly used for application authentication in Azure.
#'
#' @details
#' The credential uses the OAuth 2.0 client credentials flow to obtain access
#' tokens. It requires a registered Azure AD application with a client secret.
#' The client secret should be stored securely and not hard-coded in scripts.
#'
#' @export
#' @examples
#' # Create credential with client secret
#' cred <- ClientSecretCredential$new(
#'   tenant_id = "your-tenant-id",
#'   client_id = "your-client-id",
#'   client_secret = "your-client-secret",
#'   scope = "https://management.azure.com/.default"
#' )
#'
#' # To get a token or authenticate a request it requires
#' # valid 'client_id' and 'client_secret' credentials,
#' # otherwise it will return an error.
#' \dontrun{
#' # Get an access token
#' token <- cred$get_token()
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' resp <- httr2::req_perform(cred$req_auth(req))
#' }
ClientSecretCredential <- R6::R6Class(
  classname = "ClientSecretCredential",
  inherit = Credential,
  public = list(
    #' @description
    #' Validate the credential configuration
    #'
    #' @details
    #' Checks that the client secret is provided and not NA or NULL. Calls the parent
    #' class validation method.
    validate = function() {
      super$validate()

      if (is.null(self$.client_secret) || rlang::is_na(self$.client_secret)) {
        cli::cli_abort("Argument {.arg client_secret} cannot be NULL or NA.")
      }
    },
    #' @description
    #' Get an access token using client credentials flow
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function() {
      httr2::oauth_flow_client_credentials(
        client = self$.oauth_client,
        scope = self$.scope_str
      )
    },
    #' @description
    #' Add OAuth client credentials authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with OAuth client credentials authentication configured
    req_auth = function(req) {
      httr2::req_oauth_client_credentials(
        req = req,
        client = self$.oauth_client,
        scope = self$.scope_str
      )
    }
  )
)

#' Azure Storage API Class
#'
#' @description
#' An R6 class that extends [api_client] to provide specialized methods
#' for Azure Data Lake Storage Gen2 (ADLS Gen2) REST API operations.
#'
#' @details
#' The base URL is constructed as:
#' `https://{storageaccount}.{endpoint_suffix}`
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a storage client
#' storage <- api_storage_client$new(
#'   storageaccount = "mystorageaccount",
#'   filesystem = "mycontainer"
#' )
#'
#' # List files in the root directory
#' files <- storage$list_files()
#'
#' # List files in a specific path
#' files <- storage$list_files(path = "data/folder1")
#'
#' # List files recursively
#' files <- storage$list_files(path = "data", recursive = TRUE)
#' }
api_storage_client <- R6::R6Class(
  classname = "api_storage_client",
  inherit = api_client,
  public = list(
    #' @field .filesystem The filesystem (container) name
    .filesystem = NULL,

    #' @description
    #' Create a new Azure Storage API client instance
    #'
    #' @param storageaccount A character string specifying the Azure Storage account name.
    #' @param filesystem A character string specifying the filesystem (container) name.
    #' @param scope A character string specifying the OAuth2 scope. Defaults to
    #'   `default_azure_scope("azure_storage")`.
    #' @param endpoint_suffix A character string specifying the Azure
    #'   Storage DFS endpoint suffix. Defaults to
    #'   [default_storage_endpoint()].
    #' @param provider An optional credential provider object that inherits from
    #'   `Credential` or `DefaultCredential`. If provided, `chain` is ignored.
    #' @param chain A [credential_chain] instance for authentication. If NULL,
    #'   a default credential chain will be created using [DefaultCredential].
    #' @param tenant_id A character string specifying the Azure tenant ID. Passed to
    #'   [DefaultCredential] when `chain` is `NULL`.
    #' @param ... Additional arguments passed to the parent [api_client] constructor.
    #'
    #' @return A new `api_storage_client` object
    initialize = function(
      storageaccount,
      filesystem,
      scope = default_azure_scope("azure_storage"),
      endpoint_suffix = default_storage_endpoint(),
      provider = NULL,
      chain = NULL,
      tenant_id = NULL,
      ...
    ) {
      if (missing(storageaccount) || is.null(storageaccount)) {
        cli::cli_abort("{.arg storageaccount} must not be {.val NULL}.")
      }
      if (missing(filesystem) || is.null(filesystem)) {
        cli::cli_abort("{.arg filesystem} must not be {.val NULL}.")
      }
      if (!is.character(filesystem) || length(filesystem) != 1L) {
        cli::cli_abort("{.arg filesystem} must be a single character string.")
      }
      if (
        !is.character(endpoint_suffix) ||
          length(endpoint_suffix) != 1L ||
          !nzchar(endpoint_suffix)
      ) {
        cli::cli_abort(
          "{.arg endpoint_suffix} must be a non-empty character string."
        )
      }

      host_url <- storage_host_url(
        storageaccount = storageaccount,
        endpoint_suffix = endpoint_suffix
      )

      if (length(scope) > 1) {
        scope <- paste(scope, collapse = " ")
      }

      # Create credential provider
      if (is.null(provider)) {
        provider <- DefaultCredential$new(
          scope = scope,
          chain = chain,
          tenant_id = tenant_id
        )
      } else if (
        !R6::is.R6(provider) ||
          !(inherits(provider, "Credential") ||
            inherits(provider, "DefaultCredential"))
      ) {
        cli::cli_abort(
          "Argument {.arg provider} must inherit from {.cls Credential} or {.cls DefaultCredential}."
        )
      }

      self$.filesystem <- filesystem

      super$initialize(host_url = host_url, provider = provider, ...)

      lockBinding(".filesystem", self)
    },

    #' @description
    #' Download a file from the filesystem
    #'
    #' @param path A character string specifying the file path within the filesystem.
    #' @param dest A character string specifying the local destination path.
    #'   Defaults to a temporary file via [tempfile()].
    #'
    #' @return The local path the file was written to (invisibly).
    download_file = function(path, dest = NULL) {
      if (missing(path) || is.null(path) || !nzchar(path)) {
        cli::cli_abort("{.arg path} must be a non-empty character string.")
      }
      if (is.null(dest)) {
        dest <- tempfile()
      }

      resp <- self$.fetch(
        path = paste0(self$.filesystem, "/", path),
        method = "get",
        content = "response"
      )

      writeBin(httr2::resp_body_raw(resp), dest)

      invisible(dest)
    },

    #' @description
    #' Get the access control list (ACL) for a file or directory
    #'
    #' @param dataset A character string specifying the file or directory path within
    #'   the filesystem.
    #' @param upn A logical value. If `TRUE`, user principal names (UPN) are
    #'   returned in the `x-ms-owner`, `x-ms-group`, and `x-ms-acl` response
    #'   headers instead of object IDs. Defaults to `FALSE`.
    #'
    #' @return A data.frame with columns `group_id` and `permission`, one row per
    #'   named group entry in the `x-ms-acl` response header.
    get_access_control = function(dataset, upn = FALSE) {
      if (missing(dataset) || is.null(dataset) || !nzchar(dataset)) {
        cli::cli_abort("{.arg dataset} must be a non-empty character string.")
      }

      resp <- self$.fetch(
        path = paste0(self$.filesystem, "/{dataset}"),
        dataset = dataset,
        query = list(
          action = "getAccessControl",
          upn = tolower(as.character(upn))
        ),
        headers = list(`x-ms-version` = "2023-11-03"),
        method = "head",
        content = "headers"
      )

      acl_raw <- resp[["x-ms-acl"]]
      entries <- strsplit(acl_raw, ",")[[1]]
      group_entries <- entries[grepl("^group:[^:]+:", entries)]
      parts <- strsplit(group_entries, ":")

      data.frame(
        group_id = vapply(parts, `[[`, character(1L), 2L),
        permission = vapply(parts, `[[`, character(1L), 3L),
        stringsAsFactors = FALSE
      )
    },

    #' @description
    #' List files and directories in a path
    #'
    #' @param path A character string specifying the directory path to list.
    #'   Use empty string or NULL for the root directory. Defaults to `""`.
    #' @param recursive A logical value indicating whether to list files recursively.
    #'   Defaults to `FALSE`.
    #' @param ... Additional query parameters to pass to the API.
    #'
    #' @return A data.frame (or data.table if available) containing file and directory
    #'   information with columns such as name, contentLength, lastModified, etc.
    list_files = function(path = "", recursive = FALSE, ...) {
      if (is.null(path)) {
        path <- ""
      }

      query_params <- list(
        resource = "filesystem",
        recursive = tolower(as.character(recursive)),
        ...
      )

      if (nzchar(path)) {
        query_params$directory <- path
      }

      response <- self$.fetch(
        path = self$.filesystem,
        query = query_params,
        method = "get"
      )

      if (!is.null(response$paths) && is.data.frame(response$paths)) {
        return(response$paths)
      } else {
        cli::cli_inform("No files found in path: {.path {path}}")
        return(data.frame())
      }
    }
  )
)

storage_host_url <- function(storageaccount, endpoint_suffix) {
  endpoint_suffix <- sub("/+$", "", endpoint_suffix)
  endpoint_suffix <- sub("^https?://", "", endpoint_suffix)
  endpoint_suffix <- sub("^\\.+", "", endpoint_suffix)
  if (!nzchar(endpoint_suffix)) {
    cli::cli_abort(
      "{.arg endpoint_suffix} must be a non-empty character string."
    )
  }

  rlang::englue("https://{storageaccount}.{endpoint_suffix}")
}

#' Create an Azure Storage Client
#'
#' @description
#' A convenience wrapper around [api_storage_client] that creates a configured
#' client for Azure Data Lake Storage Gen2 (ADLS Gen2) REST API operations.
#'
#' @param storageaccount A character string specifying the Azure Storage account name.
#' @param filesystem A character string specifying the filesystem (container) name.
#' @param endpoint_suffix A character string specifying the Azure
#'   Storage DFS endpoint suffix. Defaults to
#'   [default_storage_endpoint()].
#' @param scope A character string specifying the OAuth2 scope. Defaults to
#'   `default_azure_scope("azure_storage")`.
#' @param provider An optional credential provider object that inherits from
#'   `Credential` or `DefaultCredential`. If provided, `chain` is ignored.
#' @param chain A [credential_chain] instance for authentication. Defaults to
#'   [default_credential_chain()].
#' @param tenant_id A character string specifying the Azure tenant ID. Defaults to
#'   [default_azure_tenant_id()], which reads `AZURE_TENANT_ID` from the environment.
#' @param ... Additional arguments passed to the [api_storage_client] constructor.
#'
#' @return An [api_storage_client] object.
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a storage client with default credentials
#' storage <- azr_storage_client(
#'   storageaccount = "mystorageaccount",
#'   filesystem = "mycontainer"
#' )
#'
#' # Create a storage client with a specific tenant
#' storage <- azr_storage_client(
#'   storageaccount = "mystorageaccount",
#'   filesystem = "mycontainer",
#'   tenant_id = "00000000-0000-0000-0000-000000000000"
#' )
#' }
azr_storage_client <- function(
  storageaccount,
  filesystem,
  endpoint_suffix = default_storage_endpoint(),
  scope = default_azure_scope("azure_storage"),
  provider = NULL,
  chain = default_credential_chain(),
  tenant_id = default_azure_tenant_id(),
  ...
) {
  api_storage_client$new(
    storageaccount = storageaccount,
    filesystem = filesystem,
    endpoint_suffix = endpoint_suffix,
    scope = scope,
    provider = provider,
    chain = chain,
    tenant_id = tenant_id,
    ...
  )
}

#' Cached token credential authentication
#'
#' @description
#' A credential class that retrieves tokens from the cache only, without
#' triggering interactive authentication flows. This is useful for non-interactive
#' sessions where you want to use previously cached tokens from DeviceCode or
#' AuthCode credentials.
#'
#' @details
#' This credential attempts to retrieve cached tokens from a chain of interactive
#' credentials (AuthCode and DeviceCode by default). It will not prompt for new
#' authentication - it only returns tokens that are already cached.
#'
#' This is particularly useful for:
#' \itemize{
#'   \item Non-interactive R sessions (e.g., scheduled scripts, CI/CD)
#'   \item Scenarios where you've previously authenticated interactively and want
#'     to reuse those cached tokens
#' }
#'
#' @field .scope Character string specifying the authentication scope.
#' @field .tenant_id Character string specifying the tenant ID.
#' @field .client_id Character string specifying the client ID.
#' @field .chain List of credential classes to attempt for cached tokens.
#'
#' @export
#' @examples
#' \dontrun{
#' # Create credential with default settings
#' cred <- CachedTokenCredential$new(
#'   scope = "https://graph.microsoft.com/.default",
#'   tenant_id = "my-tenant-id"
#' )
#'
#' # Get a cached token (will fail if no cached token exists)
#' token <- cred$get_token()
#'
#' # Use with httr2 request
#' req <- httr2::request("https://graph.microsoft.com/v1.0/me")
#' req <- cred$req_auth(req)
#' }
CachedTokenCredential <- R6::R6Class(
  classname = "CachedTokenCredential",
  public = list(
    .scope = NULL,
    .tenant_id = NULL,
    .client_id = NULL,
    .chain = NULL,

    #' @description
    #' Create a new CachedTokenCredential object
    #'
    #' @param scope Optional character string specifying the authentication scope.
    #' @param tenant_id Optional character string specifying the tenant ID for
    #'   authentication.
    #' @param client_id Optional character string specifying the client ID for
    #'   authentication.
    #' @param chain A list of credential classes to attempt for cached tokens.
    #'   Defaults to AuthCodeCredential and DeviceCodeCredential.
    #'
    #' @return A new `CachedTokenCredential` object
    initialize = function(
      scope = NULL,
      tenant_id = NULL,
      client_id = NULL,
      chain = cached_token_credential_chain()
    ) {
      self$.scope <- scope
      self$.tenant_id <- tenant_id
      self$.client_id <- client_id
      self$.chain <- chain
    },

    #' @description
    #' Get an access token from the cache
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function() {
      self$provider$get_token()
    },

    #' @description
    #' Add authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with authentication configured
    req_auth = function(req) {
      self$provider$req_auth(req)
    }
  ),
  active = list(
    #' @field provider Lazily initialized credential provider
    provider = function() {
      if (is.null(private$.provider_cache)) {
        # Run in fake interactive session and offline mode to only get cached tokens
        private$.provider_cache <-
          get_credential_provider(
            scope = self$.scope,
            tenant_id = self$.tenant_id,
            client_id = self$.client_id,
            chain = self$.chain,
            interactive = FALSE
          )
      }
      private$.provider_cache
    }
  ),
  private = list(
    .provider_cache = NULL
  )
)


#' Create Cached Token Credential Chain
#'
#' Creates the default chain of credentials to attempt for cached token retrieval.
#' The credentials are tried in order until one returns a valid cached token.
#' The default chain includes:
#' \enumerate{
#'   \item Authorization Code Credential - Cached tokens from browser-based authentication
#'   \item Device Code Credential - Cached tokens from device code flow
#'   \item Azure CLI Credential - Cached tokens from Azure CLI authentication
#' }
#'
#' @return A `credential_chain` object containing the sequence of
#'   credential providers to check for cached tokens.
#'
#' @seealso [CachedTokenCredential], [credential_chain()]
#'
#' @export
cached_token_credential_chain <- function() {
  credential_chain(
    auth_code = AuthCodeCredential,
    device_code = DeviceCodeCredential,
    az_cli = AzureCLICredential
  )
}

# WorkloadIdentityCredential ----
#' Workload Identity credential authentication
#'
#' @description
#' Authenticates using Azure Workload Identity by reading a federated token from
#' a file and exchanging it for an Azure AD access token. This is commonly used
#' in Kubernetes environments (AKS) where a service account token is mounted
#' into the pod.
#'
#' @details
#' The credential implements the OAuth 2.0 client credentials flow with a JWT
#' bearer assertion (`client_assertion`). It reads the federated identity token
#' from a file on each call to `get_token()` so that token rotation by the
#' runtime (e.g., Kubernetes) is automatically picked up.
#'
#' The following environment variables are used when parameters are not provided:
#' - `AZURE_CLIENT_ID`: Client (application) ID of the Azure AD application
#' - `AZURE_TENANT_ID`: Azure AD tenant ID
#' - `AZURE_FEDERATED_TOKEN_FILE`: Path to the file containing the federated token
#'
#' @export
#' @examples
#' \dontrun{
#' # Create credential using environment variables
#' # (requires AZURE_CLIENT_ID, AZURE_TENANT_ID, AZURE_FEDERATED_TOKEN_FILE)
#' cred <- WorkloadIdentityCredential$new(
#'   scope = "https://management.azure.com/.default"
#' )
#'
#' # Or supply parameters directly
#' cred <- WorkloadIdentityCredential$new(
#'   tenant_id = "your-tenant-id",
#'   client_id = "your-client-id",
#'   token_file_path = "/var/run/secrets/azure/tokens/azure-identity-token",
#'   scope = "https://management.azure.com/.default"
#' )
#'
#' # Get an access token
#' token <- cred$get_token()
#'
#' # Use with httr2 request
#' req <- httr2::request("https://management.azure.com/subscriptions")
#' resp <- httr2::req_perform(cred$req_auth(req))
#' }
WorkloadIdentityCredential <- R6::R6Class(
  classname = "WorkloadIdentityCredential",
  inherit = Credential,
  public = list(
    #' @field .token_file_path Path to the file containing the federated identity token
    .token_file_path = NULL,

    #' @description
    #' Create a new Workload Identity credential
    #'
    #' @param scope A character string specifying the OAuth2 scope. Defaults to
    #'   the Azure Resource Manager scope.
    #' @param tenant_id A character string specifying the Azure AD tenant ID.
    #'   Defaults to the `AZURE_TENANT_ID` environment variable.
    #' @param client_id A character string specifying the client (application) ID.
    #'   Defaults to the `AZURE_CLIENT_ID` environment variable.
    #' @param token_file_path A character string specifying the path to the file
    #'   containing the federated identity token. Defaults to the
    #'   `AZURE_FEDERATED_TOKEN_FILE` environment variable.
    #'
    #' @return A new `WorkloadIdentityCredential` object
    initialize = function(
      scope = NULL,
      tenant_id = Sys.getenv(
        environment_variables$azure_tenant_id,
        unset = NA_character_
      ),
      client_id = Sys.getenv(
        environment_variables$azure_client_id,
        unset = NA_character_
      ),
      token_file_path = default_federated_token_file()
    ) {
      tenant_id <- wi_required_value(
        tenant_id,
        envvar = environment_variables$azure_tenant_id,
        arg = "tenant_id"
      )
      client_id <- wi_required_value(
        client_id,
        envvar = environment_variables$azure_client_id,
        arg = "client_id"
      )
      self$.token_file_path <- token_file_path
      super$initialize(
        scope = scope,
        tenant_id = tenant_id,
        client_id = client_id
      )
      lockBinding(".token_file_path", self)
    },

    #' @description
    #' Validate the credential configuration
    #'
    #' @details
    #' Checks that `token_file_path` is provided and not NA. Calls the parent
    #' class validation method.
    validate = function() {
      super$validate()

      if (
        is.null(self$.token_file_path) || rlang::is_na(self$.token_file_path)
      ) {
        cli::cli_abort(
          c(
            "Argument {.arg token_file_path} cannot be NULL or NA.",
            "i" = "Set the {.envvar AZURE_FEDERATED_TOKEN_FILE} environment variable or pass 
            {.arg token_file_path} directly."
          ),
          class = "azr_workload_identity_missing_token_file"
        )
      }
    },

    #' @description
    #' Get an access token by exchanging the federated token
    #'
    #' @details
    #' Returns a valid in-object cached token immediately if one exists. Otherwise
    #' reads the federated token from the file and exchanges it for a new access
    #' token so that token rotation performed by the runtime is automatically
    #' reflected.
    #'
    #' @return An [httr2::oauth_token()] object containing the access token
    get_token = function() {
      cache_key <- rlang::hash(self$.scope)

      cached <- private$.token_cache[[cache_key]]
      if (private$token_is_valid(cached)) {
        return(cached)
      }

      federated_token <- wi_read_token_file(self$.token_file_path)
      token <- wi_exchange_token(
        federated_token = federated_token,
        client_id = self$.client_id,
        scope = self$.scope_str,
        token_url = self$.token_url
      )
      private$.token_cache[[cache_key]] <- token
      token
    },

    #' @description
    #' Add authentication to an httr2 request
    #'
    #' @param req An [httr2::request()] object
    #'
    #' @return The request object with a Bearer token authorization header
    req_auth = function(req) {
      token <- self$get_token()
      httr2::req_auth_bearer_token(req, token$access_token)
    }
  ),
  # private ----
  private = list(
    .token_cache = list(),
    token_is_valid = function(token) {
      if (is.null(token) || !inherits(token, "httr2_token")) {
        return(FALSE)
      }
      !token_has_expired(token)
    }
  )
)


wi_required_value <- function(value, envvar, arg) {
  if (is_empty(value)) {
    cli::cli_abort(
      c(
        "Argument {.arg {arg}} cannot be NULL or NA.",
        "i" = "Set the {.envvar {envvar}} environment variable or pass {.arg {arg}} directly."
      ),
      class = "azr_workload_identity_missing_required_value"
    )
  }

  value
}


# Read the federated identity token from a file, trimming whitespace.
wi_read_token_file <- function(path) {
  if (!file.exists(path)) {
    cli::cli_abort(
      c(
        "Federated token file not found: {.path {path}}",
        "i" = "Ensure {.envvar AZURE_FEDERATED_TOKEN_FILE} points to a valid file."
      ),
      class = "azr_workload_identity_file_not_found"
    )
  }

  token <- trimws(paste(readLines(path, warn = FALSE), collapse = ""))

  if (!nzchar(token)) {
    cli::cli_abort(
      c(
        "Federated token file is empty: {.path {path}}",
        "i" = "The file must contain a valid JWT token."
      ),
      class = "azr_workload_identity_empty_token"
    )
  }

  token
}


# Exchange a federated token for an Azure AD access token using the
# OAuth 2.0 client credentials grant with a JWT bearer assertion.
wi_exchange_token <- function(federated_token, client_id, scope, token_url) {
  resp <- rlang::try_fetch(
    httr2::request(token_url) |>
      httr2::req_body_form(
        grant_type = "client_credentials",
        client_id = client_id,
        client_assertion_type = "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
        client_assertion = federated_token,
        scope = scope
      ) |>
      httr2::req_error(is_error = \(r) FALSE) |>
      httr2::req_perform(),
    error = function(cnd) {
      cli::cli_abort(
        c(
          "Failed to reach the token endpoint.",
          "i" = "URL: {.url {token_url}}",
          "x" = conditionMessage(cnd)
        ),
        class = "azr_workload_identity_exchange_error",
        call = call("get_token")
      )
    }
  )

  body <- wi_resp_body_json(resp)

  if (!is.null(body$error)) {
    cli::cli_abort(
      c(
        "Token exchange failed: {body$error}",
        "x" = body$error_description %||% "No error description provided."
      ),
      class = "azr_workload_identity_token_error",
      call = call("get_token")
    )
  }

  httr2::oauth_token(
    access_token = body$access_token,
    token_type = body$token_type %||% "Bearer",
    expires_in = as.numeric(body$expires_in)
  )
}


wi_resp_body_json <- function(resp) {
  rlang::try_fetch(
    httr2::resp_body_json(resp),
    error = function(cnd) {
      body <- httr2::resp_body_string(resp)
      if (nchar(body) > 200) {
        body <- paste0(substr(body, 1, 200), "...")
      }

      cli::cli_abort(
        c(
          "Token endpoint returned an invalid JSON response.",
          "i" = "URL: {.url {httr2::resp_url(resp)}}",
          "i" = "Status: {httr2::resp_status(resp)}",
          "x" = conditionMessage(cnd),
          "i" = "Body: {body}"
        ),
        class = "azr_workload_identity_invalid_json_response",
        call = call("get_token")
      )
    }
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_url}
\alias{default_azure_url}
\title{Get default Azure OAuth URLs}
\usage{
default_azure_url(
  endpoint = NULL,
  oauth_host = default_azure_host(),
  tenant_id = default_azure_tenant_id()
)
}
\arguments{
\item{endpoint}{A character string specifying which endpoint URL to return.
Must be one of: \code{"authorize"}, \code{"token"}, or \code{"devicecode"}. If \code{NULL}
(default), returns a list of all endpoint URLs.}

\item{oauth_host}{A character string specifying the Azure authority host.
Defaults to \code{\link[=default_azure_host]{default_azure_host()}}.}

\item{tenant_id}{A character string specifying the tenant ID. Defaults to
\code{\link[=default_azure_tenant_id]{default_azure_tenant_id()}}.}
}
\value{
If \code{endpoint} is specified, returns a character string with the URL.
If \code{endpoint} is \code{NULL}, returns a named list of all endpoint URLs.
}
\description{
Constructs Azure OAuth 2.0 endpoint URLs for a given tenant and authority host.
}
\examples{
# Get all URLs
default_azure_url()

# Get specific endpoint
default_azure_url("token")

# Custom tenant
default_azure_url("authorize", tenant_id = "my-tenant-id")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_msal_token_cache}
\alias{default_msal_token_cache}
\title{Get default MSAL token cache path}
\usage{
default_msal_token_cache()
}
\value{
A character string with the path to the MSAL token cache file.
}
\description{
Returns the path to the MSAL token cache file shared by the Azure CLI and
Azure SDKs. Defaults to \code{msal_token_cache.json} inside the Azure config
directory (see \code{\link[=default_azure_config_dir]{default_azure_config_dir()}}).
}
\seealso{
\code{\link[=default_azure_config_dir]{default_azure_config_dir()}}, \code{\link[=write_msal_token]{write_msal_token()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_redirect_uri}
\alias{default_redirect_uri}
\title{Get default OAuth redirect URI}
\usage{
default_redirect_uri(redirect_uri = httr2::oauth_redirect_uri())
}
\arguments{
\item{redirect_uri}{A character string specifying the redirect URI. Defaults
to \code{\link[httr2:oauth_redirect_uri]{httr2::oauth_redirect_uri()}}.}
}
\value{
A character string with the redirect URI
}
\description{
Constructs a redirect URI for OAuth flows. If the provided URI doesn't have
a port, assigns a random port using \code{\link[httpuv:randomPort]{httpuv::randomPort()}}.
}
\examples{
default_redirect_uri()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_oauth_client}
\alias{default_azure_oauth_client}
\title{Create default Azure OAuth client}
\usage{
default_azure_oauth_client(
  client_id = default_azure_client_id(),
  client_secret = NULL,
  name = NULL
)
}
\arguments{
\item{client_id}{A character string specifying the client ID. Defaults to
\code{\link[=default_azure_client_id]{default_azure_client_id()}}.}

\item{client_secret}{A character string specifying the client secret. Defaults
to \code{NULL}.}

\item{name}{A character string specifying the client name. Defaults to \code{NULL}.}
}
\value{
An \code{\link[httr2:oauth_client]{httr2::oauth_client()}} object
}
\description{
Creates an \code{\link[httr2:oauth_client]{httr2::oauth_client()}} configured for Azure authentication.
}
\examples{
client <- default_azure_oauth_client()
client <- default_azure_oauth_client(
  client_id = "my-client-id",
  client_secret = "my-secret"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{environment_variables}
\alias{environment_variables}
\title{Azure Environment Variable Names}
\format{
An object of class \code{list} of length 11.
}
\usage{
environment_variables
}
\description{
Standard environment variable names used for Azure credential discovery.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-credential.R
\name{DefaultCredential}
\alias{DefaultCredential}
\title{Default credential authentication}
\description{
An R6 class that provides lazy initialization of credential providers.
The credential provider is created on first access using the default
credential chain.
}
\details{
This class wraps the credential discovery process in an R6 object with
a lazily evaluated \code{provider} field. The provider is only created when
first accessed, using the same logic as \code{\link[=get_token_provider]{get_token_provider()}}.
}
\examples{
# Create a DefaultCredential object
cred <- DefaultCredential$new(
  scope = "https://graph.microsoft.com/.default",
  tenant_id = "my-tenant-id"
)

\dontrun{
# Get a token (triggers lazy initialization)
token <- cred$get_token()

# Authenticate a request
req <- httr2::request("https://management.azure.com/subscriptions")
resp <- httr2::req_perform(cred$req_auth(req))

# Or access the provider directly
provider <- cred$provider
}

}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.scope}}{Character string specifying the authentication scope.}

\item{\code{.tenant_id}}{Character string specifying the tenant ID.}

\item{\code{.client_id}}{Character string specifying the client ID.}

\item{\code{.client_secret}}{Character string specifying the client secret.}

\item{\code{.use_cache}}{Character string indicating the caching strategy.}

\item{\code{.offline}}{Logical indicating whether to request offline access.}

\item{\code{.chain}}{A credential chain object for authentication.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{provider}}{Lazily initialized credential provider}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DefaultCredential-new}{\code{DefaultCredential$new()}}
\item \href{#method-DefaultCredential-get_token}{\code{DefaultCredential$get_token()}}
\item \href{#method-DefaultCredential-req_auth}{\code{DefaultCredential$req_auth()}}
\item \href{#method-DefaultCredential-clone}{\code{DefaultCredential$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DefaultCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-DefaultCredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new DefaultCredential object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DefaultCredential$new(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{Optional character string specifying the authentication scope.}

\item{\code{tenant_id}}{Optional character string specifying the tenant ID for
authentication.}

\item{\code{client_id}}{Optional character string specifying the client ID for
authentication.}

\item{\code{client_secret}}{Optional character string specifying the client secret
for authentication.}

\item{\code{use_cache}}{Character string indicating the caching strategy. Defaults
to \code{"disk"}. Options include \code{"disk"} for disk-based caching or \code{"memory"}
for in-memory caching.}

\item{\code{offline}}{Logical. If \code{TRUE}, adds 'offline_access' to the scope to request a 'refresh_token'.
Defaults to \code{TRUE}.}

\item{\code{chain}}{A list of credential objects, where each element must inherit
from the \code{Credential} base class. Credentials are attempted in the order
provided until \code{get_token} succeeds.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{DefaultCredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DefaultCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-DefaultCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token using the credential chain
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DefaultCredential$get_token()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DefaultCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-DefaultCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DefaultCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with authentication configured
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DefaultCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DefaultCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DefaultCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{AzureCLICredential}
\alias{AzureCLICredential}
\title{Azure CLI credential authentication}
\description{
Authenticates using the Azure CLI (\code{az}) command-line tool. This credential
requires the Azure CLI to be installed and the user to be logged in via
\verb{az login}.
}
\details{
The credential uses the \verb{az account get-access-token} command to retrieve
access tokens. It will use the currently active Azure CLI account and
subscription unless a specific tenant is specified.
}
\examples{
# Create credential with default settings
cred <- AzureCLICredential$new()

# Create credential with specific scope and tenant
cred <- AzureCLICredential$new(
  scope = "https://management.azure.com/.default",
  tenant_id = "your-tenant-id"
)

# To get a token or authenticate a request it is required that
# 'az login' is successfully executed, otherwise it will return an error.
\dontrun{
# Get an access token
token <- cred$get_token()

# Use with httr2 request
req <- httr2::request("https://management.azure.com/subscriptions")
resp <- httr2::req_perform(cred$req_auth(req))
}
}
\section{Super class}{
\code{azr::Credential} -> \code{AzureCLICredential}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{interactive}}{Logical indicating whether to check login status and
perform login if needed}

\item{\code{.process_timeout}}{Timeout in seconds for Azure CLI command execution}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-AzureCLICredential-new}{\code{AzureCLICredential$new()}}
\item \href{#method-AzureCLICredential-get_token}{\code{AzureCLICredential$get_token()}}
\item \href{#method-AzureCLICredential-req_auth}{\code{AzureCLICredential$req_auth()}}
\item \href{#method-AzureCLICredential-account_show}{\code{AzureCLICredential$account_show()}}
\item \href{#method-AzureCLICredential-login}{\code{AzureCLICredential$login()}}
\item \href{#method-AzureCLICredential-is_interactive}{\code{AzureCLICredential$is_interactive()}}
\item \href{#method-AzureCLICredential-logout}{\code{AzureCLICredential$logout()}}
\item \href{#method-AzureCLICredential-clone}{\code{AzureCLICredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="validate"><a href='../../azr/html/Credential.html#method-Credential-validate'><code>azr::Credential$validate()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new Azure CLI credential
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$new(
  scope = NULL,
  tenant_id = NULL,
  process_timeout = NULL,
  interactive = FALSE,
  use_bridge = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to
\code{NULL}, which uses the scope set during initialization.}

\item{\code{tenant_id}}{A character string specifying the Azure Active Directory
tenant ID. Defaults to \code{NULL}, which uses the default tenant from Azure CLI.}

\item{\code{process_timeout}}{A numeric value specifying the timeout in seconds
for the Azure CLI process. Defaults to \code{10}.}

\item{\code{interactive}}{A logical value indicating whether to check if the user is
logged in and perform login if needed. Defaults to \code{FALSE}.}

\item{\code{use_bridge}}{A logical value indicating whether to use the device code
bridge webpage during login. If \code{TRUE}, launches an intermediate local webpage
that displays the device code and facilitates copy-pasting before redirecting
to the Microsoft device login page. Only used when \code{interactive = TRUE}. Defaults to \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{AzureCLICredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token from Azure CLI
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$get_token(scope = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope. If \code{NULL},
uses the scope specified during initialization.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$req_auth(req, scope = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}

\item{\code{scope}}{A character string specifying the OAuth2 scope. If \code{NULL},
uses the scope specified during initialization.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with authentication header added
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-account_show"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-account_show}{}}}
\subsection{Method \code{account_show()}}{
Show the currently active Azure CLI account information
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$account_show(timeout = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{timeout}}{A numeric value specifying the timeout in seconds for the
Azure CLI command. If \code{NULL}, uses the process timeout specified during
initialization.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list containing the account information from Azure CLI
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-login"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-login}{}}}
\subsection{Method \code{login()}}{
Perform Azure CLI login using device code flow
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$login()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Invisibly returns the exit status (0 for success, non-zero for failure)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-is_interactive"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-is_interactive}{}}}
\subsection{Method \code{is_interactive()}}{
Check if the credential requires user interaction
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$is_interactive()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical indicating whether this credential is interactive
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-logout"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-logout}{}}}
\subsection{Method \code{logout()}}{
Log out from Azure CLI
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$logout()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Invisibly returns \code{NULL}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AzureCLICredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-AzureCLICredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AzureCLICredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_login}
\alias{az_cli_login}
\title{Azure CLI Device Code Login}
\usage{
az_cli_login(tenant_id = NULL, use_bridge = FALSE, verbose = FALSE)
}
\arguments{
\item{tenant_id}{A character string specifying the Azure Active Directory
tenant ID to authenticate against. If \code{NULL} (default), uses the default
tenant from Azure CLI configuration.}

\item{use_bridge}{A logical value indicating whether to use the device code
bridge webpage. If \code{TRUE}, launches an intermediate local webpage that
displays the device code and facilitates copy-pasting before redirecting
to the Microsoft device login page. If \code{FALSE} (default), copies the code
directly to the clipboard and opens the Microsoft login page.}

\item{verbose}{A logical value indicating whether to print detailed process
output to the console, including error messages from the Azure CLI process.
If \code{FALSE} (default), only essential messages are displayed.}
}
\value{
Invisibly returns the exit status (0 for success, non-zero for failure)
}
\description{
Performs an interactive Azure CLI login using device code flow.
Automatically captures the device code, copies it to the clipboard,
and opens the browser for authentication.
}
\details{
This function runs \verb{az login --use-device-code}, monitors the output
to extract the device code, copies it to the clipboard, and opens
the authentication URL in the default browser.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{azure_authority_hosts}
\alias{azure_authority_hosts}
\title{Azure Authority Host URLs}
\format{
An object of class \code{list} of length 3.
}
\usage{
azure_authority_hosts
}
\description{
Login endpoint URLs for different Azure cloud environments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_logout}
\alias{az_cli_logout}
\title{Azure CLI Logout}
\usage{
az_cli_logout()
}
\value{
Invisibly returns \code{NULL}
}
\description{
Logs out from Azure CLI by removing all stored credentials and account
information. This function runs \verb{az logout}.
}
\details{
After logging out, you will need to run \code{\link[=az_cli_login]{az_cli_login()}} again to
authenticate and use Azure CLI credentials.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_client_id}
\alias{default_azure_client_id}
\title{Get default Azure client ID}
\usage{
default_azure_client_id()
}
\value{
A character string with the client ID
}
\description{
Retrieves the Azure client ID in priority order:
\enumerate{
\item \code{AZURE_CLIENT_ID} environment variable
\item Built-in fallback (Microsoft's public Azure CLI client ID)
}
}
\examples{
default_azure_client_id()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_client_secret}
\alias{default_azure_client_secret}
\title{Get default Azure client secret}
\usage{
default_azure_client_secret()
}
\value{
A character string with the client secret, or \code{NA_character_} if not set
}
\description{
Retrieves the Azure client secret from the \code{AZURE_CLIENT_SECRET} environment
variable, or returns \code{NA_character_} if not set.
}
\examples{
default_azure_client_secret()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api-resource.R
\name{api_resource}
\alias{api_resource}
\title{Azure API Resource}
\description{
An R6 class that wraps an \code{api_client} and adds an additional path segment
(like "beta" or "v1.0") to all requests. This is useful for APIs that version
their endpoints or have different API surfaces under different paths.
}
\details{
The \code{api_resource} class creates a modified base request by appending an
endpoint path to the client's base request. All subsequent API calls through
this resource will automatically include this path prefix.
}
\examples{
\dontrun{
# Create a client
client <- api_client$new(
  host_url = "https://graph.microsoft.com"
)

# Create a resource with v1.0 API endpoint
resource_v1 <- api_resource$new(
  client = client,
  endpoint = "v1.0"
)

# Create a resource with beta API endpoint
resource_beta <- api_resource$new(
  client = client,
  endpoint = "beta"
)

# Make requests - the endpoint is automatically prepended
response <- resource_v1$.fetch(
  path = "/me",
  method = "get"
)
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.client}}{The cloned api_client instance with modified base_req}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-api_resource-new}{\code{api_resource$new()}}
\item \href{#method-api_resource-clone}{\code{api_resource$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_resource-new"></a>}}
\if{latex}{\out{\hypertarget{method-api_resource-new}{}}}
\subsection{Method \code{new()}}{
Create a new API resource instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_resource$new(client, endpoint)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{client}}{An \code{api_client} object that provides the base HTTP client
functionality. This will be cloned to avoid modifying the original.}

\item{\code{endpoint}}{A character string specifying the API endpoint or path
segment to append (e.g., \code{"v1.0"}, \code{"beta"}).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{api_resource} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_resource-clone"></a>}}
\if{latex}{\out{\hypertarget{method-api_resource-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_resource$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_federated_token_file}
\alias{default_federated_token_file}
\title{Get default federated token file path}
\usage{
default_federated_token_file()
}
\value{
A character string with the file path, or \code{NULL} if not set
}
\description{
Retrieves the path to the federated identity token file from the
\code{AZURE_FEDERATED_TOKEN_FILE} environment variable, or returns \code{NULL} if
not set. Used by \link{WorkloadIdentityCredential}.
}
\examples{
default_federated_token_file()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-interactive.R
\name{InteractiveCredential}
\alias{InteractiveCredential}
\title{Interactive credential base class}
\description{
Base class for interactive authentication credentials. This class should not
be instantiated directly; use \link{DeviceCodeCredential} or \link{AuthCodeCredential}
instead.
}
\keyword{internal}
\section{Super class}{
\code{azr::Credential} -> \code{InteractiveCredential}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{use_refresh_token}}{Logical indicating whether to use the login flow (acquire
tokens via refresh token exchange).}

\item{\code{interactive}}{Logical indicating whether this credential requires
user interaction.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-InteractiveCredential-new}{\code{InteractiveCredential$new()}}
\item \href{#method-InteractiveCredential-is_interactive}{\code{InteractiveCredential$is_interactive()}}
\item \href{#method-InteractiveCredential-get_token}{\code{InteractiveCredential$get_token()}}
\item \href{#method-InteractiveCredential-req_auth}{\code{InteractiveCredential$req_auth()}}
\item \href{#method-InteractiveCredential-clone}{\code{InteractiveCredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="validate"><a href='../../azr/html/Credential.html#method-Credential-validate'><code>azr::Credential$validate()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-InteractiveCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-InteractiveCredential-new}{}}}
\subsection{Method \code{new()}}{
Shared initializer for interactive credentials
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{InteractiveCredential$new(
  scope = NULL,
  tenant_id = NULL,
  client_id = default_azure_cli_client_id(),
  use_cache = "disk",
  offline = TRUE,
  interactive = TRUE,
  use_refresh_token = TRUE,
  flow_fun,
  req_auth_fun,
  oauth_endpoint,
  name,
  extra_flow_params = list()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope.}

\item{\code{tenant_id}}{Azure AD tenant ID.}

\item{\code{client_id}}{Application (client) ID. Defaults to the Azure CLI
public client ID.}

\item{\code{use_cache}}{Cache type: \code{"disk"} or \code{"memory"}.}

\item{\code{offline}}{Whether to request offline access (refresh tokens).}

\item{\code{interactive}}{Whether this credential requires user interaction.}

\item{\code{use_refresh_token}}{Whether to use the login flow (acquire tokens via refresh token
exchange). Set to \code{FALSE} to use the access token flow directly.}

\item{\code{flow_fun}}{The httr2 OAuth flow function (e.g. \link[httr2:req_oauth_device]{httr2::oauth_flow_device}).}

\item{\code{req_auth_fun}}{The httr2 request auth function (e.g. \link[httr2:req_oauth_device]{httr2::req_oauth_device}).}

\item{\code{oauth_endpoint}}{The OAuth endpoint name passed to the parent credential.}

\item{\code{name}}{The credential name passed to the parent credential.}

\item{\code{extra_flow_params}}{A named list of additional parameters merged into
\code{private$.flow_params} after \code{scope} and \code{auth_url}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-InteractiveCredential-is_interactive"></a>}}
\if{latex}{\out{\hypertarget{method-InteractiveCredential-is_interactive}{}}}
\subsection{Method \code{is_interactive()}}{
Check if the credential requires user interaction
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{InteractiveCredential$is_interactive()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical indicating whether this credential is interactive
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-InteractiveCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-InteractiveCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token using the flow configured by the subclass.
Returns a valid in-object cached token immediately if one exists for the
requested scope. Otherwise attempts token acquisition in three steps:
(1) return a valid httr2-cached token without any interaction; (2)
silently refresh using an existing refresh token; (3) fall back to the
configured interactive flow. When \code{reauth = TRUE} all caches are bypassed
and the interactive flow is used directly.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{InteractiveCredential$get_token(scope = NULL, reauth = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to
\code{NULL}, which uses the scope configured on the credential.}

\item{\code{reauth}}{A logical value indicating whether to force
reauthentication, bypassing the cache and silent refresh. Defaults to
\code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-InteractiveCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-InteractiveCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add OAuth authentication to an httr2 request using the flow configured
by the subclass
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{InteractiveCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with OAuth authentication configured
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-InteractiveCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-InteractiveCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{InteractiveCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_get_cached_token}
\alias{az_cli_get_cached_token}
\title{Get Cached Token from MSAL Token Cache}
\usage{
az_cli_get_cached_token(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  config_dir = default_azure_config_dir()
)
}
\arguments{
\item{scope}{A character string specifying the OAuth2 scope to filter tokens.
If \code{NULL} (default), returns the latest-expiring token regardless of scope.}

\item{tenant_id}{A character string specifying the tenant ID to filter tokens.
If \code{NULL} (default), matches any tenant.}

\item{client_id}{A character string specifying the client ID to filter tokens.
If \code{NULL} (default), matches any client.}

\item{config_dir}{A character string specifying the Azure configuration
directory. Defaults to \code{\link[=default_azure_config_dir]{default_azure_config_dir()}}.}
}
\value{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing:
\itemize{
\item \code{access_token}: The OAuth2 access token string
\item \code{token_type}: The type of token (typically "Bearer")
\item \code{.expires_at}: POSIXct timestamp when the token expires
}
}
\description{
Reads the MSAL token cache file (\code{msal_token_cache.json}) from the Azure
configuration directory and returns a matching access token as an
\code{\link[httr2:oauth_token]{httr2::oauth_token()}} object.
}
\details{
The MSAL token cache is a JSON file maintained by the Azure CLI that stores
access tokens and refresh tokens. This function reads cached access tokens
directly from the file without invoking the Azure CLI, which can be useful
in environments where the CLI is slow or unavailable but tokens have been
previously cached.

When multiple tokens are found, the function selects the token that expires
latest. If \code{scope} is provided, only tokens matching that scope/resource are
returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api-client.R
\name{default_non_auth}
\alias{default_non_auth}
\title{Default No Authentication}
\usage{
default_non_auth(req)
}
\arguments{
\item{req}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\value{
The same \code{\link[httr2:request]{httr2::request()}} object, unmodified
}
\description{
A pass-through credential function that performs no authentication.
This function returns the request object unchanged, allowing API calls
to be made without adding any authentication headers or tokens.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_storage_endpoint}
\alias{default_storage_endpoint}
\title{Get default Azure Storage DFS endpoint suffix}
\usage{
default_storage_endpoint()
}
\value{
A character string with the DFS endpoint suffix.
}
\description{
Returns the default endpoint suffix used to construct Azure Data Lake Storage
Gen2 DFS URLs.
}
\examples{
default_storage_endpoint()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_account_show}
\alias{az_cli_account_show}
\title{Show Azure CLI Account Information}
\usage{
az_cli_account_show(timeout = 10L)
}
\arguments{
\item{timeout}{An integer specifying the timeout in seconds for the Azure
CLI command. Defaults to \code{10}.}
}
\value{
A list containing the account information from Azure CLI
}
\description{
Retrieves information about the currently active Azure CLI account and
subscription. This function runs \verb{az account show} and parses the JSON
output into an R list.
}
\details{
The function returns details about the current Azure subscription including:
\itemize{
\item Subscription ID and name
\item Tenant ID
\item Account state (e.g., "Enabled")
\item User information
\item Cloud environment details
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-graph-client.R
\name{api_graph_resource}
\alias{api_graph_resource}
\title{Microsoft Graph API Resource}
\description{
An R6 class that extends \link{api_resource} to provide specialized methods
for the Microsoft Graph API. This class adds convenience methods for
common Graph operations.
}
\keyword{internal}
\section{Super class}{
\code{\link[azr:api_resource]{azr::api_resource}} -> \code{api_graph_resource}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-api_graph_resource-me}{\code{api_graph_resource$me()}}
\item \href{#method-api_graph_resource-clone}{\code{api_graph_resource$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_resource" data-id="initialize"><a href='../../azr/html/api_resource.html#method-api_resource-initialize'><code>azr::api_resource$initialize()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_graph_resource-me"></a>}}
\if{latex}{\out{\hypertarget{method-api_graph_resource-me}{}}}
\subsection{Method \code{me()}}{
Fetch the current user's profile
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_graph_resource$me(select = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{select}}{A character vector of properties to select (e.g., c("displayName", "mail")).
If NULL, all properties are returned.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The response from the /me endpoint
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_graph_resource-clone"></a>}}
\if{latex}{\out{\hypertarget{method-api_graph_resource-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_graph_resource$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-graph-client.R
\name{azr_graph_client}
\alias{azr_graph_client}
\title{Create a Microsoft Graph API Client}
\usage{
azr_graph_client(scopes = ".default", ..., chain = NULL)
}
\arguments{
\item{scopes}{A character string specifying the OAuth2 scope suffix to be appended
to the Graph API URL. Defaults to \code{".default"}, which requests all permissions
the app has been granted. The full scope will be \verb{https://graph.microsoft.com/\{scopes\}}.}

\item{...}{Additional arguments passed to the \link{api_client} constructor.}

\item{chain}{A \link{credential_chain} instance for authentication. If NULL,
a default credential chain will be created using \link{DefaultCredential}.}
}
\value{
An \link{api_service} object configured for Microsoft Graph API with
v1.0 and beta endpoints. The object is locked using \code{\link[=lockEnvironment]{lockEnvironment()}} to
prevent modification after creation. Access endpoints via \verb{$v1.0} or \verb{$beta}.
}
\description{
Creates a configured client for the Microsoft Graph API with authentication
and versioned endpoints (v1.0 and beta). This function returns an
\link{api_service} object that provides access to Microsoft Graph resources
through versioned endpoints.
}
\details{
The function creates a Microsoft Graph service using these components:
\itemize{
\item \strong{\link{api_client}}: A general-purpose API client configured with the Graph API
host URL (\verb{https://graph.microsoft.com}) and authentication provider.
\item \strong{\link{api_graph_resource}}: A specialized resource class that extends \link{api_resource}
with Microsoft Graph-specific methods. Currently implements:
\itemize{
\item \code{me(select = NULL)}: Fetch the current user's profile. The \code{select} parameter
accepts a character vector of properties to return (e.g., \code{c("displayName", "mail")}).
}
\item \strong{\link{api_service}}: A service container that combines the client and resources
with versioned endpoints (v1.0 and beta). The service is locked using
\code{\link[=lockEnvironment]{lockEnvironment()}} to prevent modification after creation.
}
}
\examples{
\dontrun{
# Create a Graph API client with default credentials
graph <- azr_graph_client()

# Fetch current user profile from v1.0 endpoint
me <- graph$v1.0$me()

# Fetch specific properties using OData $select
me <- graph$v1.0$me(select = c("displayName", "mail", "userPrincipalName"))

# Use beta endpoint for preview features
me_beta <- graph$beta$me(select = c("displayName", "mail"))

# Create with a custom credential chain
custom_chain <- credential_chain(
  AzureCLICredential$new(scope = "https://graph.microsoft.com/.default")
)
graph <- azr_graph_client(chain = custom_chain)

# Use specific scopes instead of .default
graph <- azr_graph_client(scopes = "User.Read Mail.Read")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-workload-identity.R
\name{WorkloadIdentityCredential}
\alias{WorkloadIdentityCredential}
\title{Workload Identity credential authentication}
\description{
Authenticates using Azure Workload Identity by reading a federated token from
a file and exchanging it for an Azure AD access token. This is commonly used
in Kubernetes environments (AKS) where a service account token is mounted
into the pod.
}
\details{
The credential implements the OAuth 2.0 client credentials flow with a JWT
bearer assertion (\code{client_assertion}). It reads the federated identity token
from a file on each call to \code{get_token()} so that token rotation by the
runtime (e.g., Kubernetes) is automatically picked up.

The following environment variables are used when parameters are not provided:
\itemize{
\item \code{AZURE_CLIENT_ID}: Client (application) ID of the Azure AD application
\item \code{AZURE_TENANT_ID}: Azure AD tenant ID
\item \code{AZURE_FEDERATED_TOKEN_FILE}: Path to the file containing the federated token
}
}
\examples{
\dontrun{
# Create credential using environment variables
# (requires AZURE_CLIENT_ID, AZURE_TENANT_ID, AZURE_FEDERATED_TOKEN_FILE)
cred <- WorkloadIdentityCredential$new(
  scope = "https://management.azure.com/.default"
)

# Or supply parameters directly
cred <- WorkloadIdentityCredential$new(
  tenant_id = "your-tenant-id",
  client_id = "your-client-id",
  token_file_path = "/var/run/secrets/azure/tokens/azure-identity-token",
  scope = "https://management.azure.com/.default"
)

# Get an access token
token <- cred$get_token()

# Use with httr2 request
req <- httr2::request("https://management.azure.com/subscriptions")
resp <- httr2::req_perform(cred$req_auth(req))
}
}
\section{Super class}{
\code{azr::Credential} -> \code{WorkloadIdentityCredential}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.token_file_path}}{Path to the file containing the federated identity token}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-WorkloadIdentityCredential-new}{\code{WorkloadIdentityCredential$new()}}
\item \href{#method-WorkloadIdentityCredential-validate}{\code{WorkloadIdentityCredential$validate()}}
\item \href{#method-WorkloadIdentityCredential-get_token}{\code{WorkloadIdentityCredential$get_token()}}
\item \href{#method-WorkloadIdentityCredential-req_auth}{\code{WorkloadIdentityCredential$req_auth()}}
\item \href{#method-WorkloadIdentityCredential-clone}{\code{WorkloadIdentityCredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="is_interactive"><a href='../../azr/html/Credential.html#method-Credential-is_interactive'><code>azr::Credential$is_interactive()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkloadIdentityCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-WorkloadIdentityCredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new Workload Identity credential
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkloadIdentityCredential$new(
  scope = NULL,
  tenant_id = Sys.getenv(environment_variables$azure_tenant_id, unset = NA_character_),
  client_id = Sys.getenv(environment_variables$azure_client_id, unset = NA_character_),
  token_file_path = default_federated_token_file()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to
the Azure Resource Manager scope.}

\item{\code{tenant_id}}{A character string specifying the Azure AD tenant ID.
Defaults to the \code{AZURE_TENANT_ID} environment variable.}

\item{\code{client_id}}{A character string specifying the client (application) ID.
Defaults to the \code{AZURE_CLIENT_ID} environment variable.}

\item{\code{token_file_path}}{A character string specifying the path to the file
containing the federated identity token. Defaults to the
\code{AZURE_FEDERATED_TOKEN_FILE} environment variable.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{WorkloadIdentityCredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkloadIdentityCredential-validate"></a>}}
\if{latex}{\out{\hypertarget{method-WorkloadIdentityCredential-validate}{}}}
\subsection{Method \code{validate()}}{
Validate the credential configuration
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkloadIdentityCredential$validate()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Checks that \code{token_file_path} is provided and not NA. Calls the parent
class validation method.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkloadIdentityCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-WorkloadIdentityCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token by exchanging the federated token
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkloadIdentityCredential$get_token()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Returns a valid in-object cached token immediately if one exists. Otherwise
reads the federated token from the file and exchanges it for a new access
token so that token rotation performed by the runtime is automatically
reflected.
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkloadIdentityCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-WorkloadIdentityCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkloadIdentityCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with a Bearer token authorization header
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WorkloadIdentityCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-WorkloadIdentityCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkloadIdentityCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api-client.R
\name{default_response_handler}
\alias{default_response_handler}
\title{Default response handler}
\usage{
default_response_handler(content)
}
\arguments{
\item{content}{Parsed response content from an API call.}
}
\value{
The processed content, with any \code{data.frame} objects converted to
\code{data.table} if the \code{data.table} package is installed.
}
\description{
Converts \code{data.frame} results in the parsed response to \code{data.table} objects
when the \code{data.table} package is available. Applied automatically by
\link{api_client} unless overridden via the \code{response_handler} argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_refresh_token}
\alias{default_refresh_token}
\title{Get default Azure refresh token}
\usage{
default_refresh_token()
}
\value{
A character string with the refresh token, or \code{NULL} if not set
}
\description{
Retrieves the Azure refresh token from the \code{AZURE_REFRESH_TOKEN} environment
variable, or returns \code{NULL} if not set.
}
\examples{
default_refresh_token()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api-services.R
\name{api_service}
\alias{api_service}
\title{API Service Base Class}
\description{
Base R6 class for creating API service wrappers. This class provides a
foundation for building service-specific API clients with authentication,
endpoint management, and configuration.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.client}}{An \link{api_client} instance for making API requests}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-api_service-new}{\code{api_service$new()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_service-new"></a>}}
\if{latex}{\out{\hypertarget{method-api_service-new}{}}}
\subsection{Method \code{new()}}{
Create a new API service instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_service$new(
  client = NULL,
  chain = NULL,
  endpoints = list(),
  config = list()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{client}}{An \link{api_client} instance. If \code{NULL}, a new client will be created.}

\item{\code{chain}}{A \link{credential_chain} instance for authentication. Optional.}

\item{\code{endpoints}}{A named list where names are endpoint paths (e.g., "v1.0", "beta")
and values are R6 class objects (not instances) to use for creating resources.
Defaults to an empty list. If the value is \code{NULL}, \link{api_resource} will be used.}

\item{\code{config}}{A list of configuration options. Defaults to an empty list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{api_service} object
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-cached-token.R
\name{cached_token_credential_chain}
\alias{cached_token_credential_chain}
\title{Create Cached Token Credential Chain}
\usage{
cached_token_credential_chain()
}
\value{
A \code{credential_chain} object containing the sequence of
credential providers to check for cached tokens.
}
\description{
Creates the default chain of credentials to attempt for cached token retrieval.
The credentials are tried in order until one returns a valid cached token.
The default chain includes:
\enumerate{
\item Authorization Code Credential - Cached tokens from browser-based authentication
\item Device Code Credential - Cached tokens from device code flow
\item Azure CLI Credential - Cached tokens from Azure CLI authentication
}
}
\seealso{
\link{CachedTokenCredential}, \code{\link[=credential_chain]{credential_chain()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{write_msal_token}
\alias{write_msal_token}
\title{Write an httr2 Token to the MSAL Token Cache}
\usage{
write_msal_token(token, cache_file = default_msal_token_cache())
}
\arguments{
\item{token}{An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object. Must contain \code{access_token},
\code{token_type}, and \code{.expires_at}. May optionally contain \code{refresh_token}
and \code{scope}. All cache fields (\code{home_account_id}, \code{tenant_id}, \code{username},
\code{client_id}, \code{scope}, \code{environment}) are derived from the JWT claims
(\code{oid}, \code{tid}, \code{upn}/\code{preferred_username}, \code{appid}/\code{azp}, \code{scp}/\code{scope},
\code{iss}) and the token object itself.}

\item{cache_file}{Path to the MSAL token cache JSON file. Defaults to
\code{\link[=default_msal_token_cache]{default_msal_token_cache()}}.}
}
\value{
Invisibly returns the path to the cache file.
}
\description{
Writes an \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object into the MSAL token cache JSON file
(\code{msal_token_cache.json}) shared by the Azure SDK and Azure CLI. The
resulting entry is readable by other Azure tools (Python SDK, Azure CLI,
and the rest of this package via \code{\link[=az_cli_get_cached_token]{az_cli_get_cached_token()}}).
}
\details{
The function adds or overwrites \code{AccessToken}, \code{RefreshToken} (when the
token carries a refresh token), \code{Account}, and \code{AppMetadata} sections.
Existing entries for other accounts or clients are preserved.

The \code{home_account_id} follows the MSAL convention
\code{"<object_id>.<tenant_id>"} where \code{object_id} is the Azure AD OID of the
authenticated principal. Cache entry keys are built in the same format used
by the Azure CLI and MSAL Python:
\itemize{
\item AccessToken: \verb{<home_account_id>-<environment>-accesstoken-<client_id>-<realm>-<target>}
\item RefreshToken: \verb{<home_account_id>-<environment>-refreshtoken-<client_id>--}
\item Account: \verb{<home_account_id>-<environment>-<realm>}
\item AppMetadata: \verb{appmetadata-<environment>-<client_id>}
}
}
\seealso{
\code{\link[=az_cli_get_cached_token]{az_cli_get_cached_token()}}, \code{\link[httr2:oauth_token]{httr2::oauth_token()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-credential.R
\name{credential_chain}
\alias{credential_chain}
\title{Create Custom Credential Chain}
\usage{
credential_chain(...)
}
\arguments{
\item{...}{Named credential objects or credential classes. Each element
should be a credential class (e.g., \code{ClientSecretCredential}) or an
instantiated credential object that inherits from the \code{Credential}
base class. The names are used for identification purposes.}
}
\value{
A \code{credential_chain} object containing the specified sequence
of credential providers.
}
\description{
Creates a custom chain of credential providers to attempt during
authentication. Credentials are tried in the order they are provided
until one successfully authenticates. This allows you to customize
the authentication flow beyond the default credential chain.
}
\examples{
# Create a custom chain with only non-interactive credentials
custom_chain <- credential_chain(
  client_secret = ClientSecretCredential,
  azure_cli = AzureCLICredential
)

# Use the custom chain to get a token
\dontrun{
token <- get_token(
  scope = "https://graph.microsoft.com/.default",
  chain = custom_chain
)
}

}
\seealso{
\code{\link[=default_credential_chain]{default_credential_chain()}}, \code{\link[=get_token_provider]{get_token_provider()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-client-secret.R
\name{ClientSecretCredential}
\alias{ClientSecretCredential}
\title{Client secret credential authentication}
\description{
Authenticates a service principal using a client ID and client secret.
This credential is commonly used for application authentication in Azure.
}
\details{
The credential uses the OAuth 2.0 client credentials flow to obtain access
tokens. It requires a registered Azure AD application with a client secret.
The client secret should be stored securely and not hard-coded in scripts.
}
\examples{
# Create credential with client secret
cred <- ClientSecretCredential$new(
  tenant_id = "your-tenant-id",
  client_id = "your-client-id",
  client_secret = "your-client-secret",
  scope = "https://management.azure.com/.default"
)

# To get a token or authenticate a request it requires
# valid 'client_id' and 'client_secret' credentials,
# otherwise it will return an error.
\dontrun{
# Get an access token
token <- cred$get_token()

# Use with httr2 request
req <- httr2::request("https://management.azure.com/subscriptions")
resp <- httr2::req_perform(cred$req_auth(req))
}
}
\section{Super class}{
\code{azr::Credential} -> \code{ClientSecretCredential}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ClientSecretCredential-validate}{\code{ClientSecretCredential$validate()}}
\item \href{#method-ClientSecretCredential-get_token}{\code{ClientSecretCredential$get_token()}}
\item \href{#method-ClientSecretCredential-req_auth}{\code{ClientSecretCredential$req_auth()}}
\item \href{#method-ClientSecretCredential-clone}{\code{ClientSecretCredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="initialize"><a href='../../azr/html/Credential.html#method-Credential-initialize'><code>azr::Credential$initialize()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="is_interactive"><a href='../../azr/html/Credential.html#method-Credential-is_interactive'><code>azr::Credential$is_interactive()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ClientSecretCredential-validate"></a>}}
\if{latex}{\out{\hypertarget{method-ClientSecretCredential-validate}{}}}
\subsection{Method \code{validate()}}{
Validate the credential configuration
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ClientSecretCredential$validate()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Checks that the client secret is provided and not NA or NULL. Calls the parent
class validation method.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ClientSecretCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-ClientSecretCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token using client credentials flow
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ClientSecretCredential$get_token()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ClientSecretCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-ClientSecretCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add OAuth client credentials authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ClientSecretCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with OAuth client credentials authentication configured
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ClientSecretCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ClientSecretCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ClientSecretCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-interactive.R
\name{AuthCodeCredential}
\alias{AuthCodeCredential}
\title{Authorization code credential authentication}
\description{
Authenticates a user through the OAuth 2.0 authorization code flow. This
flow opens a web browser for the user to sign in.
}
\details{
The authorization code flow is the standard OAuth 2.0 flow for interactive
authentication. It requires a web browser and is suitable for applications
where the user can interact with a browser window.

The credential supports token caching to avoid repeated authentication.
Tokens can be cached to disk or in memory. A redirect URI is required for
the OAuth flow to complete.
}
\examples{
# AuthCodeCredential requires an interactive session
\dontrun{
# Create credential with default settings
cred <- AuthCodeCredential$new(
  tenant_id = "your-tenant-id",
  client_id = "your-client-id",
  scope = "https://management.azure.com/.default"
)

# Get an access token (will open browser for authentication)
token <- cred$get_token()

# Force reauthentication
token <- cred$get_token(reauth = TRUE)

# Use with httr2 request
req <- httr2::request("https://management.azure.com/subscriptions")
req <- cred$req_auth(req)
}
}
\section{Super classes}{
\code{azr::Credential} -> \code{\link[azr:InteractiveCredential]{azr::InteractiveCredential}} -> \code{AuthCodeCredential}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-AuthCodeCredential-new}{\code{AuthCodeCredential$new()}}
\item \href{#method-AuthCodeCredential-clone}{\code{AuthCodeCredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="validate"><a href='../../azr/html/Credential.html#method-Credential-validate'><code>azr::Credential$validate()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="InteractiveCredential" data-id="get_token"><a href='../../azr/html/InteractiveCredential.html#method-InteractiveCredential-get_token'><code>azr::InteractiveCredential$get_token()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="InteractiveCredential" data-id="is_interactive"><a href='../../azr/html/InteractiveCredential.html#method-InteractiveCredential-is_interactive'><code>azr::InteractiveCredential$is_interactive()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="InteractiveCredential" data-id="req_auth"><a href='../../azr/html/InteractiveCredential.html#method-InteractiveCredential-req_auth'><code>azr::InteractiveCredential$req_auth()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthCodeCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-AuthCodeCredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new authorization code credential
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthCodeCredential$new(
  scope = NULL,
  tenant_id = NULL,
  client_id = default_azure_cli_client_id(),
  use_cache = "disk",
  offline = TRUE,
  redirect_uri = default_redirect_uri(),
  interactive = TRUE,
  use_refresh_token = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to \code{NULL}.}

\item{\code{tenant_id}}{A character string specifying the Azure Active Directory
tenant ID. Defaults to \code{NULL}.}

\item{\code{client_id}}{A character string specifying the application (client) ID.
Defaults to the Azure CLI public client ID.}

\item{\code{use_cache}}{A character string specifying the cache type. Use \code{"disk"}
for disk-based caching or \code{"memory"} for in-memory caching. Defaults to \code{"disk"}.}

\item{\code{offline}}{A logical value indicating whether to request offline access
(refresh tokens). Defaults to \code{TRUE}.}

\item{\code{redirect_uri}}{A character string specifying the redirect URI registered
with the application. Defaults to \code{\link[=default_redirect_uri]{default_redirect_uri()}}.}

\item{\code{interactive}}{A logical value indicating whether this credential
requires user interaction. Defaults to \code{TRUE}.}

\item{\code{use_refresh_token}}{A logical value indicating whether to use the login flow
(acquire tokens via refresh token exchange). Defaults to \code{TRUE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{AuthCodeCredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-AuthCodeCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-AuthCodeCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{AuthCodeCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_config_dir}
\alias{default_azure_config_dir}
\title{Get default Azure configuration directory}
\usage{
default_azure_config_dir()
}
\value{
A character string with the Azure configuration directory path
}
\description{
Retrieves the Azure configuration directory from the \code{AZURE_CONFIG_DIR}
environment variable, or falls back to the platform-specific default.
}
\examples{
default_azure_config_dir()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_host}
\alias{default_azure_host}
\title{Get default Azure authority host}
\usage{
default_azure_host()
}
\value{
A character string with the authority host URL
}
\description{
Retrieves the Azure authority host in priority order:
\enumerate{
\item \code{AZURE_AUTHORITY_HOST} environment variable
\item Built-in fallback (\code{login.microsoftonline.com})
}
}
\examples{
default_azure_host()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-storage-client.R
\name{azr_storage_client}
\alias{azr_storage_client}
\title{Create an Azure Storage Client}
\usage{
azr_storage_client(
  storageaccount,
  filesystem,
  endpoint_suffix = default_storage_endpoint(),
  scope = default_azure_scope("azure_storage"),
  provider = NULL,
  chain = default_credential_chain(),
  tenant_id = default_azure_tenant_id(),
  ...
)
}
\arguments{
\item{storageaccount}{A character string specifying the Azure Storage account name.}

\item{filesystem}{A character string specifying the filesystem (container) name.}

\item{endpoint_suffix}{A character string specifying the Azure
Storage DFS endpoint suffix. Defaults to
\code{\link[=default_storage_endpoint]{default_storage_endpoint()}}.}

\item{scope}{A character string specifying the OAuth2 scope. Defaults to
\code{default_azure_scope("azure_storage")}.}

\item{provider}{An optional credential provider object that inherits from
\code{Credential} or \code{DefaultCredential}. If provided, \code{chain} is ignored.}

\item{chain}{A \link{credential_chain} instance for authentication. Defaults to
\code{\link[=default_credential_chain]{default_credential_chain()}}.}

\item{tenant_id}{A character string specifying the Azure tenant ID. Defaults to
\code{\link[=default_azure_tenant_id]{default_azure_tenant_id()}}, which reads \code{AZURE_TENANT_ID} from the environment.}

\item{...}{Additional arguments passed to the \link{api_storage_client} constructor.}
}
\value{
An \link{api_storage_client} object.
}
\description{
A convenience wrapper around \link{api_storage_client} that creates a configured
client for Azure Data Lake Storage Gen2 (ADLS Gen2) REST API operations.
}
\examples{
\dontrun{
# Create a storage client with default credentials
storage <- azr_storage_client(
  storageaccount = "mystorageaccount",
  filesystem = "mycontainer"
)

# Create a storage client with a specific tenant
storage <- azr_storage_client(
  storageaccount = "mystorageaccount",
  filesystem = "mycontainer",
  tenant_id = "00000000-0000-0000-0000-000000000000"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-credential.R
\name{default_credential_chain}
\alias{default_credential_chain}
\title{Create Default Credential Chain}
\usage{
default_credential_chain()
}
\value{
A \code{credential_chain} object containing the default sequence of
credential providers.
}
\description{
Creates the default chain of credentials to attempt during authentication.
The credentials are tried in order until one successfully authenticates.
The default chain includes:
\enumerate{
\item Client Secret Credential - Uses client ID and secret
\item Authorization Code Credential - Interactive browser-based authentication
\item Device Code Credential - Interactive device code flow
\item Azure CLI Credential - Uses credentials from Azure CLI
}
}
\seealso{
\code{\link[=credential_chain]{credential_chain()}}, \code{\link[=get_token_provider]{get_token_provider()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{azure_storage_endpoints}
\alias{azure_storage_endpoints}
\title{Azure Storage Endpoint Defaults}
\format{
An object of class \code{list} of length 1.
}
\usage{
azure_storage_endpoints
}
\description{
Default endpoint suffixes for Azure Storage services.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_tenant_id}
\alias{default_azure_tenant_id}
\title{Get default Azure tenant ID}
\usage{
default_azure_tenant_id()
}
\value{
A character string with the tenant ID
}
\description{
Retrieves the Azure tenant ID in priority order:
\enumerate{
\item \code{AZURE_TENANT_ID} environment variable
\item Built-in fallback (\code{"common"})
}
}
\examples{
default_azure_tenant_id()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import-funs.R
\name{is_hosted_session}
\alias{is_hosted_session}
\title{Detect if running in a hosted session}
\usage{
is_hosted_session()
}
\value{
A logical value: \code{TRUE} if running in a hosted session (Google Colab,
VS Code, Kubernetes, or remote RStudio Server), \code{FALSE} otherwise.
}
\description{
Determines whether the current R session is running in a hosted environment
such as Google Colab, VS Code, Kubernetes, or RStudio Server (non-localhost).
}
\details{
This function checks for (in order):
\itemize{
\item Option override: if \code{azr.hosted} option is set, returns \code{isTRUE()} of its value
\item Google Colab: presence of the \code{COLAB_RELEASE_TAG} environment variable
\item VS Code: presence of the \code{VSCODE_INJECTION} or \code{VSCODE_PROXY_URI} environment variable
\item Kubernetes: presence of the \code{KUBERNETES_SERVICE_HOST} environment variable
\item RStudio Server: \code{RSTUDIO_PROGRAM_MODE} is "server" and
\code{RSTUDIO_HTTP_REFERER} does not contain "localhost"
}
}
\examples{
if (is_hosted_session()) {
  message("Running in a hosted environment")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-cached-token.R
\name{CachedTokenCredential}
\alias{CachedTokenCredential}
\title{Cached token credential authentication}
\description{
A credential class that retrieves tokens from the cache only, without
triggering interactive authentication flows. This is useful for non-interactive
sessions where you want to use previously cached tokens from DeviceCode or
AuthCode credentials.
}
\details{
This credential attempts to retrieve cached tokens from a chain of interactive
credentials (AuthCode and DeviceCode by default). It will not prompt for new
authentication - it only returns tokens that are already cached.

This is particularly useful for:
\itemize{
\item Non-interactive R sessions (e.g., scheduled scripts, CI/CD)
\item Scenarios where you've previously authenticated interactively and want
to reuse those cached tokens
}
}
\examples{
\dontrun{
# Create credential with default settings
cred <- CachedTokenCredential$new(
  scope = "https://graph.microsoft.com/.default",
  tenant_id = "my-tenant-id"
)

# Get a cached token (will fail if no cached token exists)
token <- cred$get_token()

# Use with httr2 request
req <- httr2::request("https://graph.microsoft.com/v1.0/me")
req <- cred$req_auth(req)
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.scope}}{Character string specifying the authentication scope.}

\item{\code{.tenant_id}}{Character string specifying the tenant ID.}

\item{\code{.client_id}}{Character string specifying the client ID.}

\item{\code{.chain}}{List of credential classes to attempt for cached tokens.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{provider}}{Lazily initialized credential provider}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-CachedTokenCredential-new}{\code{CachedTokenCredential$new()}}
\item \href{#method-CachedTokenCredential-get_token}{\code{CachedTokenCredential$get_token()}}
\item \href{#method-CachedTokenCredential-req_auth}{\code{CachedTokenCredential$req_auth()}}
\item \href{#method-CachedTokenCredential-clone}{\code{CachedTokenCredential$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CachedTokenCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-CachedTokenCredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new CachedTokenCredential object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CachedTokenCredential$new(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  chain = cached_token_credential_chain()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{scope}}{Optional character string specifying the authentication scope.}

\item{\code{tenant_id}}{Optional character string specifying the tenant ID for
authentication.}

\item{\code{client_id}}{Optional character string specifying the client ID for
authentication.}

\item{\code{chain}}{A list of credential classes to attempt for cached tokens.
Defaults to AuthCodeCredential and DeviceCodeCredential.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{CachedTokenCredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CachedTokenCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-CachedTokenCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token from the cache
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CachedTokenCredential$get_token()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CachedTokenCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-CachedTokenCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CachedTokenCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with authentication configured
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CachedTokenCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CachedTokenCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CachedTokenCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{azure_scopes}
\alias{azure_scopes}
\title{Common Azure OAuth Scopes}
\format{
An object of class \code{list} of length 5.
}
\usage{
azure_scopes
}
\description{
Predefined OAuth scopes for common Azure services.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-refresh-token.R
\name{RefreshTokenCredential}
\alias{RefreshTokenCredential}
\title{Refresh token credential authentication}
\description{
Authenticates using an existing refresh token. This credential is useful when
you have obtained a refresh token through another authentication flow and want
to use it to get new access tokens without interactive authentication.
}
\details{
The refresh token credential uses the OAuth 2.0 refresh token flow to obtain
new access tokens. It requires a valid refresh token that was previously
obtained through an interactive flow (e.g., authorization code or device code).

This is particularly useful for:
\itemize{
\item Non-interactive sessions where you have a pre-obtained refresh token
\item Long-running applications that need to refresh tokens automatically
\item Scenarios where you want to avoid repeated interactive authentication
}
}
\examples{
\dontrun{
# Create credential with a refresh token
cred <- RefreshTokenCredential$new(
  refresh_token = "your-refresh-token",
  scope = "https://management.azure.com/.default",
  tenant_id = "your-tenant-id",
  client_id = "your-client-id"
)

# Get an access token
token <- cred$get_token()

# Use with httr2 request
req <- httr2::request("https://management.azure.com/subscriptions")
resp <- httr2::req_perform(cred$req_auth(req))
}
}
\section{Super class}{
\code{azr::Credential} -> \code{RefreshTokenCredential}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.refresh_token}}{Character string containing the refresh token.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RefreshTokenCredential-new}{\code{RefreshTokenCredential$new()}}
\item \href{#method-RefreshTokenCredential-validate}{\code{RefreshTokenCredential$validate()}}
\item \href{#method-RefreshTokenCredential-get_token}{\code{RefreshTokenCredential$get_token()}}
\item \href{#method-RefreshTokenCredential-req_auth}{\code{RefreshTokenCredential$req_auth()}}
\item \href{#method-RefreshTokenCredential-clone}{\code{RefreshTokenCredential$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="is_interactive"><a href='../../azr/html/Credential.html#method-Credential-is_interactive'><code>azr::Credential$is_interactive()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="Credential" data-id="print"><a href='../../azr/html/Credential.html#method-Credential-print'><code>azr::Credential$print()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RefreshTokenCredential-new"></a>}}
\if{latex}{\out{\hypertarget{method-RefreshTokenCredential-new}{}}}
\subsection{Method \code{new()}}{
Create a new refresh token credential
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RefreshTokenCredential$new(
  refresh_token = default_refresh_token(),
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{refresh_token}}{A character string containing the refresh token.
Defaults to \code{\link[=default_refresh_token]{default_refresh_token()}} which reads from the
\code{AZURE_REFRESH_TOKEN} environment variable.}

\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to \code{NULL}.}

\item{\code{tenant_id}}{A character string specifying the Azure Active Directory
tenant ID. Defaults to \code{NULL}.}

\item{\code{client_id}}{A character string specifying the application (client) ID.
Defaults to \code{NULL}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{RefreshTokenCredential} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RefreshTokenCredential-validate"></a>}}
\if{latex}{\out{\hypertarget{method-RefreshTokenCredential-validate}{}}}
\subsection{Method \code{validate()}}{
Validate the credential configuration
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RefreshTokenCredential$validate()}\if{html}{\out{</div>}}
}

\subsection{Details}{
Checks that the refresh token is provided and not NA or NULL. Calls the
parent class validation method.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RefreshTokenCredential-get_token"></a>}}
\if{latex}{\out{\hypertarget{method-RefreshTokenCredential-get_token}{}}}
\subsection{Method \code{get_token()}}{
Get an access token using the refresh token flow
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RefreshTokenCredential$get_token()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing the access token
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RefreshTokenCredential-req_auth"></a>}}
\if{latex}{\out{\hypertarget{method-RefreshTokenCredential-req_auth}{}}}
\subsection{Method \code{req_auth()}}{
Add OAuth refresh token authentication to an httr2 request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RefreshTokenCredential$req_auth(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The request object with OAuth refresh token authentication configured
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RefreshTokenCredential-clone"></a>}}
\if{latex}{\out{\hypertarget{method-RefreshTokenCredential-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RefreshTokenCredential$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{azure_client}
\alias{azure_client}
\title{Azure Default Client Configuration}
\format{
An object of class \code{list} of length 2.
}
\usage{
azure_client
}
\description{
Default client ID and tenant ID used for Azure authentication when not
explicitly provided. The client ID is Microsoft's public Azure CLI client ID.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_is_login}
\alias{az_cli_is_login}
\title{Check if User is Logged in to Azure CLI}
\usage{
az_cli_is_login(timeout = 10L)
}
\arguments{
\item{timeout}{A numeric value specifying the timeout in seconds for the
Azure CLI command. Defaults to \code{10}.}
}
\value{
A logical value: \code{TRUE} if the user is logged in, \code{FALSE} otherwise
}
\description{
Checks whether the user is currently logged in to Azure CLI by attempting
to retrieve account information.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azure-storage-client.R
\name{api_storage_client}
\alias{api_storage_client}
\title{Azure Storage API Class}
\description{
An R6 class that extends \link{api_client} to provide specialized methods
for Azure Data Lake Storage Gen2 (ADLS Gen2) REST API operations.
}
\details{
The base URL is constructed as:
\verb{https://\{storageaccount\}.\{endpoint_suffix\}}
}
\examples{
\dontrun{
# Create a storage client
storage <- api_storage_client$new(
  storageaccount = "mystorageaccount",
  filesystem = "mycontainer"
)

# List files in the root directory
files <- storage$list_files()

# List files in a specific path
files <- storage$list_files(path = "data/folder1")

# List files recursively
files <- storage$list_files(path = "data", recursive = TRUE)
}
}
\section{Super class}{
\code{\link[azr:api_client]{azr::api_client}} -> \code{api_storage_client}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.filesystem}}{The filesystem (container) name}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-api_storage_client-new}{\code{api_storage_client$new()}}
\item \href{#method-api_storage_client-download_file}{\code{api_storage_client$download_file()}}
\item \href{#method-api_storage_client-get_access_control}{\code{api_storage_client$get_access_control()}}
\item \href{#method-api_storage_client-list_files}{\code{api_storage_client$list_files()}}
\item \href{#method-api_storage_client-clone}{\code{api_storage_client$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".build_request"><a href='../../azr/html/api_client.html#method-api_client-.build_request'><code>azr::api_client$.build_request()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".fetch"><a href='../../azr/html/api_client.html#method-api_client-.fetch'><code>azr::api_client$.fetch()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".get_token"><a href='../../azr/html/api_client.html#method-api_client-.get_token'><code>azr::api_client$.get_token()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".resp_body_content"><a href='../../azr/html/api_client.html#method-api_client-.resp_body_content'><code>azr::api_client$.resp_body_content()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".resp_content"><a href='../../azr/html/api_client.html#method-api_client-.resp_content'><code>azr::api_client$.resp_content()</code></a></span></li>
<li><span class="pkg-link" data-pkg="azr" data-topic="api_client" data-id=".send_request"><a href='../../azr/html/api_client.html#method-api_client-.send_request'><code>azr::api_client$.send_request()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_storage_client-new"></a>}}
\if{latex}{\out{\hypertarget{method-api_storage_client-new}{}}}
\subsection{Method \code{new()}}{
Create a new Azure Storage API client instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_storage_client$new(
  storageaccount,
  filesystem,
  scope = default_azure_scope("azure_storage"),
  endpoint_suffix = default_storage_endpoint(),
  provider = NULL,
  chain = NULL,
  tenant_id = NULL,
  ...
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{storageaccount}}{A character string specifying the Azure Storage account name.}

\item{\code{filesystem}}{A character string specifying the filesystem (container) name.}

\item{\code{scope}}{A character string specifying the OAuth2 scope. Defaults to
\code{default_azure_scope("azure_storage")}.}

\item{\code{endpoint_suffix}}{A character string specifying the Azure
Storage DFS endpoint suffix. Defaults to
\code{\link[=default_storage_endpoint]{default_storage_endpoint()}}.}

\item{\code{provider}}{An optional credential provider object that inherits from
\code{Credential} or \code{DefaultCredential}. If provided, \code{chain} is ignored.}

\item{\code{chain}}{A \link{credential_chain} instance for authentication. If NULL,
a default credential chain will be created using \link{DefaultCredential}.}

\item{\code{tenant_id}}{A character string specifying the Azure tenant ID. Passed to
\link{DefaultCredential} when \code{chain} is \code{NULL}.}

\item{\code{...}}{Additional arguments passed to the parent \link{api_client} constructor.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{api_storage_client} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_storage_client-download_file"></a>}}
\if{latex}{\out{\hypertarget{method-api_storage_client-download_file}{}}}
\subsection{Method \code{download_file()}}{
Download a file from the filesystem
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_storage_client$download_file(path, dest = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{A character string specifying the file path within the filesystem.}

\item{\code{dest}}{A character string specifying the local destination path.
Defaults to a temporary file via \code{\link[=tempfile]{tempfile()}}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The local path the file was written to (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_storage_client-get_access_control"></a>}}
\if{latex}{\out{\hypertarget{method-api_storage_client-get_access_control}{}}}
\subsection{Method \code{get_access_control()}}{
Get the access control list (ACL) for a file or directory
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_storage_client$get_access_control(dataset, upn = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dataset}}{A character string specifying the file or directory path within
the filesystem.}

\item{\code{upn}}{A logical value. If \code{TRUE}, user principal names (UPN) are
returned in the \code{x-ms-owner}, \code{x-ms-group}, and \code{x-ms-acl} response
headers instead of object IDs. Defaults to \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A data.frame with columns \code{group_id} and \code{permission}, one row per
named group entry in the \code{x-ms-acl} response header.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_storage_client-list_files"></a>}}
\if{latex}{\out{\hypertarget{method-api_storage_client-list_files}{}}}
\subsection{Method \code{list_files()}}{
List files and directories in a path
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_storage_client$list_files(path = "", recursive = FALSE, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{A character string specifying the directory path to list.
Use empty string or NULL for the root directory. Defaults to \code{""}.}

\item{\code{recursive}}{A logical value indicating whether to list files recursively.
Defaults to \code{FALSE}.}

\item{\code{...}}{Additional query parameters to pass to the API.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A data.frame (or data.table if available) containing file and directory
information with columns such as name, contentLength, lastModified, etc.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_storage_client-clone"></a>}}
\if{latex}{\out{\hypertarget{method-api_storage_client-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_storage_client$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{default_azure_scope}
\alias{default_azure_scope}
\title{Get default Azure OAuth scope}
\usage{
default_azure_scope(resource = "azure_arm")
}
\arguments{
\item{resource}{A character string specifying the Azure resource. Must be one of:
\code{"azure_arm"} (Azure Resource Manager), \code{"azure_graph"} (Microsoft Graph),
\code{"azure_storage"} (Azure Storage), or \code{"azure_key_vault"} (Azure Key Vault).
Defaults to \code{"azure_arm"}.}
}
\value{
A character string with the OAuth scope URL
}
\description{
Returns the default OAuth scope for a specified Azure resource.
}
\examples{
default_azure_scope()
default_azure_scope("azure_graph")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/azr-package.R
\docType{package}
\name{azr-package}
\alias{azr}
\alias{azr-package}
\title{azr: Credential Chain for Seamless 'OAuth 2.0' Authentication to 'Azure Services'}
\description{
Implements a credential chain for 'Azure OAuth 2.0' authentication based on the package 'httr2''s 'OAuth' framework. Sequentially attempts authentication methods until one succeeds. During development allows interactive browser-based flows ('Device Code' and 'Auth Code' flows) and non-interactive flow ('Client Secret') in batch mode.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pedrobtz.github.io/azr/}
  \item \url{https://github.com/pedrobtz/azr}
  \item Report bugs at \url{https://github.com/pedrobtz/azr/issues}
}

}
\author{
\strong{Maintainer}: Pedro Baltazar \email{pedrobtz@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential-azure-cli.R
\name{az_cli_get_token}
\alias{az_cli_get_token}
\title{Get Access Token from Azure CLI}
\usage{
az_cli_get_token(scope, tenant_id = NULL, timeout = 10L)
}
\arguments{
\item{scope}{A character string specifying the OAuth2 scope for which to
request the access token (e.g., \code{"https://management.azure.com/.default"}).}

\item{tenant_id}{A character string specifying the Azure Active Directory
tenant ID. If \code{NULL}, uses the default tenant from Azure CLI. Defaults to \code{NULL}.}

\item{timeout}{A numeric value specifying the timeout in seconds for the
Azure CLI process. Defaults to \code{10}.}
}
\value{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object containing:
\itemize{
\item \code{access_token}: The OAuth2 access token string
\item \code{token_type}: The type of token (typically "Bearer")
\item \code{.expires_at}: POSIXct timestamp when the token expires
}
}
\description{
Retrieves an access token from Azure CLI using the \verb{az account get-access-token}
command. This is a lower-level function that directly interacts with the Azure
CLI to obtain OAuth2 tokens.
}
\details{
This function executes the Azure CLI command and parses the JSON response to
create an httr2 OAuth token object. The token includes the access token,
token type, and expiration time.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-credential.R
\name{get_credential_provider}
\alias{get_credential_provider}
\title{Get Credential Provider}
\usage{
get_credential_provider(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  oauth_host = NULL,
  oauth_endpoint = NULL,
  chain = NULL,
  interactive = TRUE,
  verbose = getOption("azr.verbose", FALSE)
)
}
\arguments{
\item{scope}{Optional character string specifying the authentication scope.}

\item{tenant_id}{Optional character string specifying the tenant ID for
authentication.}

\item{client_id}{Optional character string specifying the client ID for
authentication.}

\item{client_secret}{Optional character string specifying the client secret
for authentication.}

\item{use_cache}{Character string indicating the caching strategy. Defaults
to \code{"disk"}. Options include \code{"disk"} for disk-based caching or \code{"memory"}
for in-memory caching.}

\item{offline}{Logical. If \code{TRUE}, adds 'offline_access' to the scope to
request a 'refresh_token'. Defaults to \code{TRUE}.}

\item{oauth_host}{Optional character string specifying the OAuth host URL.}

\item{oauth_endpoint}{Optional character string specifying the OAuth endpoint.}

\item{chain}{A list of credential objects, where each element must inherit
from the \code{Credential} base class. Credentials are attempted in the order
provided until \code{get_token} succeeds. If \code{NULL}, uses
\code{\link[=default_credential_chain]{default_credential_chain()}}.}

\item{interactive}{A logical value indicating whether interactive credentials
are allowed. Defaults to \code{TRUE}.}

\item{verbose}{A logical value indicating whether to print verbose messages
during credential discovery. Defaults to \code{getOption("azr.verbose", FALSE)}.}
}
\value{
A credential object that inherits from the \code{Credential} class and
has successfully authenticated.
}
\description{
Discovers and returns an authenticated credential object from a chain of
credential providers. This function attempts each credential in the chain
until one successfully authenticates, returning the first successful
credential object.
}
\examples{
\dontrun{
# Get a credential provider with default settings
cred <- get_credential_provider(
  scope = "https://graph.microsoft.com/.default",
  tenant_id = "my-tenant-id"
)

# Use the credential to get a token
token <- cred$get_token()
}

}
\seealso{
\code{\link[=get_token_provider]{get_token_provider()}}, \code{\link[=get_request_authorizer]{get_request_authorizer()}},
\code{\link[=default_credential_chain]{default_credential_chain()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-credential.R
\name{get_request_authorizer}
\alias{get_request_authorizer}
\title{Get Default Request Authorizer Function}
\usage{
get_request_authorizer(
  scope = NULL,
  tenant_id = NULL,
  client_id = NULL,
  client_secret = NULL,
  use_cache = "disk",
  offline = TRUE,
  chain = default_credential_chain()
)
}
\arguments{
\item{scope}{Optional character string specifying the authentication scope.}

\item{tenant_id}{Optional character string specifying the tenant ID for
authentication.}

\item{client_id}{Optional character string specifying the client ID for
authentication.}

\item{client_secret}{Optional character string specifying the client secret
for authentication.}

\item{use_cache}{Character string indicating the caching strategy. Defaults
to \code{"disk"}. Options include \code{"disk"} for disk-based caching or \code{"memory"}
for in-memory caching.}

\item{offline}{Logical. If \code{TRUE}, adds 'offline_access' to the scope to request a 'refresh_token'.
Defaults to \code{TRUE}.}

\item{chain}{A list of credential objects, where each element must inherit
from the \code{Credential} base class. Credentials are attempted in the order
provided until \code{get_token} succeeds.}
}
\value{
A function that authorizes HTTP requests with appropriate credentials
when called.
}
\description{
Creates a request authorizer function that retrieves authentication credentials
and returns a callable request authorization method. This function handles the
credential discovery process and returns the request authentication method
from the discovered credential object.
}
\examples{
# In non-interactive sessions, this function will return an error if the
# environment is not setup with valid credentials. And in an interactive session
# the user will be prompted to attempt one of the interactive authentication flows.
\dontrun{
req_auth <- get_request_authorizer(
  scope = "https://graph.microsoft.com/.default"
)
req <- req_auth(httr2::request("https://graph.microsoft.com/v1.0/me"))
}

}
\seealso{
\code{\link[=get_token_provider]{get_token_provider()}}, \code{\link[=get_token]{get_token()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api-client.R
\name{api_client}
\alias{api_client}
\title{Azure API Client}
\description{
An R6 class that provides a base HTTP client for interacting with Azure APIs.
This client handles authentication, request building, retry logic, logging,
and error handling for Azure API requests.
}
\details{
The \code{api_client} class is designed to be a base class for Azure service-specific
clients. It provides:
\itemize{
\item Automatic authentication using Azure credentials
\item Configurable retry logic with exponential backoff
\item Request and response logging
\item JSON, XML, and HTML content type handling
\item Standardized error handling
}
}
\examples{
\dontrun{
# Create a client with default credentials
client <- api_client$new(
  host_url = "https://management.azure.com"
)

# Create a client with a credential provider
cred_provider <- get_credential_provider(
  scope = "https://management.azure.com/.default"
)
client <- api_client$new(
  host_url = "https://management.azure.com",
  provider = cred_provider
)

# Create a client with custom credentials function
client <- api_client$new(
  host_url = "https://management.azure.com",
  credentials = my_credential_function,
  timeout = 120,
  max_tries = 3
)

# Create a client with custom response handler
custom_handler <- function(content) {
  # Custom processing logic - e.g., keep data frames as-is
  content
}
client <- api_client$new(
  host_url = "https://management.azure.com",
  response_handler = custom_handler
)

# Make a GET request
response <- client$.fetch(
  path = "/subscriptions/{subscription_id}/resourceGroups",
  subscription_id = "my-subscription-id",
  query = list(`api-version` = "2021-04-01"),
  method = "get"
)
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{.host_url}}{Base URL for the API}

\item{\code{.base_req}}{Base httr2 request object}

\item{\code{.provider}}{Credential provider R6 object}

\item{\code{.credentials}}{Credentials function for authentication}

\item{\code{.options}}{Request options (timeout, connecttimeout, max_tries)}

\item{\code{.response_handler}}{Optional callback function to process response content}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-api_client-new}{\code{api_client$new()}}
\item \href{#method-api_client-.fetch}{\code{api_client$.fetch()}}
\item \href{#method-api_client-.resp_content}{\code{api_client$.resp_content()}}
\item \href{#method-api_client-.build_request}{\code{api_client$.build_request()}}
\item \href{#method-api_client-.send_request}{\code{api_client$.send_request()}}
\item \href{#method-api_client-.resp_body_content}{\code{api_client$.resp_body_content()}}
\item \href{#method-api_client-.get_token}{\code{api_client$.get_token()}}
\item \href{#method-api_client-clone}{\code{api_client$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-new"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-new}{}}}
\subsection{Method \code{new()}}{
Create a new API client instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$new(
  host_url,
  provider = NULL,
  credentials = NULL,
  timeout = 60L,
  connecttimeout = 30L,
  max_tries = 5L,
  response_handler = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{host_url}}{A character string specifying the base URL for the API
(e.g., \code{"https://management.azure.com"}).}

\item{\code{provider}}{An R6 credential provider object that inherits from the
\code{Credential} or \code{DefaultCredential} class. If provided, the credential's
\code{req_auth} method will be used for authentication. Takes precedence over
\code{credentials}.}

\item{\code{credentials}}{A function that adds authentication to requests. If
both \code{provider} and \code{credentials} are \code{NULL}, uses \code{\link[=default_non_auth]{default_non_auth()}}.
The function should accept an httr2 request object and return a modified
request with authentication.}

\item{\code{timeout}}{An integer specifying the request timeout in seconds.
Defaults to \code{60}.}

\item{\code{connecttimeout}}{An integer specifying the connection timeout in
seconds. Defaults to \code{30}.}

\item{\code{max_tries}}{An integer specifying the maximum number of retry
attempts for failed requests. Defaults to \code{5}.}

\item{\code{response_handler}}{An optional function to process the parsed response
content. The function should accept one argument (the parsed response) and
return the processed content. If \code{NULL}, uses \code{\link[=default_response_handler]{default_response_handler()}}
which converts data frames to data.table objects. Defaults to \code{NULL}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{api_client} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.fetch"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.fetch}{}}}
\subsection{Method \code{.fetch()}}{
Make an HTTP request to the API
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.fetch(
  path,
  ...,
  query = NULL,
  body = NULL,
  headers = NULL,
  method = "get",
  verbosity = 0L,
  content = c("body", "headers", "response", "request"),
  content_type = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{A character string specifying the API endpoint path. Supports
\code{\link[rlang:englue]{rlang::englue()}} syntax for variable interpolation using named arguments
passed via \code{...}.}

\item{\code{...}}{Named arguments used for path interpolation with \code{\link[rlang:englue]{rlang::englue()}}.}

\item{\code{query}}{A named list of query parameters to append to the URL.}

\item{\code{body}}{Request body data. Sent as JSON in the request body.
Can be a list or character string (JSON).}

\item{\code{headers}}{A named list of additional HTTP headers to include in the
request.}

\item{\code{method}}{A character string specifying the HTTP method. One of
\code{"get"}, \code{"post"}, \code{"put"}, \code{"patch"}, or \code{"delete"}. Defaults to \code{"get"}.}

\item{\code{verbosity}}{An integer specifying the verbosity level for request
debugging (passed to \code{\link[httr2:req_perform]{httr2::req_perform()}}). Defaults to \code{0}.}

\item{\code{content}}{A character string specifying what to return. One of:
\itemize{
\item \code{"body"} (default): Return the parsed response body
\item \code{"headers"}: Return response headers
\item \code{"response"}: Return the full httr2 response object
\item \code{"request"}: Return the prepared request object without executing it
}}

\item{\code{content_type}}{A character string specifying how to parse the response
body. If \code{NULL}, uses the response's Content-Type header. Common values:
\code{"application/json"}, \code{"application/xml"}, \code{"text/html"}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Depends on the \code{content} parameter:
\itemize{
\item \code{"body"}: Parsed response body (list, data.frame, or character)
\item \code{"headers"}: List of response headers
\item \code{"response"}: Full \code{\link[httr2:response]{httr2::response()}} object
\item \code{"request"}: \code{\link[httr2:request]{httr2::request()}} object
}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.resp_content"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.resp_content}{}}}
\subsection{Method \code{.resp_content()}}{
Extract content from a response object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.resp_content(resp, content, content_type = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{resp}}{An \code{\link[httr2:response]{httr2::response()}} object}

\item{\code{content}}{A character string specifying what to return. One of:
\itemize{
\item \code{"body"}: Return the parsed response body
\item \code{"headers"}: Return response headers
\item \code{"response"}: Return the full httr2 response object
}}

\item{\code{content_type}}{A character string specifying how to parse the response
body. Only used when \code{content = "body"}. If \code{NULL}, uses the response's
Content-Type header.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Depends on the \code{content} parameter:
\itemize{
\item \code{"body"}: Parsed response body (list, data.frame, or character)
\item \code{"headers"}: List of response headers
\item \code{"response"}: Full \code{\link[httr2:response]{httr2::response()}} object
}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.build_request"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.build_request}{}}}
\subsection{Method \code{.build_request()}}{
Build an HTTP request object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.build_request(
  path,
  ...,
  query = NULL,
  body = NULL,
  headers = NULL,
  method = "get"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{A character string specifying the API endpoint path. Supports
\code{\link[rlang:englue]{rlang::englue()}} syntax for variable interpolation using named arguments
passed via \code{...}.}

\item{\code{...}}{Named arguments used for path interpolation with \code{\link[rlang:englue]{rlang::englue()}}.}

\item{\code{query}}{A named list of query parameters to append to the URL.}

\item{\code{body}}{Request body data. Sent as JSON in the request body.
Can be a list or character string (JSON).}

\item{\code{headers}}{A named list of additional HTTP headers to include in the
request.}

\item{\code{method}}{A character string specifying the HTTP method. One of
\code{"get"}, \code{"post"}, \code{"put"}, \code{"patch"}, or \code{"delete"}. Defaults to \code{"get"}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{\link[httr2:request]{httr2::request()}} object ready for execution
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.send_request"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.send_request}{}}}
\subsection{Method \code{.send_request()}}{
Perform an HTTP request and log the results
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.send_request(req, verbosity)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{An \code{\link[httr2:request]{httr2::request()}} object to execute}

\item{\code{verbosity}}{An integer specifying the verbosity level for request
debugging (passed to \code{\link[httr2:req_perform]{httr2::req_perform()}}). Defaults to \code{0}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{\link[httr2:response]{httr2::response()}} object containing the API response
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.resp_body_content"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.resp_body_content}{}}}
\subsection{Method \code{.resp_body_content()}}{
Extract and parse response content
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.resp_body_content(resp, content_type = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{resp}}{An \code{\link[httr2:response]{httr2::response()}} object}

\item{\code{content_type}}{A character string specifying how to parse the response
body. If \code{NULL}, uses the response's Content-Type header. Common values:
\code{"application/json"}, \code{"application/xml"}, \code{"text/html"}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Parsed response body. Format depends on content type:
\itemize{
\item JSON: List or data.frame
\item XML: xml2 document
\item HTML: xml2 document
\item Other: Character string
}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-.get_token"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-.get_token}{}}}
\subsection{Method \code{.get_token()}}{
Get authentication token from the credential provider
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$.get_token()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An \code{\link[httr2:oauth_token]{httr2::oauth_token()}} object if a provider is available,
otherwise returns \code{NULL} with a warning.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-api_client-clone"></a>}}
\if{latex}{\out{\hypertarget{method-api_client-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{api_client$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
